/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_SKELETONS_PARALLEX_HPX_FIBONACCI_H_INCLUDED
#define SSTMAC_SOFTWARE_SKELETONS_PARALLEX_HPX_FIBONACCI_H_INCLUDED

#include <hpx/hpx_app.h>
#include <hpx/runtime/actions/hpx_action.h>
#include <sstmac/common/messages/valuepayload.h>
#include <sstmac/sstmacro.h>

namespace hpx
{
  namespace fibonacci
  {

    class hpx_fibonacci : public hpx_app
    {
    public:
      typedef boost::intrusive_ptr<hpx_fibonacci> ptr;

      hpx_fibonacci(sstmac::sw::softwareid id) :
          hpx_app(id)
      {

      }

      virtual
      ~hpx_fibonacci()
      {
      }

      virtual void
      hpx_main();

      static ptr
      construct(sstmac::sw::softwareid id)
      {
        ptr retval(new hpx_fibonacci(id));
        return retval;
      }

      virtual app::ptr
      clone(sstmac::sw::softwareid id)
      {
        return construct(id);
      }

      virtual void
      consume_params(const sstmac::sim_parameters::ptr &params);

      virtual std::string
      to_string() const
      {
        return "hpx_fibonacci";
      }


    protected:
      long n_value_;

      virtual void
      register_actions();

      virtual void
      register_components();

    };
  }
} //end of namespace sstmac

#endif
