/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_SKELETONS_PARALLEX_HPX_RANDOM_MEM_ACCESS_H_
#define SSTMAC_SOFTWARE_SKELETONS_PARALLEX_HPX_RANDOM_MEM_ACCESS_H_

#include <hpx/hpx_app.h>

#include <sstmac/common/rng.h>
#include <sstmac/sstmacro.h>

namespace hpx
{
  namespace rma
  {

    class hpx_random_mem_access : public hpx::hpx_app
    {
    protected:
      int array_size_;
      int iterations_;
      int seed_;

      RNG::SHR3::ptr rng_;

      class hpx_rma_element;

      virtual void
      register_actions();

      virtual void
      register_components();

    public:
      typedef boost::intrusive_ptr<hpx_random_mem_access> ptr;

      hpx_random_mem_access(sstmac::sw::softwareid id) :
          hpx_app(id)
      {

      }

      virtual
      ~hpx_random_mem_access()
      {
      }

      virtual void
      hpx_main();

      static ptr
      construct(sstmac::sw::softwareid id)
      {
        ptr retval(new hpx_random_mem_access(id));
        return retval;
      }

      virtual app::ptr
      clone(sstmac::sw::softwareid id)
      {
        return construct(id);
      }

      virtual void
      consume_params(const sstmac::sim_parameters::ptr &params);

      virtual std::string
      to_string() const
      {
        return "hpx_random_mem_access";
      }

    };

  } 
}
#endif /* HPXDATAFLOW_H_ */
