//#if HAVE_CONFIG_H
# include "sstmac/common/sstmac-config.h"
//#endif

#include "tascel/tascel.h"
#include "ftypes.h"
#include "gacpp.h"

#include <mpi/sstmac_mpi.h>

#include <cmath>
#include <iostream>

using namespace tascel;
#define MT_DBL 0
#define MT_INT 1
#define ichunk 20

static ProcGroup* pgrp;

extern "C"
{
  void
  rsg_(Integer* nm, Integer* n, DoublePrecision* a, DoublePrecision* b,
      DoublePrecision* w, Integer* matz, DoublePrecision* z,
      DoublePrecision* fv1, DoublePrecision* fv2, Integer* ierr);
  void
  rs_(Integer* nm, Integer* n, DoublePrecision* a, DoublePrecision* w,
      Integer* matz, DoublePrecision* z, DoublePrecision* fv1,
      DoublePrecision* fv2, Integer* ierr);
  void
  dgemm_(char* transa, char* transb, Integer* m, Integer* n, Integer* k,
      DoublePrecision* alpha, DoublePrecision* a, Integer* lda,
      DoublePrecision* b, Integer* ldb, DoublePrecision* beta,
      DoublePrecision* c, Integer* ldc);
}

Integer
gaCreateHandle()
{
  Integer han = theRegistry().addEntry<void*>(0);
  DEBUG_PRINT("gaCreateHandle: han = %d\n", han);
  amBarrier();
  return han;
}

/*void tascelInit() {
 TascelConfig::initialize(1, MPI_COMM_WORLD);
 pgrp = ProcGroup::construct();
 #if defined(DATA_TRACE)
 traceEventTypes["gaCreate"] = trace_proj_register_event("gaCreate");
 traceEventTypes["gaZero"] = trace_proj_register_event("gaZero");
 traceEventTypes["gaReadInc"] = trace_proj_register_event("gaReadInc");
 traceEventTypes["gaSymmetrize"] = trace_proj_register_event("gaSymmetrize");
 traceEventTypes["gaDiagSeq"] = trace_proj_register_event("gaDiagSeq");
 traceEventTypes["gaDiagStdSeq"] = trace_proj_register_event("gaDiagStdSeq");
 traceEventTypes["gaDgemm"] = trace_proj_register_event("gaDgemm");
 traceEventTypes["gaCopy"] = trace_proj_register_event("gaCopy");
 traceEventTypes["gaAdd"] = trace_proj_register_event("gaAdd");
 #endif
 }

 void tascelFinalize() {
 TascelConfig::finalize();
 delete pgrp;
 }*/

Integer
getNodeID()
{
  Integer nodeID = static_cast<Integer>(theTwoSided().getProcRank().toInt());
  return nodeID;
}

Integer
getNProc()
{
  return static_cast<Integer>(theTwoSided().numProcRanks().toInt());
}

void
gaCreate(Integer g_a, Integer ndim, Integer *dims, Integer type)
{
#if defined(DATA_TRACE)
  trace_proj_event_start(Timer::timestamp(), traceEventTypes["gaCreate"]);
#endif

  void*& pairEntry = theRegistry().lookup<void*>(g_a);

  size_t sizeType = type == MT_INT ? sizeof(Integer) : sizeof(DoublePrecision);

  amBarrier();

  DEBUG_PRINT("gaCreate: %dd array, handle = %d\n", ndim, g_a);

  if (ndim == 1)
    {
      Array<Offset, 1> adims, blocks;
      adims[0] = dims[0];
      blocks[0] = ichunk;

      Array<int, 1> proc_grid;
      proc_grid[0] = theTwoSided().numProcRanks().toInt();
      ;

      //DEBUG_PRINT("allocating 1d array, handle = %d\n", g_a);
      GArray<1>* garray = new GArray<1>(adims, blocks, proc_grid,
          kBlockCyclicMultiBlock, kRowMajor, sizeType);
      std::pair<int, void*>* p = new std::pair<int, void*>();
      p->first = 1;
      p->second = garray;
      pairEntry = p;
    }
  else if (ndim == 2)
    {
      Array<Offset, 2> adims, blocks;
      adims[0] = dims[0];
      adims[1] = dims[1];
      assert(sizeof(Integer) == sizeof(long));
      //DEBUG_PRINT("dims[0] = %d, dims[1] = %d\n", dims[0], ((long*)dims)[1]);
      blocks[0] = ichunk;
      blocks[1] = ichunk;

      int nproc = theTwoSided().numProcRanks().toInt();
      int dim1, dim2 = nproc;
      for (dim1 = (int) sqrt((double) nproc); dim1 > 1; dim1--)
        {
          if (nproc % dim1 == 0)
            {
              dim2 = nproc / dim1;
              break;
            }
        }

      DEBUG_PRINT("gaCreate: dim1 = %d, dim2 = %d\n", dim1, dim2);

      Array<int, 2> proc_grid;
      proc_grid[0] = dim1;
      proc_grid[1] = dim2;

      //DEBUG_PRINT("allocating 2d array, handle = %d\n", g_a);
      GArray<2>* garray = new GArray<2>(adims, blocks, proc_grid,
          kBlockCyclicMultiBlock, kColumnMajor, sizeType);

      std::pair<int, void*>* p = new std::pair<int, void*>();
      p->first = 2;
      p->second = garray;
      pairEntry = p;
    }
  else
    {
      std::cerr << "dimensionality not implemented" << std::endl;
      exit(1);
    }

  amBarrier();

#if defined(DATA_TRACE)
  trace_proj_event_stop(Timer::timestamp(), traceEventTypes["gaCreate"]);
#endif
}

void
gaDestroy(Integer g_a)
{
  std::pair<int, void*>* p = theRegistry().lookup<std::pair<int, void*>*>(g_a);

  DEBUG_PRINT("gaDestroy %d\n", g_a);

  if (p->first == 1)
    {
      GArray<1>& ga = *reinterpret_cast<GArray<1>*>(p->second);
      delete &ga;
    }
  else if (p->first == 2)
    {
      GArray<2>& ga = *reinterpret_cast<GArray<2>*>(p->second);
      delete &ga;
    }
  else
    {
      std::cerr << "dimensionality not implemented" << std::endl;
      exit(1);
    }

  delete p;
  theRegistry().removeEntry(g_a);
}

void
gaZero(Integer g_a)
{
#if defined(DATA_TRACE)
  trace_proj_event_start(Timer::timestamp(), traceEventTypes["gaZero"]);
#endif

  amBarrier();

  std::pair<int, void*>* p = theRegistry().lookup<std::pair<int, void*>*>(g_a);
  DEBUG_PRINT("gaZero: g_a = %d (non-collective)\n", g_a);

  if (p->first == 1)
    {
      GArray<1>& ga = *reinterpret_cast<GArray<1>*>(p->second);
      ga.zero();
    }
  else if (p->first == 2)
    {
      GArray<2>& ga = *reinterpret_cast<GArray<2>*>(p->second);
      ga.zero();
    }
  else
    {
      std::cerr << "dimensionality not implemented" << std::endl;
      exit(1);
    }

  amBarrier();

#if defined(DATA_TRACE)
  trace_proj_event_stop(Timer::timestamp(), traceEventTypes["gaZero"]);
#endif
}

Integer
gaReadInc(Integer g_a, Integer subscript, Integer inc)
{
#if defined(DATA_TRACE)
  trace_proj_event_start(Timer::timestamp(), traceEventTypes["gaReadInc"]);
#endif

  std::pair<int, void*>* p = theRegistry().lookup<std::pair<int, void*>*>(g_a);
  //DEBUG_PRINT("calling read inc on g_a = %d\n", g_a);

  if (p->first == 1)
    {
      GArray<1>& ga = *reinterpret_cast<GArray<1>*>(p->second);
      int ret = -1;
      GArrayRequest* req1 = GArrayRequest::construct(), *req2 =
          GArrayRequest::construct();

      ga.readInc<int>(&ret, 0, 1, req1, req2);

      Dispatcher<NullMutex> dispatcher;
      dispatcher.registerCodelet(req1);
      dispatcher.registerCodelet(req2);

      while (!dispatcher.empty())
        {
          Codelet* codelet;
          if ((codelet = dispatcher.progress()) != NULL)
            {
              codelet->execute();
            }
          AmListenObjCodelet<NullMutex>* lcodelet;
          if ((lcodelet = theAm().amListeners[0]->progress()) != NULL)
            {
              lcodelet->execute();
            }
        }

      delete req1;
      delete req2;

      //DEBUG_PRINT("read inc g_a = %d, returing %d\n", g_a, ret);

      return ret;
    }
  else
    {
      std::cerr << "dimensionality not implemented" << std::endl;
      exit(1);
    }

#if defined(DATA_TRACE)
  trace_proj_event_stop(Timer::timestamp(), traceEventTypes["gaReadInc"]);
#endif
}

void
amBarrier()
{
  int epoch = pgrp->signalBarrier();
  while (!pgrp->testBarrier(epoch))
    {
      AmListenObjCodelet<NullMutex>* codelet;
      if ((codelet = theAm().amListeners[0]->progress()) != NULL)
        {
          codelet->execute();
        }
    }
}

void
gaPut(Integer g_a, Integer *lo, Integer *hi, void *buf, Integer *ld)
{
  std::pair<int, void*>* p = theRegistry().lookup<std::pair<int, void*>*>(g_a);

  if (p->first == 2)
    {
      GArray<2>& ga = *reinterpret_cast<GArray<2>*>(p->second);
      GArrayRequest* req1 = GArrayRequest::construct(), *req2 =
          GArrayRequest::construct();

      DEBUG_PRINT("gaPut: g_a = %d, lo = [%d,%d], hi = [%d,%d], ld = %d\n",
          g_a, lo[0], lo[1], hi[0], hi[1], ld[0]);

      Array<Offset, 2> glo, ghi;
      glo[0] = lo[0] - 1;
      ghi[0] = hi[0];
      glo[1] = lo[1] - 1;
      ghi[1] = hi[1];
      Box<2> gbox(glo, ghi);

      Array<Offset, 2> llo, lhi;
      llo[0] = 0;
      lhi[0] = ghi[0] - glo[0];
      llo[1] = 0;
      lhi[1] = ghi[1] - glo[1];
      Box<2> lbox(llo, lhi);

      assert(lhi[0] - llo[0] == *ld);

      //std::cout << "global box: " << gbox << std::endl;
      //std::cout << "local box: " << lbox << std::endl;

      ga.put(gbox, buf, lbox, lbox, req1, req2);

      Dispatcher<NullMutex> dispatcher;
      dispatcher.registerCodelet(req1);
      dispatcher.registerCodelet(req2);

      while (!dispatcher.empty())
        {
          Codelet* codelet;
          if ((codelet = dispatcher.progress()) != NULL)
            {
              codelet->execute();
            }
          AmListenObjCodelet<NullMutex>* lcodelet;
          if ((lcodelet = theAm().amListeners[0]->progress()) != NULL)
            {
              lcodelet->execute();
            }
        }

      delete req1;
      delete req2;
    }
  else
    {
      std::cerr << "dimensionality not implemented" << std::endl;
      exit(1);
    }
}

void
gaPutI(int g_a, int *lo, int *hi, void *buf, int *ld)
{
  std::pair<int, void*>* p = theRegistry().lookup<std::pair<int, void*>*>(g_a);

  if (p->first == 2)
    {
      Integer loi[2] =
        { lo[1] + 1, lo[0] + 1 };
      Integer hii[2] =
        { hi[1] + 1, hi[0] + 1 };
      Integer ldi[1] =
        { ld[0] };

      gaPut(g_a, loi, hii, buf, ldi);
    }
  else
    {
      assert(0);
    }
}

void
gaGetI(int g_a, int *lo, int *hi, void *buf, int *ld)
{
  std::pair<int, void*>* p = theRegistry().lookup<std::pair<int, void*>*>(g_a);

  if (p->first == 2)
    {
      Integer loi[2] =
        { lo[1] + 1, lo[0] + 1 };
      Integer hii[2] =
        { hi[1] + 1, hi[0] + 1 };
      Integer ldi[1] =
        { ld[0] };

      gaGet(g_a, loi, hii, buf, ldi);
    }
  else
    {
      assert(0);
    }
}

void
gaGet(Integer g_a, Integer *lo, Integer *hi, void *buf, Integer *ld)
{
  std::pair<int, void*>* p = theRegistry().lookup<std::pair<int, void*>*>(g_a);

  if (p->first == 2)
    {
      GArray<2>& ga = *reinterpret_cast<GArray<2>*>(p->second);
      GArrayRequest* req1 = GArrayRequest::construct(), *req2 =
          GArrayRequest::construct();

      DEBUG_PRINT("gaGet: g_a = %d, lo = [%d,%d], hi = [%d,%d], ld = %d\n",
          g_a, lo[0], lo[1], hi[0], hi[1], ld[0]);

      Array<Offset, 2> glo, ghi;
      glo[0] = lo[0] - 1;
      ghi[0] = hi[0];
      glo[1] = lo[1] - 1;
      ghi[1] = hi[1];
      Box<2> gbox(glo, ghi);

      Array<Offset, 2> llo, lhi;
      llo[0] = 0;
      lhi[0] = ghi[0] - glo[0];
      llo[1] = 0;
      lhi[1] = ghi[1] - glo[1];
      Box<2> lbox(llo, lhi);

      assert(lhi[0] - llo[0] == *ld);

      //std::cout << "global box: " << gbox << std::endl;
      //std::cout << "local box: " << lbox << std::endl;
      //std::cout << "ga.get:: gbox = " << gbox << ", lbox = " << lbox << std::endl;
      ga.get(gbox, buf, lbox, lbox, req1, req2);

      Dispatcher<NullMutex> dispatcher;
      dispatcher.registerCodelet(req1);
      dispatcher.registerCodelet(req2);

      while (!dispatcher.empty())
        {
          Codelet* codelet;
          if ((codelet = dispatcher.progress()) != NULL)
            {
              codelet->execute();
            }
          AmListenObjCodelet<NullMutex>* lcodelet;
          if ((lcodelet = theAm().amListeners[0]->progress()) != NULL)
            {
              lcodelet->execute();
            }
        }

      delete req1;
      delete req2;
    }
  else
    {
      std::cerr << "dimensionality not implemented" << std::endl;
      exit(1);
    }
}

void
gaAccI(int g_a, int *lo, int *hi, void *buf, int *ld)
{
  std::pair<int, void*>* p = theRegistry().lookup<std::pair<int, void*>*>(g_a);

  if (p->first == 2)
    {
      Integer loi[2] =
        { lo[1] + 1, lo[0] + 1 };
      Integer hii[2] =
        { hi[1] + 1, hi[0] + 1 };
      Integer ldi[1] =
        { ld[0] };

      gaAcc(g_a, loi, hii, buf, ldi);
    }
  else
    {
      assert(0);
    }
}

void
gaAcc(Integer g_a, Integer *lo, Integer *hi, void *buf, Integer *ld)
{
  std::pair<int, void*>* p = theRegistry().lookup<std::pair<int, void*>*>(g_a);

  if (p->first == 2)
    {
      GArray<2>& ga = *reinterpret_cast<GArray<2>*>(p->second);
      GArrayRequest* req1 = GArrayRequest::construct(), *req2 =
          GArrayRequest::construct();

      DEBUG_PRINT("gaAcc: g_a = %d, lo = [%d,%d], hi = [%d,%d], ld = %d\n",
          g_a, lo[0], lo[1], hi[0], hi[1], ld[0]);

      Array<Offset, 2> glo, ghi;
      glo[0] = lo[0] - 1;
      ghi[0] = hi[0];
      glo[1] = lo[1] - 1;
      ghi[1] = hi[1];
      Box<2> gbox(glo, ghi);

      Array<Offset, 2> llo, lhi;
      llo[0] = 0;
      lhi[0] = ghi[0] - glo[0];
      llo[1] = 0;
      lhi[1] = ghi[1] - glo[1];
      Box<2> lbox(llo, lhi);

      assert(lhi[0] - llo[0] == *ld);

      ga.acc<Rma::DoubleSum>(gbox, buf, lbox, lbox, req1, req2);

      Dispatcher<NullMutex> dispatcher;
      dispatcher.registerCodelet(req1);
      dispatcher.registerCodelet(req2);

      while (!dispatcher.empty())
        {
          Codelet* codelet;
          if ((codelet = dispatcher.progress()) != NULL)
            {
              codelet->execute();
            }
          AmListenObjCodelet<NullMutex>* lcodelet;
          if ((lcodelet = theAm().amListeners[0]->progress()) != NULL)
            {
              lcodelet->execute();
            }
        }

      delete req1;
      delete req2;
    }
  else
    {
      std::cerr << "dimensionality not implemented" << std::endl;
      exit(1);
    }
}

void
gaSymmetrize(Integer g_a)
{
  // @todo Assume that is this was called collectively and only one PE
  // should actually execute

  // @todo Assume that g_a is a array of doubles

  amBarrier();

  if (theTwoSided().getProcRank() != 0)
    {
      return;
    }

#if defined(DATA_TRACE)
  trace_proj_event_start(Timer::timestamp(), traceEventTypes["gaSymmetrize"]);
#endif

  DEBUG_PRINT("gaSymmetrize: g_a = %d\n", g_a);

  std::pair<int, void*>* p = theRegistry().lookup<std::pair<int, void*>*>(g_a);

  if (p->first == 2)
    {
      GArray<2>& ga = *reinterpret_cast<GArray<2>*>(p->second);
      Array<Offset, 2> lo, hi;
      lo[0] = lo[1] = 0;
      hi[0] = ga.procGrid_.cg_.dims_[0];
      hi[1] = ga.procGrid_.cg_.dims_[1];
      Box<2> gbox(lo, hi);

      assert(hi[0] == hi[1]);
      assert(ga.elsize_ == sizeof(double));

      //DEBUG_PRINT("whole box size = %d, byte count = %d\n",
      //gbox.size(), gbox.size() * ga.elsize_.toInt());

      double* ptr = (double*) alloca(ga.elsize_.toInt() * gbox.size());
      GArrayRequest* req1 = GArrayRequest::construct(), *req2 =
          GArrayRequest::construct();

      //std::cout << "global box: " << gbox << std::endl;

      ga.get(gbox, ptr, gbox, gbox, req1, req2);

      Dispatcher<NullMutex> dispatcher;
      dispatcher.registerCodelet(req1);
      dispatcher.registerCodelet(req2);

      while (!dispatcher.empty())
        {
          Codelet* codelet;
          if ((codelet = dispatcher.progress()) != NULL)
            {
              codelet->execute();
            }
          AmListenObjCodelet<NullMutex>* lcodelet;
          if ((lcodelet = theAm().amListeners[0]->progress()) != NULL)
            {
              lcodelet->execute();
            }
        }

      double* copyPtr = (double*) alloca(ga.elsize_.toInt() * gbox.size());

      for (unsigned i = 0; i < hi[0]; i++)
        {
          for (unsigned j = 0; j < hi[1]; j++)
            {
              copyPtr[i * hi[0] + j] = ptr[j * hi[0] + i];
            }
        }

      for (unsigned i = 0; i < gbox.size(); i++)
        {
          ptr[i] = ptr[i] * 0.5 + copyPtr[i] * 0.5;
        }

      ga.put(gbox, ptr, gbox, gbox, req1, req2);

      dispatcher.registerCodelet(req1);
      dispatcher.registerCodelet(req2);

      while (!dispatcher.empty())
        {
          Codelet* codelet;
          if ((codelet = dispatcher.progress()) != NULL)
            {
              codelet->execute();
            }
          AmListenObjCodelet<NullMutex>* lcodelet;
          if ((lcodelet = theAm().amListeners[0]->progress()) != NULL)
            {
              lcodelet->execute();
            }
        }

      delete req1;
      delete req2;

      DEBUG_PRINT("gaSymmetrize: finished g_a = %d\n", g_a);
    }
  else
    {
      std::cerr << "dimensionality not implemented" << std::endl;
      exit(1);
    }

#if defined(DATA_TRACE)
  trace_proj_event_stop(Timer::timestamp(), traceEventTypes["gaSymmetrize"]);
#endif
}

void
gaDiagSeq(Integer g_a, Integer g_s, Integer g_v, DoublePrecision *evals)
{
#if defined(DATA_TRACE)
  trace_proj_event_start(Timer::timestamp(), traceEventTypes["gaDiagSeq"]);
#endif

  std::pair<int, void*>* p_a = theRegistry().lookup<std::pair<int, void*>*>(
      g_a);
  std::pair<int, void*>* p_s = theRegistry().lookup<std::pair<int, void*>*>(
      g_s);
  std::pair<int, void*>* p_v = theRegistry().lookup<std::pair<int, void*>*>(
      g_v);

  if (theTwoSided().getProcRank() != 0)
    {
      assert(p_a->first == 2);
      GArray<2>& ga_a = *reinterpret_cast<GArray<2>*>(p_a->second);
      Offset len = ga_a.procGrid_.cg_.dims_[0];
      amBarrier();
      DEBUG_PRINT("%d: bcast\n", theTwoSided().getProcRank().toInt());

      MPI::Set_Payload<DoublePrecision>(*evals);
      MPI::Bcast(NULL, len, MPI_DOUBLE, 0, MPI_COMM_WORLD);
      *evals = MPI::Get_Payload<DoublePrecision>();

#if defined(DATA_TRACE)
      trace_proj_event_stop(Timer::timestamp(), traceEventTypes["gaDiagSeq"]);
#endif
      return;
    }

  if (p_a->first == 2 && p_s->first == 2 && p_v->first == 2)
    {
      DEBUG_PRINT("gaDiagSeq: [%d,%d,%d]\n", g_a, g_s, g_v);
      GArray<2>& ga_a = *reinterpret_cast<GArray<2>*>(p_a->second);

      Integer lo[2] =
        { 1, 1 };
      Integer hi[2] =
        { ga_a.procGrid_.cg_.dims_[0], ga_a.procGrid_.cg_.dims_[1] };

      assert(hi[0] == hi[1]);
      assert(ga_a.elsize_ == sizeof(double));

      DoublePrecision* k_a = new DoublePrecision[hi[0] * hi[1]];
      DoublePrecision* k_s = new DoublePrecision[hi[0] * hi[1]];
      DoublePrecision* k_v = new DoublePrecision[hi[0] * hi[1]];
      DoublePrecision *k_fv1 = new DoublePrecision[hi[0]];
      DoublePrecision *k_fv2 = new DoublePrecision[hi[0]];

      gaGet(g_a, lo, hi, k_a, hi);
      gaGet(g_s, lo, hi, k_s, hi);

      Integer nm = hi[0], n = hi[0], matz = 1, ierr = 0;

      rsg_(&nm, &n, k_a, k_s, evals, &matz, k_v, k_fv1, k_fv2, &ierr);

      gaPut(g_v, lo, hi, k_v, hi);

      delete[] k_a;
      delete[] k_s;
      delete[] k_v;
      delete[] k_fv1;
      delete[] k_fv2;

      amBarrier();

      DEBUG_PRINT("%d: bcast\n", theTwoSided().getProcRank().toInt());
      MPI::Set_Payload<DoublePrecision>(*evals);
      MPI::Bcast(NULL, hi[0], MPI_DOUBLE, 0, MPI_COMM_WORLD);
      *evals = MPI::Get_Payload<DoublePrecision>();
      DEBUG_PRINT("gaDiagSeq: finished on %d\n", g_a);
    }
  else
    {
      std::cerr << "dimensionality not implemented" << std::endl;
      exit(1);
    }

#if defined(DATA_TRACE)
  trace_proj_event_stop(Timer::timestamp(), traceEventTypes["gaDiagSeq"]);
#endif
}

void
gaDiagStdSeq(Integer g_a, Integer g_v, DoublePrecision *evals)
{
#if defined(DATA_TRACE)
  trace_proj_event_start(Timer::timestamp(), traceEventTypes["gaDiagStdSeq"]);
#endif

  std::pair<int, void*>* p_a = theRegistry().lookup<std::pair<int, void*>*>(
      g_a);
  std::pair<int, void*>* p_v = theRegistry().lookup<std::pair<int, void*>*>(
      g_v);

  if (theTwoSided().getProcRank() != 0)
    {
      assert(p_a->first == 2);
      GArray<2>& ga_a = *reinterpret_cast<GArray<2>*>(p_a->second);
      Offset len = ga_a.procGrid_.cg_.dims_[0];
      amBarrier();
      MPI::Set_Payload<DoublePrecision>(*evals);
      MPI::Bcast(NULL, len, MPI_DOUBLE, 0, MPI_COMM_WORLD);
      *evals = MPI::Get_Payload<DoublePrecision>();
#if defined(DATA_TRACE)
      trace_proj_event_stop(Timer::timestamp(), traceEventTypes["gaDiagStdSeq"]);
#endif
      return;
    }

  if (p_a->first == 2 && p_v->first == 2)
    {
      DEBUG_PRINT("gaDiagStdSeq: [%d,%d]\n", theTwoSided().getProcRank().toInt(),
          g_a, g_v);

      GArray<2>& ga_a = *reinterpret_cast<GArray<2>*>(p_a->second);
      GArray<2>& ga_v = *reinterpret_cast<GArray<2>*>(p_v->second);

      Integer lo[2] =
        { 1, 1 };
      Integer hi[2] =
        { ga_a.procGrid_.cg_.dims_[0], ga_a.procGrid_.cg_.dims_[1] };

      assert(hi[0] == hi[1]);
      assert(ga_a.elsize_ == sizeof(double));

      DoublePrecision* k_a = new DoublePrecision[hi[0] * hi[1]];
      DoublePrecision* k_v = new DoublePrecision[hi[0] * hi[1]];
      DoublePrecision *k_fv1 = new DoublePrecision[hi[0]];
      DoublePrecision *k_fv2 = new DoublePrecision[hi[0]];

      gaGet(g_a, lo, hi, k_a, hi);

      Integer nm = hi[0], n = hi[0], matz = 1, ierr = 0;

      rs_(&nm, &n, k_a, evals, &matz, k_v, k_fv1, k_fv2, &ierr);

      gaPut(g_v, lo, hi, k_v, hi);

      delete[] k_a;
      delete[] k_v;
      delete[] k_fv1;
      delete[] k_fv2;

      amBarrier();
      MPI::Set_Payload<DoublePrecision>(*evals);
      MPI::Bcast(NULL, hi[0], MPI_DOUBLE, 0, MPI_COMM_WORLD);
      *evals = MPI::Get_Payload<DoublePrecision>();
      DEBUG_PRINT("gaDiagStdSeq: finished on %d\n", g_a);
    }
  else
    {
      std::cerr << "dimensionality not implemented" << std::endl;
      exit(1);
    }
#if defined(DATA_TRACE)
  trace_proj_event_stop(Timer::timestamp(), traceEventTypes["gaDiagStdSeq"]);
#endif
}

struct DGEMMOperator
{
  GArray<2>& gArray_;
  const Box<2> localFullBox_;
  const Box<2> localBox_;
  Integer g_a, g_b;
  bool transa;

  DGEMMOperator(GArray<2>& gArray, Box<2>& localFullBox, Box<2>& localBox,
      Integer g_a, Integer g_b, bool transa) :
      gArray_(gArray), localFullBox_(localFullBox.lo(), localFullBox.hi()), localBox_(
          localBox.lo(), localBox.hi()), g_a(g_a), g_b(g_b), transa(transa)
  {
  }

  void
  operator()(const ProcGrid<2>& pg, const Box<2>& localBox,
      const Box<2>& globalBox, const Box<2>& curLocalBox,
      const Box<2>& curGlobalBox)
  {
    ProcRank remoteRank;
    unsigned owner;
    Offset local_offset;
    Array<Offset, 2> local_strides;
    Offset global_offset;
    Array<Offset, 2> global_strides;
    Array<Offset, 2> count;

    computeBoxPairParams<2>(pg, globalBox, curGlobalBox, localFullBox_,
        localBox, curLocalBox, owner, local_offset, local_strides,
        global_offset, global_strides, count);

    remoteRank = owner;

    if (remoteRank.toInt() != theTwoSided().getProcRank().toInt())
      {
        return;
      }

    RmaPtr d(gArray_.allocId(), global_offset * gArray_.elsize());
    double* ptr = reinterpret_cast<double*>(theRma().lookupPointer(d));

    // Assume that beta is 0.0
    for (unsigned i = 0; i < ichunk * ichunk; i++)
      {
        ptr[i] = 0.0;
      }

    // Assume square
    for (unsigned i = 0; i < globalBox.hi()[0]; i += ichunk)
      {
        //printf("%d: traversing blocks ([%d,%d],[%d,%d]) and ([%d,%d],[%d,%d])\n",
        //theTwoSided().getProcRank().toInt(),
        //curGlobalBox.lo()[0], i,
        //curGlobalBox.lo()[0] + ichunk, i + ichunk,
        //i, curGlobalBox.lo()[1],
        //i + ichunk, curGlobalBox.lo()[1] + ichunk);

        double* buf1 = new double[ichunk * ichunk];
        double* buf2 = new double[ichunk * ichunk];

        int lo1[2];

        if (transa)
          {
            lo1[0] = curGlobalBox.lo()[0];
            lo1[1] = i;
          }
        else
          {
            lo1[0] = i;
            lo1[1] = curGlobalBox.lo()[0];
          }

        int hi1[2] =
          { lo1[0] + ichunk - 1, lo1[1] + ichunk - 1 };
        int lo2[2] =
          { curGlobalBox.lo()[1], i };
        int hi2[2] =
          { lo2[0] + ichunk - 1, lo2[1] + ichunk - 1 };
        int ld = ichunk;

        gaGetI(g_a, lo1, hi1, buf1, &ld);
        gaGetI(g_b, lo2, hi2, buf2, &ld);

        for (unsigned j = 0; j < ichunk; j++)
          {
            for (unsigned k = 0; k < ichunk; k++)
              {
                double sum = 0.0;
                for (unsigned m = 0; m < ichunk; m++)
                  {
                    // @todo This branching is probably terrible for performance, should be hoisted
                    sum +=
                        (transa ? buf1[m + j * ichunk] : buf1[j + m * ichunk])
                            * buf2[m + k * ichunk];
                  }
                ptr[j + k * ichunk] += sum;
              }
          }

        delete[] buf1;
        delete[] buf2;
      }
  }
};

void
gaDgemm(char* transa, char* transb, Integer m, Integer n, Integer k,
    DoublePrecision* alpha, Integer g_a, Integer g_b, DoublePrecision* beta,
    Integer g_c, int talen, int tblen)
{
#if defined(DATA_TRACE)
  trace_proj_event_start(Timer::timestamp(), traceEventTypes["gaDgemm"]);
#endif

  amBarrier();

  assert(sizeof(DoublePrecision) == sizeof(double));

  // Assume same data type
  std::pair<int, void*>* p_a = theRegistry().lookup<std::pair<int, void*>*>(
      g_a);
  std::pair<int, void*>* p_b = theRegistry().lookup<std::pair<int, void*>*>(
      g_b);
  std::pair<int, void*>* p_c = theRegistry().lookup<std::pair<int, void*>*>(
      g_c);

  DEBUG_PRINT("gaDgemm: transa = %s, transb = %s, m = %d, n = %d, k = %d, "
      "g_a = %d, g_b = %d, g_c = %d, alpha = %f, beta = %f, "
      "talen = %d, tblend = %d\n",
      transa, transb, m, n, k, g_a, g_b, g_c, alpha[0], beta[0],
      talen, tblen);

  assert(*transb == 'n');
  assert(*alpha == 1.0);
  assert(*beta == 0.0);

  GArray<2>& ga_c = *reinterpret_cast<GArray<2>*>(p_c->second);
  Array<Offset, 2> lo, hi;
  lo[0] = lo[1] = 0;
  hi[0] = ga_c.procGrid_.cg_.dims_[0];
  hi[1] = ga_c.procGrid_.cg_.dims_[1];
  Box<2> gbox(lo, hi);

  DGEMMOperator op(ga_c, gbox, gbox, g_a, g_b, *transa == 't');
  ProcGridIter<DGEMMOperator, 2>::iterate(ga_c.procGrid_, gbox, gbox, op);

  amBarrier();

#if defined(DATA_TRACE)
  trace_proj_event_stop(Timer::timestamp(), traceEventTypes["gaDgemm"]);
#endif
}

void
gaCopy(Integer g_a, Integer g_b)
{
#if defined(DATA_TRACE)
  trace_proj_event_start(Timer::timestamp(), traceEventTypes["gaCopy"]);
#endif

  amBarrier();

  DEBUG_PRINT("gaCopy: from %d to %d\n", theTwoSided().getProcRank().toInt(),
      g_a, g_b);
  std::pair<int, void*>* p_a = theRegistry().lookup<std::pair<int, void*>*>(
      g_a);
  std::pair<int, void*>* p_b = theRegistry().lookup<std::pair<int, void*>*>(
      g_b);

  // Assume same distribution of arrays
  if (p_a->first == 2 && p_b->first == 2)
    {
      GArray<2>& ga_a = *reinterpret_cast<GArray<2>*>(p_a->second);
      GArray<2>& ga_b = *reinterpret_cast<GArray<2>*>(p_b->second);

      RmaPtr d1(ga_a.allocId(), 0);
      RmaPtr d2(ga_b.allocId(), 0);
      void* ptr1 = theRma().lookupPointer(d1);
      void* ptr2 = theRma().lookupPointer(d2);

      memcpy(ptr2, ptr1, ga_a.size() * ga_a.elsize());
    }
  else
    {
      std::cerr << "dimensionality not implemented" << std::endl;
      exit(1);
    }

  amBarrier();

#if defined(DATA_TRACE)
  trace_proj_event_stop(Timer::timestamp(), traceEventTypes["gaCopy"]);
#endif
}

void
gaAdd(DoublePrecision alpha, Integer g_a, DoublePrecision beta, Integer g_b,
    Integer g_c)
{
#if defined(DATA_TRACE)
  trace_proj_event_start(Timer::timestamp(), traceEventTypes["gaAdd"]);
#endif

  // Assume the same distribution, array of doubles
  std::pair<int, void*>* p_a = theRegistry().lookup<std::pair<int, void*>*>(
      g_a);
  std::pair<int, void*>* p_b = theRegistry().lookup<std::pair<int, void*>*>(
      g_b);
  std::pair<int, void*>* p_c = theRegistry().lookup<std::pair<int, void*>*>(
      g_c);

  DEBUG_PRINT("gaAdd: [%d,%d,%d], alpha = %f, beta = %f\n",
      g_a, g_b, g_c, alpha, beta);

  if (p_a->first == 2 && p_b->first == 2 && p_c->first == 2)
    {
      GArray<2>& ga_a = *reinterpret_cast<GArray<2>*>(p_a->second);
      GArray<2>& ga_b = *reinterpret_cast<GArray<2>*>(p_b->second);
      GArray<2>& ga_c = *reinterpret_cast<GArray<2>*>(p_c->second);

      RmaPtr rptr_a(ga_a.allocId(), 0);
      RmaPtr rptr_b(ga_b.allocId(), 0);
      RmaPtr rptr_c(ga_c.allocId(), 0);

      double* buf_a = reinterpret_cast<double*>(theRma().lookupPointer(rptr_a));
      double* buf_b = reinterpret_cast<double*>(theRma().lookupPointer(rptr_b));
      double* buf_c = reinterpret_cast<double*>(theRma().lookupPointer(rptr_c));

      for (unsigned i = 0; i < ga_a.size(); i++)
        {
          buf_c[i] = buf_a[i] * alpha + buf_b[i] * beta;
        }
    }
  else
    {
      std::cerr << "dimensionality not implemented" << std::endl;
      exit(1);
    }

  amBarrier();

#if defined(DATA_TRACE)
  trace_proj_event_stop(Timer::timestamp(), traceEventTypes["gaAdd"]);
#endif
}
