//#if HAVE_CONFIG_H
# include "sstmac/common/sstmac-config.h"
//#endif
#include "scf_defs.h"
#include "rpl_gac.h"
#include "ftypes.h"
#include "gacpp.h"

#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include <assert.h>
#include <list>
#include <map>
#include <set>

#include <mpi/sstmac_mpi.h>

#define ctwoel_               SSTMAC_F77_FUNC_(ctwoel,CTWOEL)
#define contract_matrices_    SSTMAC_F77_FUNC_(contract_matrices,CONTRACT_MATRICES)
#define g_                    SSTMAC_F77_FUNC_(g,G)

using namespace std;
using namespace scf_tascel;

#define maxatom  286
#define maxnbfn  (15*maxatom)
#define mxiter   30
#define maxnnbfn ((maxnbfn)*(maxnbfn+1)/2)
#define pi       3.141592653589793d0
#define tol      (0.5e-3)
#define tol2e    (1.0e-6)

#define ichunk 20

extern std::map<int, std::list<long int> > taskSample;
extern double peLoad;
extern int peLoadCount;

#include "tascel/grid/Box.h"

#ifdef __cplusplus
extern "C"
{
#endif
  extern int armci_me;
  void
  nga_get_(Integer *g_a, Integer *lo, Integer *hi, void *buf, Integer *ld);

  void g_(double *gg, Integer *i, Integer *j, Integer *k, Integer *l);
  double contract_matrices_(Integer *g_a, Integer *g_b);
  void
  new_nga_acc_(Integer *g_a, Integer *lo, Integer *hi, void *buf, Integer *ld,
      void *alpha);

  void
  nga_acc_(Integer *g_a, Integer *lo, Integer *hi, void *buf, Integer *ld,
      void *alpha);

  void ctwoel_(Integer *g_schwarz, Integer *g_dens, Integer *g_fock,
      double *schwmax, double *etwo, long long *pnbfn, long long *icut1,
      long long *icut2, long long *icut3, long long *icut4);

#ifdef __cplusplus
}
#endif

static void
cpptwoel(Integer g_schwarz, Integer g_dens, Integer g_fock, double schwmax,
    Integer nbfn, long long *icut1, long long *icut2, long long *icut3,
    long long *icut4);

void ctwoel_(Integer *g_schwarz, Integer *g_dens, Integer *g_fock,
    double *schwmax, double *etwo, long long *pnbfn, long long *icut1,
    long long *icut2, long long *icut3, long long *icut4)
{
  const Integer nbfn = (Integer) *pnbfn;
  long long ijcnt, klcnt, ijklcnt;

  ijcnt = *icut1;
  klcnt = *icut2;
  ijklcnt = *icut3;

  cpptwoel(*g_schwarz, *g_dens, *g_fock, *schwmax, nbfn, icut1, icut2, icut3,
      icut4);

  *etwo = 0.5 * contract_matrices_(g_fock, g_dens);

  ijcnt = *icut1 - ijcnt;
  klcnt = *icut2 - klcnt;
  ijklcnt = (*icut3) - ijklcnt;
  *icut4 = *icut3;
  if (*icut3 <= 0L)
    {
      /*     no integrals may be calculated if there is no work for  */
      /*     this node (ichunk too big), or, something is wrong */
      printf("no two-electron integrals computed by node %d\n", GA_Nodeid());
      fflush(stdout);
    }

  return;
}

/*************************************************************************
 *
 * the code involving task pool stuff follows
 *
 *************************************************************************/

#include "tascel/utc/UniformTaskCollSplitHybrid.h"
#include "tascel/utc/DataCollRegistry.h"
#include "tascel/utc/AccessModeRegistry.h"
#include <algorithm>

using namespace std;
using namespace tascel;

static int
next_4chunk(int g_counter, Integer nbfn, Integer lo[4], Integer hi[4],
    Integer *ilo, Integer *jlo, Integer *klo, Integer *llo)
{
  const long one = 1;
  const int zero = 0;
  long imax, itask;
  Integer itmp;
  long long total_ntsks;
  imax = nbfn / ichunk;
  if (nbfn - ichunk * imax > 0)
    {
      imax = imax + 1;
    }
  total_ntsks = imax * imax * imax * imax;

  itask = NGA_Read_inc(g_counter, (int *) &zero, one);
  /*   printf("%d: itask=%d total_ntsks=%lf\n", armci_me, (int)itask, (double)total_ntsks); */
  if (itask < 0)
    {
      printf("next_4chunk: itask negative: %d imax:%d nbfn:%d ichunk:%d\n",
          (int) itask, (int) imax, (int) nbfn, (int) ichunk);
      printf("probable GA integer precision problem if %lf > 2^31\n",
          pow(imax * 1.0, 4));
      fflush(stdout);
      assert(0);
    }
  if (itask < total_ntsks)
    {
      itmp = itask;
      *ilo = itmp % imax;
      itmp /= imax;
      *jlo = itmp % imax;
      itmp /= imax;
      *klo = itmp % imax;
      itmp /= imax;
      assert(itmp < imax);
      *llo = itmp;

      lo[0] = *ilo * ichunk;
      lo[1] = *jlo * ichunk;
      lo[2] = *klo * ichunk;
      lo[3] = *llo * ichunk;
      hi[0] = min((*ilo + 1) * ichunk, nbfn) - 1;
      hi[1] = min((*jlo + 1) * ichunk, nbfn) - 1;
      hi[2] = min((*klo + 1) * ichunk, nbfn) - 1;
      hi[3] = min((*llo + 1) * ichunk, nbfn) - 1;

      return 1;
    }
  else
    {
      return 0;
    }
}

static void
clean_chunk(double *chunk)
{
  int i;
  for (i = 0; i < ichunk * ichunk; i++)
    {
      chunk[i] = 0.0;
    }
}
class MemPoolManager
{
private:
  const size_t size;
  const int nbufs;
  std::list<void *> bufs;
  std::set<void *> managed_bufs; //FIXME: for testing
public:
  MemPoolManager(int size_, int nbufs_) :
      size(size_), nbufs(nbufs_)
  {
    for (int i = 0; i < nbufs; i++)
      {
        void *ptr = malloc(size);
        assert(ptr!=NULL);
        bufs.insert(bufs.begin(), ptr);
        managed_bufs.insert(ptr);
      }
  }
  ~MemPoolManager()
  {
    int i = 0;
    for (list<void*>::iterator itr = bufs.begin(); itr != bufs.end();)
      {
        assert(*itr != NULL);
        free(*itr);
        itr = bufs.erase(itr);
        i++;
      }
    assert(i==nbufs);
    //no memory leaks
  }
  bool
  hasBufs()
  {
    return bufs.begin() != bufs.end();
  }
  void *
  request()
  {
    assert(hasBufs());
    void *ret = bufs.front();
    bufs.pop_front();
    return ret;
  }
  void
  release(void *buf)
  {
    if (buf != NULL)
      {
        assert(managed_bufs.find(buf) != managed_bufs.end());
        bufs.push_front(buf);
      }
  }
};

typedef struct
{
  double buf[ichunk][ichunk];
} buf_t;

typedef struct
{
  long int taskid;
} task_dscr_t;

struct bufs_t
{
  double f_ij[ichunk][ichunk], d_kl[ichunk][ichunk];
  double f_ik[ichunk][ichunk], d_jl[ichunk][ichunk];
  double s_ij[ichunk][ichunk], s_kl[ichunk][ichunk];
};

typedef struct
{
  Integer nbfn;
  double schwmax;
  int cnt1, cnt2;
  long long *icut1, *icut2, *icut3, *icut4;
  Integer g_schwarz, g_dens, g_fock;
  MemPoolManager *bmgr;
} task_plo_t;

static void
compute_lo_hi(long itask, long nbfn, long lo[4], long hi[4], long *ilo,
    long *jlo, long *klo, long *llo)
{
  long imax;
  long itmp;
  long long total_ntsks;
  imax = nbfn / ichunk;
  if (nbfn - ichunk * imax > 0)
    {
      imax = imax + 1;
    }
  total_ntsks = imax * imax * imax * imax;

  if (itask < 0)
    {
      printf("next_4chunk: itask negative: %d imax:%d nbfn:%d ichunk:%d\n",
          (int) itask, (int) imax, (int) nbfn, (int) ichunk);
      printf("probable GA integer precision problem if %lf > 2^31\n",
          pow(imax * 1.0, 4));
      fflush(stdout);
      assert(0);
    }assert(itask < total_ntsks);

  itmp = itask;
  *ilo = itmp % imax;
  itmp /= imax;
  *jlo = itmp % imax;
  itmp /= imax;
  *klo = itmp % imax;
  itmp /= imax;
  assert(itmp < imax);
  *llo = itmp;

  lo[0] = *ilo * ichunk;
  lo[1] = *jlo * ichunk;
  lo[2] = *klo * ichunk;
  lo[3] = *llo * ichunk;
  hi[0] = min((*ilo + 1) * ichunk, nbfn) - 1;
  hi[1] = min((*jlo + 1) * ichunk, nbfn) - 1;
  hi[2] = min((*klo + 1) * ichunk, nbfn) - 1;
  hi[3] = min((*llo + 1) * ichunk, nbfn) - 1;
}

static int
compute_owner(long itask, const vector<DataColl*> &colls, int arrid, long nbfn)
{
  const int size = 2 * sizeof(int);
  char idx[size];
  long int ilo, jlo, klo, llo, lo[4], hi[4], tmp;
  int *iidx = (int *) idx;

  compute_lo_hi(itask, nbfn, lo, hi, &ilo, &jlo, &klo, &llo);

  switch (arrid)
    {
  case 0:
    iidx[0] = lo[1];
    iidx[1] = lo[0];
    break;
  case 1:
    iidx[0] = lo[3] / ichunk;
    iidx[1] = lo[2] / ichunk;
    break;
  case 2:
    iidx[0] = lo[3] / ichunk;
    iidx[1] = lo[2] / ichunk;
    break;
  case 3:
    iidx[0] = lo[3] / ichunk;
    iidx[1] = lo[1] / ichunk;
    break;
  case 4:
    iidx[0] = lo[1] / ichunk;
    iidx[1] = lo[0] / ichunk;
    break;
  case 5:
    iidx[0] = lo[2] / ichunk;
    iidx[1] = lo[0] / ichunk;
    break;
  default:
    assert(0);
    }
}

static void
compute_index(void *dscr, int dscr_len, void *pldata, int pldata_len, int arrid,
    void *idx, int idxlen)
{
  assert(dscr_len == sizeof(task_dscr_t));
  assert(arrid >=0 && arrid<6);
  assert(idxlen == 2*sizeof(int));
  const long itask = ((task_dscr_t*) dscr)->taskid;
  int *iidx = (int *) idx;
  task_plo_t *ptplo = (task_plo_t*) pldata;
  assert(pldata_len == sizeof(task_plo_t));
  const long int nbfn = ptplo->nbfn;
  long int ilo, jlo, klo, llo, lo[4], hi[4], tmp;

  compute_lo_hi(itask, nbfn, lo, hi, &ilo, &jlo, &klo, &llo);

  switch (arrid)
    {
  case 0:
    iidx[0] = lo[1] / ichunk;
    iidx[1] = lo[0] / ichunk;
    break;
  case 1:
    iidx[0] = lo[3] / ichunk;
    iidx[1] = lo[2] / ichunk;
    break;
  case 2:
    iidx[0] = lo[3] / ichunk;
    iidx[1] = lo[2] / ichunk;
    break;
  case 3:
    iidx[0] = lo[3] / ichunk;
    iidx[1] = lo[1] / ichunk;
    break;
  case 4:
    iidx[0] = lo[1] / ichunk;
    iidx[1] = lo[0] / ichunk;
    break;
  case 5:
    iidx[0] = lo[2] / ichunk;
    iidx[1] = lo[0] / ichunk;
    break;
  default:
    assert(0);
    }
}

static void
twoel_task(UniformTaskCollection *utc, void *bigd, int bigd_len, void *pldata,
    int pldata_len, vector<void *> data_bufs);

#define TRACE_LIMIT 10
static int numTraced = 0;

enum
{
  SCHWARZ, SCHWARZ_2, DENS, DENS_2, FOCK, FOCK_2
};
static int dataColls[6];
static UniformTaskCollSplitHybrid* utc = 0;

#include <list>
#include <vector>
#include <algorithm>

static AmHandle rhan, ldhan, rpehan;
static AmHandle rsamplePE, ldhanSample, rdistSamplePE, ldhanDist, treeHan,
    orderHan, downTree;

Dispatcher<NullMutex> dispatcher;

inline void
dispatchUntilEmpty(bool executeExternal = true)
{
  // All the requests enqueued need to complete before the computation
  // proceeds
  while (!dispatcher.empty())
    {
      Codelet* codelet;
      if ((codelet = dispatcher.progress()) != NULL)
        {
          codelet->execute();
          // Assume that codelet was an AmRequest and needs to be freed
          delete reinterpret_cast<AmRequest*>(codelet);
        }
      if (executeExternal)
        {
          AmListenObjCodelet<NullMutex>* lcodelet;
          if ((lcodelet = theAm().amListeners[0]->progress()) != NULL)
            {
              lcodelet->execute();
            }
        }
    }
}

struct LBRecord
{
  long int rec;
  double time;
};

struct PE
{
  vector<LBRecord> recs;
  int pe;
  double load;

  PE() :
      pe(0), load(0.0)
  {
  }
};

struct PESample
{
  std::map<int, std::list<long int> > recs;
  int pe;
  double load;
  int count;

  PESample() :
      pe(0), load(0.0)
  {
  }

  PESample(int pe) :
      pe(pe), load(0.0)
  {
  }
};

struct PEDist
{
  vector<long int> recs;
  int pe;
  double load;

  PEDist() :
      pe(0), load(0.0)
  {
  }
};

static std::vector<LBRecord> lbData;
static int peRec = 0;

void
recvData(const AmContext* const context)
{
  peRec++;
  LBRecord* recs = reinterpret_cast<LBRecord*>(context->arg->data);
  for (unsigned i = 0; i < context->arg->dlen.toInt() / sizeof(LBRecord); i++)
    {
      lbData.push_back(*(recs + i));
    }
}

static vector<PE> pes;
static vector<PESample> pesm;
static vector<PEDist> ped;
static double totalLoad = 0.0;

struct PEAmArg : public AmArg
{
  int pe;
  double totalLoad;
};

void
recvPEData(const AmContext* const context)
{
  peRec++;
  PEAmArg& peArg = *reinterpret_cast<PEAmArg*>(context->arg);
  LBRecord* recs = reinterpret_cast<LBRecord*>(context->arg->data);

  pes[peArg.pe].load = 0.0;
  for (unsigned i = 0; i < context->arg->dlen.toInt() / sizeof(LBRecord); i++)
    {
      pes[peArg.pe].load += recs[i].time;
      pes[peArg.pe].recs.push_back(recs[i]);
    }

  totalLoad += pes[peArg.pe].load;
}

static bool loadDataRecv = false;

void
newLoad(const AmContext* const context)
{
  loadDataRecv = true;

  // Reset temp queue inside UTC
  utc->nonNullLength = 0;

  LBRecord* recs = reinterpret_cast<LBRecord*>(context->arg->data);
  for (unsigned i = 0; i < context->arg->dlen.toInt() / sizeof(LBRecord); i++)
    {
      task_dscr_t dscr;
      dscr.taskid = recs[i].rec;
      utc->isNotReadOnly(&dscr, sizeof(dscr));
    }
}

template<typename T>
  struct CompPE
  {
    bool
    operator()(const T& pe1, const T& pe2) const
    {
      return pe1.load > pe2.load;
    }
  };

template<typename T>
  struct CompPEA
  {
    bool
    operator()(const T& pe1, const T& pe2) const
    {
      return pe1.load / pe1.count > pe2.load / pe2.count;
    }
  };

struct CompLBRecord
{
  bool
  operator()(const LBRecord& pe1, const LBRecord& pe2) const
  {
    return pe1.time < pe2.time;
  }
};

struct CompLBRecordDist
{
  double val;

  CompLBRecordDist(double val) :
      val(val)
  {
  }

  bool
  operator()(const LBRecord& pe1, const LBRecord& pe2) const
  {
    return fabs(pe1.time - val) < fabs(pe2.time - val);
  }
};

struct CompLBRecordMin
{
  bool
  operator()(const LBRecord& pe1, const LBRecord& pe2) const
  {
    return pe1.time > pe2.time;
  }
};

static std::map<int, std::list<long int> > combinedSample;

void
recvPESample(const AmContext* const context)
{
  peRec++;
  PEAmArg& peArg = *reinterpret_cast<PEAmArg*>(context->arg);
  long int* recs = reinterpret_cast<long int*>(context->arg->data);

  pesm[peArg.pe].load = peArg.totalLoad;

  int currentBin = -1;

  for (unsigned i = 0; i < context->arg->dlen.toInt() / sizeof(long int); i++)
    {
      if (recs[i] == -1)
        {
          currentBin = recs[i + 1];
          i++;
        }
      else
        {
          pesm[peArg.pe].recs[currentBin].push_back(recs[i]);
        }
    }

  totalLoad += pesm[peArg.pe].load;
}

void
newLoadSample(const AmContext* const context)
{
  loadDataRecv = true;

  // Reset temp queue inside UTC
  utc->nonNullLength = 0;

  long int* recs = reinterpret_cast<long int*>(context->arg->data);
  for (unsigned i = 0; i < context->arg->dlen.toInt() / sizeof(long int); i++)
    {
      task_dscr_t dscr;
      dscr.taskid = recs[i];
      utc->isNotReadOnly(&dscr, sizeof(dscr));
    }
}

void
recvPEDist(const AmContext* const context)
{
  peRec++;
  PEAmArg& peArg = *reinterpret_cast<PEAmArg*>(context->arg);
  long int* recs = reinterpret_cast<long int*>(context->arg->data);

  ped[peArg.pe].load = peArg.totalLoad;

  int currentBin = -1;

  for (unsigned i = 0; i < context->arg->dlen.toInt() / sizeof(long int); i++)
    {
      if (recs[i] == -1)
        {
          currentBin = recs[i + 1];
          i++;
        }
      else
        {
          combinedSample[currentBin].push_back(recs[i]);
        }
    }
}

void
newLoadDist(const AmContext* const context)
{
  loadDataRecv = true;

  // Reset temp queue inside UTC
  utc->nonNullLength = 0;

  long int* recs = reinterpret_cast<long int*>(context->arg->data);
  for (unsigned i = 0; i < context->arg->dlen.toInt() / sizeof(long int); i++)
    {
      taskSample[-2].push_back(recs[i]);
    }

  for (std::map<int, std::list<long int> >::iterator iter = taskSample.begin();
      iter != taskSample.end(); ++iter)
    {
      for (std::list<long int>::iterator iter2 = iter->second.begin();
          iter2 != iter->second.end(); ++iter2)
        {
          task_dscr_t dscr;
          dscr.taskid = *iter2;
          utc->isNotReadOnly(&dscr, sizeof(dscr));
        }
    }
}

void
refineLBSample()
{
  //double sampleTime = Timer::timestamp();
  printf("refineLBSample()\n");
  printf("number of bins = %d\n", taskSample.size());

  if (theTwoSided().getProcRank() != 0)
    {
      std::vector<long int> sampleSerial;

      for (std::map<int, std::list<long int> >::iterator iter =
          taskSample.begin(); iter != taskSample.end(); ++iter)
        {
          sampleSerial.push_back(-1);
          sampleSerial.push_back(iter->first);

          for (std::list<long int>::iterator iter2 = iter->second.begin();
              iter2 != iter->second.end(); ++iter2)
            {
              sampleSerial.push_back(*iter2);
            }
        }

      PEAmArg arg;
      arg.data = &sampleSerial[0];
      arg.dlen = sizeof(long int) * sampleSerial.size();
      arg.pe = theTwoSided().getProcRank().toInt();
      arg.totalLoad = peLoad;

      AmRequest* req = AmRequest::construct();
      theAm().amPut(0, arg, rsamplePE, req);
      dispatcher.registerCodelet(req);
      dispatchUntilEmpty();

      while (!loadDataRecv)
        {
          AmListenObjCodelet<NullMutex>* lcodelet;
          if ((lcodelet = theAm().amListeners[0]->progress()) != NULL)
            {
              lcodelet->execute();
            }
        }
    }
  else
    {
      pesm[0].recs = taskSample;
      pesm[0].load = peLoad;

      totalLoad += pesm[0].load;

      while (peRec != theTwoSided().numProcRanks().toInt() - 1)
        {
          AmListenObjCodelet<NullMutex>* lcodelet;
          if ((lcodelet = theAm().amListeners[0]->progress()) != NULL)
            {
              lcodelet->execute();
            }
        }

      double avgLoad = totalLoad / theTwoSided().numProcRanks().toInt();

      //printf("avgLoad = %f\n", avgLoad);

      for (unsigned i = 0; i < theTwoSided().numProcRanks().toInt(); i++)
        {
          PESample& pe = pesm[i];
          //printf("pe %d: load = %f, 1.003 * avgLoad = %f\n", i, pe.load, 1.003 * avgLoad);

          std::map<int, std::list<long int> >::iterator item = pe.recs.begin();
          while (pe.load > 1.003 * avgLoad && item != pe.recs.end())
            {
              assert(item->second.size() != 0);

              int bin = item->first;
              combinedSample[item->first].push_back(item->second.back());
              /*printf("adding item to pool: size %d, id = %d\n",
               item->first, item->second.back());*/
              item->second.pop_back();
              if (item->second.size() == 0)
                {
                  item++;
                }

              pe.load -= bin;
            }
        }

      make_heap(pesm.begin(), pesm.end(), CompPE<PESample>());

      /*if (combinedSample.rbegin() != combinedSample.rend())
       printf("max elem = %d milliseconds\n", combinedSample.rbegin()->first);*/

      //printf("min load = %f\n", pesm.front().load);
      // Iterate from largest to smallest item
      std::map<int, std::list<long int> >::reverse_iterator cIter =
          combinedSample.rbegin();

      while (cIter != combinedSample.rend())
        {
          assert(cIter->second.size() != 0);

          PESample minPE = pesm.front();
          //printf("minPE = %d, load = %f\n", minPE.pe, minPE.load);
          pop_heap(pesm.begin(), pesm.end(), CompPE<PESample>());
          pesm.pop_back();

          minPE.recs[cIter->first].push_back(cIter->second.back());
          minPE.load += cIter->first;
          cIter->second.pop_back();

          if (cIter->second.size() == 0)
            {
              cIter++;
            }

          pesm.push_back(minPE);
          push_heap(pesm.begin(), pesm.end(), CompPE<PESample>());
        }

      for (unsigned i = 0; i < theTwoSided().numProcRanks().toInt(); i++)
        {
          //printf("%d load = %f\n", pesm[i].pe, pesm[i].load);

          vector<long int> tasks;
          for (std::map<int, std::list<long int> >::iterator iter =
              pesm[i].recs.begin(); iter != pesm[i].recs.end(); ++iter)
            {
              for (std::list<long int>::iterator iter2 = iter->second.begin();
                  iter2 != iter->second.end(); ++iter2)
                {
                  tasks.push_back(*iter2);
                }
            }

          AmArg arg;
          arg.data = &tasks[0];
          arg.dlen = sizeof(long int) * tasks.size();
          AmRequest* req = AmRequest::construct();
          theAm().amPut(pesm[i].pe, arg, ldhanSample, req);
          dispatcher.registerCodelet(req);
          dispatchUntilEmpty();
        }

      while (!loadDataRecv)
        {
          AmListenObjCodelet<NullMutex>* lcodelet;
          if ((lcodelet = theAm().amListeners[0]->progress()) != NULL)
            {
              lcodelet->execute();
            }
        }
    }

  //printf("%d: finished, time = %f\n", theTwoSided().getProcRank(),
  // Timer::timestamp() - sampleTime);
}

void
refineLBDistribSample()
{
  double sampleTime = Timer::timestamp();
  //printf("refineLBDistribSample()\n");

  double sumLoad = 0.0;

  MPI::Set_Payload<double>(peLoad);
  MPI::Allreduce(NULL, NULL, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  sumLoad = MPI::Get_Payload<double>();
  double avgLoad = sumLoad / theTwoSided().numProcRanks().toInt();

  if (theTwoSided().getProcRank() == 0)
    printf("avgLoad = %f\n", avgLoad);

  double peNewLoad = peLoad;
  vector<long int> localTaskPool;
  std::map<int, std::list<long int> >::iterator item = taskSample.begin();

  if (item != taskSample.end())
    {
      localTaskPool.push_back(-1);
      localTaskPool.push_back(item->first);
    }

  /*printf("pe %d: load = %f, 1.003 * avgLoad = %f\n", theTwoSided().getProcRank(),
   peLoad, 1.003 * avgLoad);*/

  while (peNewLoad > 1.003 * avgLoad && item != taskSample.end())
    {
      assert(item->second.size() != 0);

      if (theTwoSided().getProcRank() != 0)
        {
          localTaskPool.push_back(item->second.back());
        }
      else
        {
          combinedSample[item->first].push_back(item->second.back());
        }
      /*printf("adding item to pool: size %d, id = %d\n",
       item->first, item->second.back());*/
      item->second.pop_back();
      peNewLoad -= item->first;

      if (item->second.size() == 0)
        {
          item++;
          if (item != taskSample.end())
            {
              localTaskPool.push_back(-1);
              localTaskPool.push_back(item->first);
            }
        }
    }

  if (theTwoSided().getProcRank().toInt() != 0)
    {
      PEAmArg arg;
      arg.data = &localTaskPool[0];
      arg.dlen = sizeof(long int) * localTaskPool.size();
      arg.pe = theTwoSided().getProcRank().toInt();
      arg.totalLoad = peNewLoad;
      AmRequest* req = AmRequest::construct();
      theAm().amPut(0, arg, rdistSamplePE, req);
      dispatcher.registerCodelet(req);
      dispatchUntilEmpty();
    }
  else
    {
      ped[0].load = peNewLoad;
    }

  if (theTwoSided().getProcRank() == 0)
    {
      while (peRec != theTwoSided().numProcRanks().toInt() - 1)
        {
          AmListenObjCodelet<NullMutex>* lcodelet;
          if ((lcodelet = theAm().amListeners[0]->progress()) != NULL)
            {
              lcodelet->execute();
            }
        }

      make_heap(ped.begin(), ped.end(), CompPE<PEDist>());

      /*if (combinedSample.rbegin() != combinedSample.rend())
       printf("max elem = %d millisecond bin\n", combinedSample.rbegin()->first);
       printf("min load = %f\n", ped.front().load);*/

      int taskCount = 0;

      // Iterate from largest to smallest item
      std::map<int, std::list<long int> >::reverse_iterator cIter =
          combinedSample.rbegin();

      while (cIter != combinedSample.rend())
        {
          assert(cIter->second.size() != 0);

          PEDist minPE = ped.front();
          //printf("minPE = %d, load = %f\n", minPE.pe, minPE.load);
          pop_heap(ped.begin(), ped.end(), CompPE<PEDist>());
          ped.pop_back();

          minPE.recs.push_back(cIter->second.back());
          minPE.load += cIter->first;
          cIter->second.pop_back();
          taskCount++;

          if (cIter->second.size() == 0)
            {
              cIter++;
            }

          ped.push_back(minPE);
          push_heap(ped.begin(), ped.end(), CompPE<PEDist>());
        }

      printf("redistributed %d tasks\n", taskCount);

      for (unsigned i = 0; i < theTwoSided().numProcRanks().toInt(); i++)
        {
          //printf("%d load = %f\n", ped[i].pe, ped[i].load);

          AmArg arg;
          arg.data = &ped[i].recs[0];
          arg.dlen = sizeof(long int) * ped[i].recs.size();
          AmRequest* req = AmRequest::construct();
          theAm().amPut(ped[i].pe, arg, ldhanDist, req);
          dispatcher.registerCodelet(req);
          dispatchUntilEmpty();
        }
    }

  while (!loadDataRecv)
    {
      AmListenObjCodelet<NullMutex>* lcodelet;
      if ((lcodelet = theAm().amListeners[0]->progress()) != NULL)
        {
          lcodelet->execute();
        }
    }

  printf("%d: finished, time = %f\n", theTwoSided().getProcRank().toInt(),
      Timer::timestamp() - sampleTime);
}

#if 0

void refineLB()
  {
    //printf("refineLB()\n");

    if (theTwoSided().getProcRank() != 0)
      {
        //printf("taskDuration.size() = %d\n", taskDuration.size());
        LBRecord* recs = (LBRecord*)alloca(sizeof(LBRecord) * taskDuration.size());

        unsigned i = 0;
        for (std::map<long int, double>::iterator iter = taskDuration.begin();
            iter != taskDuration.end(); ++iter)
          {
            recs[i].rec = iter->first;
            recs[i].time = iter->second;
            i++;
          }

        PEAmArg arg;
        arg.data = recs;
        arg.dlen = sizeof(LBRecord) * taskDuration.size();
        arg.pe = theTwoSided().getProcRank();

        AmRequest req = theAm().amPut(0, arg, rpehan);
        req.wait();

        while (!loadDataRecv)
          {
            theAm().dispatch();
          }
      }
    else
      {

        for (std::map<long int, double>::iterator iter = taskDuration.begin();
            iter != taskDuration.end(); ++iter)
          {
            LBRecord lrec;
            lrec.rec = iter->first;
            lrec.time = iter->second;
            pes[theTwoSided().getProcRank()].recs.push_back(lrec);
            pes[theTwoSided().getProcRank()].load += lrec.time;
          }

        totalLoad += pes[theTwoSided().getProcRank()].load;

        while (peRec != theTwoSided().numProcRanks().toInt() - 1)
          {
            theAm().dispatch();
          }

        double avgLoad = totalLoad / theTwoSided().numProcRanks().toInt();

        //printf("avgLoad = %f\n", avgLoad);

        for (unsigned i = 0; i < theTwoSided().numProcRanks().toInt(); i++)
          {
            PE& pe = pes[i];
            //printf("pe %d: load = %f, 1.003 * avgLoad = %f\n", i, pe.load, 1.003 * avgLoad);
            while (pe.load > 1.003 * avgLoad)
              {
                make_heap(pe.recs.begin(), pe.recs.end(), CompLBRecordMin());

                lbData.push_back(pe.recs.front());

                //printf("adding item to pool: %f\n", pe.recs.front().time);

                pe.load -= pe.recs.front().time;

                pop_heap(pe.recs.begin(), pe.recs.end(), CompLBRecordMin());
                pe.recs.pop_back();
              }
          }

        make_heap(pes.begin(), pes.end(), CompPE<PE>());

        /*printf("max elem = %f\n", lbData.front().time);
         printf("min load = %f\n", pes.front().load);*/

        int lbSize = lbData.size();
        for (unsigned i = 0; i < lbSize; i++)
          {
            LBRecord maxRec = lbData.front();
            pop_heap(lbData.begin(), lbData.end(), CompLBRecord());
            lbData.pop_back();

            PE minPE = pes.front();
            //printf("minPE = %d, load = %f\n", minPE.pe, minPE.load);
            pop_heap(pes.begin(), pes.end(), CompPE<PE>());
            pes.pop_back();

            minPE.recs.push_back(maxRec);
            minPE.load += maxRec.time;

            pes.push_back(minPE);
            push_heap(pes.begin(), pes.end(), CompPE<PE>());
          }

        for (unsigned i = 0; i < theTwoSided().numProcRanks().toInt(); i++)
          {
            //printf("%d load = %f\n", pes[i].pe, pes[i].load);
            AmArg arg;
            arg.data = &pes[i].recs[0];
            arg.dlen = sizeof(LBRecord) * pes[i].recs.size();
            AmRequest req = theAm().amPut(pes[i].pe, arg, ldhan);
            req.wait();
          }
      }
  }

#endif

struct LBTreeMsg : public AmArg
{
  double amt;
  int pe;
  int count;

  LBTreeMsg(double amt, int pe) :
      amt(amt), pe(pe), count(0)
  {
  }
};

struct LBFinalMsg : public AmArg
{
  bool final;
};

static int numChildren = -1;
static double aggregateLoad = 0.0;
static vector<PESample> childrenLoad;
static map<int, list<long int> > givenTasks;
static double totalChildLoad = 0.0;
static bool hasOrders = false, reachedLeaf = false;
static vector<int> children;
static vector<int> finalChildren;
static double avgLoad = 0.0;
static int upCount = 0;

void
treeFun(const AmContext* const context)
{
  //trace_proj_event_start(Timer::timestamp(), traceEventTypes["LB_prep_up"]);
  //printf("%d: treefun numChildren = %d\n",
  //theTwoSided().getProcRank().toInt(), numChildren);
  //assert(numChildren != -1);
  numChildren--;

  LBTreeMsg& lbmsg = *reinterpret_cast<LBTreeMsg*>(context->arg);
  long int* recs = reinterpret_cast<long int*>(context->arg->data);

  if (lbmsg.amt == -1)
    {
      for (vector<PESample>::iterator iter = childrenLoad.begin();
          iter != childrenLoad.end(); ++iter)
        {
          if (iter->pe == lbmsg.pe)
            {
              //assert(iter->recs.size() == 0);
              childrenLoad.erase(iter);
              break;
            }
        }
      //trace_proj_event_stop(Timer::timestamp(), traceEventTypes["LB_prep_up"]);
      return;
    }

  upCount += lbmsg.count;

  for (int i = 0; i < childrenLoad.size(); i++)
    {
      if (childrenLoad[i].pe == lbmsg.pe)
        {
          childrenLoad[i].load = lbmsg.amt;
          childrenLoad[i].count = lbmsg.count;
        }
    }

  totalChildLoad += lbmsg.amt;

  int currentBin = -1, count = 0;

  //printf("number of bytes = %d\n", context->arg->dlen.toInt());

  for (unsigned i = 0; i < context->arg->dlen.toInt() / sizeof(long int); i++)
    {
      if (recs[i] == -1)
        {
          currentBin = recs[i + 1];
          i++;
        }
      else
        {
          givenTasks[currentBin].push_back(recs[i]);
          count++;
        }
    }

  //printf("%d: recv from child %d: %d tasks\n", theTwoSided().getProcRank().toInt(),
  //lbmsg.pe, count);
  //trace_proj_event_stop(Timer::timestamp(), traceEventTypes["LB_prep_up"]);
}

void
orderFun(const AmContext* const context)
{
}

void
downTreeFun(const AmContext* const context)
{
  //trace_proj_event_start(Timer::timestamp(), traceEventTypes["LB_down_tree"]);

  LBFinalMsg& msg = *reinterpret_cast<LBFinalMsg*>(context->arg);
  long int* recs = reinterpret_cast<long int*>(context->arg->data);

  int currentBin = -1, count = 0;
  double loadAdded = 0.0;

  for (unsigned i = 0; i < context->arg->dlen.toInt() / sizeof(long int); i++)
    {
      if (recs[i] == -1)
        {
          currentBin = recs[i + 1];
          i++;
        }
      else
        {
          count++;
          //printf("adding task %ld into bin = %d\n", recs[i], currentBin);
          givenTasks[currentBin].push_back(recs[i]);
          loadAdded += currentBin;
        }
    }

  if (totalChildLoad == 0.0 && !msg.final)
    {
      //trace_proj_event_stop(Timer::timestamp(), traceEventTypes["LB_down_tree"]);
      return;
    }

  //printf("%d: downTree, number of tasks = %d, final = %s\n",
  //theTwoSided().getProcRank().toInt(), count, msg.final ? "true" : "false");

  if (!msg.final)
    {
      make_heap(childrenLoad.begin(), childrenLoad.end(), CompPEA<PESample>());

      // Iterate from largest to smallest item
      std::map<int, std::list<long int> >::reverse_iterator cIter =
          givenTasks.rbegin();

      while (cIter != givenTasks.rend())
        {
          while (cIter != givenTasks.rend() && cIter->second.size() == 0)
            {
              cIter++;
            }

          if (cIter == givenTasks.rend())
            {
              break;
            }

          PESample minPE = childrenLoad.front();
          //printf("YYY minPE = %d, load = %f\n", minPE.pe, minPE.load);
          pop_heap(childrenLoad.begin(), childrenLoad.end(),
              CompPEA<PESample>());
          childrenLoad.pop_back();

          minPE.recs[cIter->first].push_back(cIter->second.back());
          minPE.load += cIter->first;
          totalChildLoad += cIter->first;
          cIter->second.pop_back();

          if (cIter->second.size() == 0)
            {
              cIter++;
            }

          childrenLoad.push_back(minPE);
          push_heap(childrenLoad.begin(), childrenLoad.end(),
              CompPEA<PESample>());
        }

      for (unsigned i = 0; i < childrenLoad.size(); i++)
        {
          //printf("\t %d: sending to child %d load = %f, final = %s\n", theTwoSided().getProcRank(),
          //childrenLoad[i].pe, childrenLoad[i].load, children.size() == 0 ? "true" : "false");

          if (childrenLoad[i].pe == theTwoSided().getProcRank().toInt())
            {
              for (std::map<int, std::list<long int> >::iterator iter =
                  childrenLoad[i].recs.begin();
                  iter != childrenLoad[i].recs.end(); ++iter)
                {
                  for (std::list<long int>::iterator iter2 =
                      iter->second.begin(); iter2 != iter->second.end();
                      ++iter2)
                    {
                      //printf("adding task, id = %d\n", *iter2);
                      taskSample[iter->first].push_back(*iter2);
                      peLoad += iter->first;
                    }
                }
              childrenLoad[i].recs.clear();
              reachedLeaf = true;
            }
          else
            {
              vector<long int> tasks;
              for (std::map<int, std::list<long int> >::iterator iter =
                  childrenLoad[i].recs.begin();
                  iter != childrenLoad[i].recs.end(); ++iter)
                {
                  if (iter->second.size() != 0)
                    {
                      tasks.push_back(-1);
                      tasks.push_back(iter->first);
                    }
                  for (std::list<long int>::iterator iter2 =
                      iter->second.begin(); iter2 != iter->second.end();
                      ++iter2)
                    {
                      //printf("\t\t%d: sending child task %ld\n", theTwoSided().getProcRank(), *iter2);
                      tasks.push_back(*iter2);
                    }
                }

              childrenLoad[i].recs.clear();

              LBFinalMsg arg;
              arg.data = &tasks[0];
              arg.dlen = sizeof(long int) * tasks.size();

              arg.final = children.size() == 0;

              if (arg.final)
                {
                  //printf("final send from %d to %d\n", theTwoSided().getProcRank(), childrenLoad[i].pe);
                }

              AmRequest* req = AmRequest::construct();
              theAm().amPut(childrenLoad[i].pe, arg, downTree, req);
              assert(childrenLoad[i].pe != theTwoSided().getProcRank().toInt());
              dispatcher.registerCodelet(req);
              dispatchUntilEmpty(false);
            }
        }
    }
  else
    {
      reachedLeaf = true;

      for (std::map<int, std::list<long int> >::iterator iter =
          givenTasks.begin(); iter != givenTasks.end(); ++iter)
        {
          for (std::list<long int>::iterator iter2 = iter->second.begin();
              iter2 != iter->second.end(); ++iter2)
            {
              //task_dscr_t dscr;
              //dscr.taskid = *iter2;
              //utc->isNotReadOnly(&dscr, sizeof(dscr));
              taskSample[iter->first].push_back(*iter2);
              peLoad += iter->first;
              //printf("%d: bin = %d, adding task = %ld\n", theTwoSided().getProcRank(), iter->first, *iter2);
            }
        }

      givenTasks.clear();
    }
  //trace_proj_event_stop(Timer::timestamp(), traceEventTypes["LB_down_tree"]);
}

void
refineLBHierSample()
{
  //printf("refineLBHierSample()\n");

  numChildren = -1;
  aggregateLoad = 0.0;
  childrenLoad.clear();
  givenTasks.clear();
  totalChildLoad = 0.0;
  hasOrders = false;
  reachedLeaf = false;
  children.clear();
  finalChildren.clear();
  avgLoad = 0.0;
  upCount = 0;

  //trace_proj_event_start(Timer::timestamp(), traceEventTypes["LB_avg"]);
  double sumLoad = 0.0;

  MPI::Set_Payload<double>(peLoad);
  MPI::Allreduce(NULL, NULL, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  sumLoad = MPI::Get_Payload<double>();
  avgLoad = sumLoad / theTwoSided().numProcRanks().toInt();
  //trace_proj_event_stop(Timer::timestamp(), traceEventTypes["LB_avg"]);

  //trace_proj_event_start(Timer::timestamp(), traceEventTypes["LB_build_tree"]);

  if (theTwoSided().getProcRank() == 0)
    printf("avgLoad = %f\n", avgLoad);

  const int Nary = 3;
  const int rank = theTwoSided().getProcRank().toInt();

  for (unsigned i = 0; i < Nary; i++)
    {
      int child = rank * Nary + i + 1;
      if (child < theTwoSided().numProcRanks().toInt())
        {
          children.push_back(child);
          childrenLoad.push_back(PESample(child));
          //printf("\t%d: child = %d\n", theTwoSided().getProcRank().toInt(), child);
        }
    }

  if (children.size() == 0)
    {
      for (unsigned i = 0; i < Nary; i++)
        {
          int factorProcs = theTwoSided().numProcRanks().toInt() / Nary * Nary;
          if (factorProcs < theTwoSided().numProcRanks().toInt())
            factorProcs += Nary;
          int child = (rank * Nary + i + 1) - factorProcs - 1;
          //int child = (rank + 1 + theTwoSided().numProcRanks().toInt() - 1) / Nary;
          if (child < theTwoSided().numProcRanks().toInt() && child >= 0)
            {
              finalChildren.push_back(child);
              childrenLoad.push_back(PESample(child));
              //printf("\t\t%d: final child = %d\n",
              //theTwoSided().getProcRank().toInt(), child);
            }
        }
    }

  numChildren += children.size() + finalChildren.size() + 1;

  // Make sure that all processors have set numChildren before AM arrive
  //theColl().dispatchBarrier();

  const int parent = (theTwoSided().getProcRank().toInt() - 1) / Nary;

  int factorProcs = theTwoSided().numProcRanks().toInt() / Nary * Nary;
  if (factorProcs < theTwoSided().numProcRanks().toInt())
    factorProcs += Nary;
  const int bottomParent = ((theTwoSided().getProcRank().toInt() + 1
      + factorProcs) - 1) / Nary;

  //trace_proj_event_stop(Timer::timestamp(), traceEventTypes["LB_build_tree"]);

  //trace_proj_event_start(Timer::timestamp(), traceEventTypes["LB_leaf_donate"]);

  if (peLoad > 1.003 * avgLoad)
    {
#if 0
      map<int, list<long int> > toSend;

      while (peLoad > 1.003 * avgLoad && accessCount.size() > 0)
        {
          map<long int, pair<int, int> >::iterator found;
          double ratio = -1;
          for (map<long int, pair<int, int> >::iterator iter = accessCount.begin();
              iter != accessCount.end(); ++iter)
            {
              double total = iter->second.first + iter->second.second;
              int remote = iter->second.second;
              if (remote / total > ratio)
                {
                  ratio = remote / total;
                  found = iter;
                }
            }

          for (map<int, list<long int> >::iterator iter = taskSample.begin();
              iter != taskSample.end(); ++iter)
            {
              for (list<long int>::iterator iter2 = iter->second.begin();
                  iter2 != iter->second.end(); ++iter2)
                {
                  if (*iter2 == found->first)
                    {
                      peLoad -= iter->first;
                      toSend[iter->first].push_back(*iter2);
                      iter->second.remove(*iter2);
                      break;
                    }
                }
            }

          accessCount.erase(found);
        }

      vector<long int> localTaskPool;
      for (map<int, list<long int> >::iterator iter = toSend.begin();
          iter != toSend.end(); ++iter)
        {
          localTaskPool.push_back(-1);
          localTaskPool.push_back(iter->first);
          for (list<long int>::iterator iter2 = iter->second.begin();
              iter2 != iter->second.end(); ++iter2)
            {
              localTaskPool.push_back(*iter2);
            }
        }

#else
      vector<long int> localTaskPool;
      std::map<int, std::list<long int> >::iterator item = taskSample.begin();

      if (item != taskSample.end())
        {
          localTaskPool.push_back(-1);
          localTaskPool.push_back(item->first);
        }

      int taskCount = 0;

      while (peLoad > 1.003 * avgLoad && item != taskSample.end())
        {
          assert(item->second.size() != 0);

          taskCount++;
          localTaskPool.push_back(item->second.back());
          //printf("%d: adding item to pool: size %d, id = %d\n",
          //theTwoSided().getProcRank(), item->first, item->second.back());
          item->second.pop_back();
          peLoad -= item->first;

          if (item->second.size() == 0)
            {
              item++;
              if (item != taskSample.end())
                {
                  localTaskPool.push_back(-1);
                  localTaskPool.push_back(item->first);
                }
            }
        }

      //printf("%d: sending %d tasks up, length = %d, bytes = %d\n",
      //theTwoSided().getProcRank().toInt(), taskCount, localTaskPool.size(),
      //sizeof(long int) * localTaskPool.size());
#endif

      LBTreeMsg arg(peLoad, rank);
      arg.data = &localTaskPool[0];
      arg.dlen = sizeof(long int) * localTaskPool.size();
      arg.count = 1;
      AmRequest* req = AmRequest::construct();
      theAm().amPut(bottomParent, arg, treeHan, req);
      dispatcher.registerCodelet(req);
      dispatchUntilEmpty();
    }
  else
    {
      LBTreeMsg msg(peLoad, rank);
      msg.count = 1;
      AmRequest* req = AmRequest::construct();
      theAm().amPut(bottomParent, msg, treeHan, req);
      dispatcher.registerCodelet(req);
      dispatchUntilEmpty();
    }

  //trace_proj_event_stop(Timer::timestamp(), traceEventTypes["LB_leaf_donate"]);

  // Wait for all children to send their loads
  while (numChildren != 0)
    {
      AmListenObjCodelet<NullMutex>* lcodelet;
      if ((lcodelet = theAm().amListeners[0]->progress()) != NULL)
        {
          lcodelet->execute();
        }
    }

  //trace_proj_event_start(Timer::timestamp(), traceEventTypes["LB_send_up"]);

  if (childrenLoad.size() == 0)
    {
      LBTreeMsg msg(-1, rank);
      AmRequest* req = AmRequest::construct();
      theAm().amPut(parent, msg, treeHan, req);
      dispatcher.registerCodelet(req);
      dispatchUntilEmpty();
      assert(parent != theTwoSided().getProcRank().toInt());
    }
  else
    {
      make_heap(childrenLoad.begin(), childrenLoad.end(), CompPEA<PESample>());

      // Iterate from largest to smallest item
      std::map<int, std::list<long int> >::reverse_iterator cIter =
          givenTasks.rbegin();

      //printf("minLoad = %f\n", childrenLoad.front().load);

      while (cIter != givenTasks.rend()
          && (theTwoSided().getProcRank().toInt() == parent
              || childrenLoad.front().load / childrenLoad.front().count
                  < 0.95 * avgLoad))
        {
          assert(cIter->second.size() != 0);

          PESample minPE = childrenLoad.front();
          //printf("XXX minPE = %d, load = %f, adding task %d\n",
          //minPE.pe, minPE.load / minPE.count, cIter->second.back());
          pop_heap(childrenLoad.begin(), childrenLoad.end(),
              CompPEA<PESample>());
          childrenLoad.pop_back();

          minPE.recs[cIter->first].push_back(cIter->second.back());
          minPE.load += cIter->first;
          totalChildLoad += cIter->first;
          cIter->second.pop_back();

          if (cIter->second.size() == 0)
            {
              cIter++;
            }

          childrenLoad.push_back(minPE);
          push_heap(childrenLoad.begin(), childrenLoad.end(),
              CompPEA<PESample>());
        }

      int size = 0;
      for (std::map<int, std::list<long int> >::iterator iter =
          givenTasks.begin(); iter != givenTasks.end(); ++iter)
        {
          size += iter->second.size();
        }

      // Not root
      if (rank != parent)
        {
          if (size > 0)
            {
              std::vector<long int> extraTasks;

              for (std::map<int, std::list<long int> >::iterator iter =
                  givenTasks.begin(); iter != givenTasks.end(); ++iter)
                {
                  extraTasks.push_back(-1);
                  extraTasks.push_back(iter->first);

                  for (std::list<long int>::iterator iter2 =
                      iter->second.begin(); iter2 != iter->second.end();
                      ++iter2)
                    {
                      //printf("extra task id = %ld\n", *iter2);
                      extraTasks.push_back(*iter2);
                    }
                }

              LBTreeMsg arg(totalChildLoad, rank);
              //printf("%d upCount = %d\n", theTwoSided().getProcRank(), upCount);
              arg.count = upCount;
              arg.data = &extraTasks[0];
              arg.dlen = sizeof(long int) * extraTasks.size();
              AmRequest* req = AmRequest::construct();
              theAm().amPut(parent, arg, treeHan, req);
              dispatcher.registerCodelet(req);
              dispatchUntilEmpty();
              assert(parent != theTwoSided().getProcRank().toInt());
              givenTasks.clear();
            }
          else
            {
              LBTreeMsg msg(totalChildLoad, rank);
              //printf("%d upCount = %d\n", theTwoSided().getProcRank(), upCount);
              msg.count = upCount;
              AmRequest* req = AmRequest::construct();
              theAm().amPut(parent, msg, treeHan, req);
              dispatcher.registerCodelet(req);
              dispatchUntilEmpty();
              assert(parent != theTwoSided().getProcRank().toInt());
            }
        }
      else
        {
          assert(size == 0);

          // printf("root, proc = %d, task count = %d\n",
          // theTwoSided().getProcRank().toInt(), givenTasks.size());

          // Root: start sending down
          for (unsigned i = 0; i < childrenLoad.size(); i++)
            {
              //printf("\tat root %d load = %f\n", childrenLoad[i].pe, childrenLoad[i].load);

              if (childrenLoad[i].pe == theTwoSided().getProcRank().toInt())
                {
                  for (std::map<int, std::list<long int> >::iterator iter =
                      childrenLoad[i].recs.begin();
                      iter != childrenLoad[i].recs.end(); ++iter)
                    {
                      for (std::list<long int>::iterator iter2 =
                          iter->second.begin(); iter2 != iter->second.end();
                          ++iter2)
                        {
                          taskSample[iter->first].push_back(*iter2);
                          peLoad += iter->first;
                          //printf("adding task, id = %d\n", *iter2);
                        }
                    }
                  childrenLoad[i].recs.clear();
                }
              else
                {
                  vector<long int> tasks;
                  for (std::map<int, std::list<long int> >::iterator iter =
                      childrenLoad[i].recs.begin();
                      iter != childrenLoad[i].recs.end(); ++iter)
                    {
                      if (iter->second.size() != 0)
                        {
                          tasks.push_back(-1);
                          tasks.push_back(iter->first);
                        }
                      for (std::list<long int>::iterator iter2 =
                          iter->second.begin(); iter2 != iter->second.end();
                          ++iter2)
                        {
                          tasks.push_back(*iter2);
                        }
                    }

                  childrenLoad[i].recs.clear();

                  LBFinalMsg arg;
                  arg.data = &tasks[0];
                  arg.dlen = sizeof(long int) * tasks.size();
                  arg.final = false;
                  AmRequest* req = AmRequest::construct();
                  theAm().amPut(childrenLoad[i].pe, arg, downTree, req);
                  dispatcher.registerCodelet(req);
                  dispatchUntilEmpty();
                }
            }
        }
    }

  //trace_proj_event_stop(Timer::timestamp(), traceEventTypes["LB_send_up"]);

  while (!reachedLeaf)
    {
      AmListenObjCodelet<NullMutex>* lcodelet;
      if ((lcodelet = theAm().amListeners[0]->progress()) != NULL)
        {
          lcodelet->execute();
        }
    }

  printf("%d: NEW load = %f\n", theTwoSided().getProcRank().toInt(), peLoad);

  //trace_proj_event_start(Timer::timestamp(), traceEventTypes["LB_add_queue"]);

  // Reset temp queue inside UTC
  utc->nonNullLength = 0;

  for (std::map<int, std::list<long int> >::iterator iter = taskSample.begin();
      iter != taskSample.end(); ++iter)
    {
      for (std::list<long int>::iterator iter2 = iter->second.begin();
          iter2 != iter->second.end(); ++iter2)
        {
          task_dscr_t dscr;
          dscr.taskid = *iter2;
          utc->isNotReadOnly(&dscr, sizeof(dscr));
          //printf("%d: bin = %d, final task = %ld\n", theTwoSided().getProcRank(), iter->first, *iter2);
        }
    }

  amBarrier();

  //trace_proj_event_stop(Timer::timestamp(), traceEventTypes["LB_add_queue"]);
}

void
greedyLB()
{
  //printf("greedyLB()\n");
  if (theTwoSided().getProcRank() != 0)
    {
      LBRecord* recs =
          (LBRecord*) alloca(sizeof(LBRecord) * taskDuration.size());

      unsigned i = 0;
      for (std::map<long int, double>::iterator iter = taskDuration.begin();
          iter != taskDuration.end(); ++iter)
        {
          recs[i].rec = iter->first;
          recs[i].time = iter->second;
          i++;
        }

      AmArg arg;
      arg.data = recs;
      arg.dlen = sizeof(LBRecord) * taskDuration.size();

      AmRequest* req = AmRequest::construct();
      theAm().amPut(0, arg, rhan, req);
      dispatcher.registerCodelet(req);
      dispatchUntilEmpty();

      while (!loadDataRecv)
        {
          AmListenObjCodelet<NullMutex>* lcodelet;
          if ((lcodelet = theAm().amListeners[0]->progress()) != NULL)
            {
              lcodelet->execute();
            }
        }
    }
  else
    {
      for (std::map<long int, double>::iterator iter = taskDuration.begin();
          iter != taskDuration.end(); ++iter)
        {
          LBRecord lrec;
          lrec.rec = iter->first;
          lrec.time = iter->second;
          lbData.push_back(lrec);
        }

      while (peRec != theTwoSided().numProcRanks().toInt() - 1)
        {
          AmListenObjCodelet<NullMutex>* lcodelet;
          if ((lcodelet = theAm().amListeners[0]->progress()) != NULL)
            {
              lcodelet->execute();
            }
        }

      make_heap(lbData.begin(), lbData.end(), CompLBRecord());

      vector<PE> pes(theTwoSided().numProcRanks().toInt());

      for (unsigned i = 0; i < theTwoSided().numProcRanks().toInt(); i++)
        {
          pes[i].pe = i;
          pes[i].load = 0.0;
        }

      make_heap(pes.begin(), pes.end(), CompPE<PE>());

      /*printf("max elem = %f\n", lbData.front().time);
       printf("max load = %f\n", pes.front().load);*/

      int lbSize = lbData.size();
      for (unsigned i = 0; i < lbSize; i++)
        {
          LBRecord maxRec = lbData.front();
          pop_heap(lbData.begin(), lbData.end(), CompLBRecord());
          lbData.pop_back();

          PE minPE = pes.front();
          pop_heap(pes.begin(), pes.end(), CompPE<PE>());
          pes.pop_back();

          minPE.recs.push_back(maxRec);
          minPE.load += maxRec.time;

          pes.push_back(minPE);
          push_heap(pes.begin(), pes.end(), CompPE<PE>());
        }

      for (unsigned i = 0; i < theTwoSided().numProcRanks().toInt(); i++)
        {
          //printf("%d load = %f\n", pes[i].pe, pes[i].load);
          AmArg arg;
          arg.data = &pes[i].recs[0];
          arg.dlen = sizeof(LBRecord) * pes[i].recs.size();
          AmRequest* req = AmRequest::construct();
          theAm().amPut(pes[i].pe, arg, ldhanSample, req);
          dispatcher.registerCodelet(req);
          dispatchUntilEmpty();
        }
    }

  amBarrier();
}

static bool isInited = false;

static void
cpptwoel(Integer g_schwarz, Integer g_dens, Integer g_fock, double _schwmax,
    Integer nbfn, long long *_icut1, long long *_icut2, long long *_icut3,
    long long *_icut4)
{

  if (!isInited)
    {
      rhan = theAm().amRegister(0, recvData, 0, 0, 0);
      rpehan = theAm().amRegister(0, recvPEData, 0, 0, 0);
      ldhanSample = theAm().amRegister(0, newLoadSample, 0, 0, 0);
      ldhanDist = theAm().amRegister(0, newLoadDist, 0, 0, 0);
      ldhan = theAm().amRegister(0, newLoad, 0, 0, 0);
      rsamplePE = theAm().amRegister(0, recvPESample, 0, 0, 0);
      rdistSamplePE = theAm().amRegister(0, recvPEDist, 0, 0, 0);
      treeHan = theAm().amRegister(0, treeFun, 0, 0, 0);
      orderHan = theAm().amRegister(0, orderFun, 0, 0, 0);
      downTree = theAm().amRegister(0, downTreeFun, 0, 0, 0);
      pes.resize(theTwoSided().numProcRanks().toInt());
      pesm.resize(theTwoSided().numProcRanks().toInt());
      ped.resize(theTwoSided().numProcRanks().toInt());
      for (unsigned i = 0; i < theTwoSided().numProcRanks().toInt(); i++)
        {
          pes[i].pe = i;
          pesm[i].pe = i;
          ped[i].pe = i;
        }
      isInited = true;
    }

  int dotask;
  int g_counter, ione = 1;

  const long me = GA_Nodeid();
  const long nproc = GA_Nnodes();
  const long imax = nbfn / ichunk + ((nbfn % ichunk) ? 1 : 0);
  const long total_ntasks = imax * imax * imax * imax;
  const long ntasks_per_proc = (long) ceil(1.0 * total_ntasks / nproc);
  const long tasklo = ntasks_per_proc * me;
  const long taskhi = min(tasklo + ntasks_per_proc, total_ntasks) - 1;

    {
      TslFuncRegTbl* frt = 0;
      if (!utc)
        {
          task_plo_t* tplo = (task_plo_t*) malloc(sizeof(task_plo_t));
          tplo->nbfn = nbfn;
          tplo->schwmax = _schwmax;
          tplo->icut1 = _icut1;
          tplo->icut2 = _icut2;
          tplo->icut3 = _icut3;
          tplo->icut4 = _icut4;
          tplo->g_schwarz = g_schwarz;
          tplo->g_dens = g_dens;
          tplo->g_fock = g_fock;
          tplo->bmgr = new MemPoolManager(ichunk * ichunk * sizeof(double), 6);

          int block[2] =
            { ichunk, ichunk };
          vector<DataColl*> colls;

          task_dscr_t tdscr;
          frt = new TslFuncRegTbl();
          TslFunc tf = frt->add(twoel_task);
          TaskCollProps props;
          props.functions(tf, frt).taskSize(sizeof(task_dscr_t))
#define EVEN_DISTRIBUTION 0
#if EVEN_DISTRIBUTION
              .maxTasks(ntasks_per_proc)
#else
              .maxTasks(total_ntasks)
#endif
              .localData(tplo, sizeof(*tplo));

          utc = new UniformTaskCollSplitHybrid(props,
              sstmac::sw::tascel_app::get_current_config());

          long numAdded = 0;

          for (long i = tasklo; i <= taskhi; ++i)
            {
              if (1 || me == compute_owner(i, colls, 0, nbfn))
                {
                  tdscr.taskid = i;
#if 1
                  utc->addTask(&tdscr, sizeof(tdscr));
                  numAdded++;
#else
                  vector<void *> bufs;
                  twoel_task(utc, &tdscr, sizeof(tdscr), tplo, sizeof(*tplo), bufs);
#endif
                }
            }

          utc->persist();
        }

      GA_Sync();

      // Clear previous timing data for load balance
      taskDuration.clear();
      taskSample.clear();
      peLoad = 0.0;
      peLoadCount = 0;

      utc->process();

      printf("%d: iter %d load %f\n", theTwoSided().getProcRank().toInt(),
          iterUTC, peLoad);
      printf("%d: iter %d load count %d\n", theTwoSided().getProcRank().toInt(),
          iterUTC, peLoadCount);

      if (frt && iterUTC + 1 == mxiter)
        {
          delete frt;
        }
    }

  amBarrier();

  loadDataRecv = false;
  peRec = 0;
  pes.clear();
  pesm.clear();
  ped.clear();
  combinedSample.clear();
  totalLoad = 0.0;
  lbData.clear();
  pes.resize(theTwoSided().numProcRanks().toInt());
  pesm.resize(theTwoSided().numProcRanks().toInt());
  ped.resize(theTwoSided().numProcRanks().toInt());
  for (unsigned i = 0; i < theTwoSided().numProcRanks().toInt(); i++)
    {
      pes[i].pe = i;
      pesm[i].pe = i;
      ped[i].pe = i;
    }

  iterUTC++;

  if (iterUTC == mxiter)
    {
      delete utc;
    }

#if defined(DATA_TRACE)
  if (iterUTC > 2 && (iterUTC + 1) % 3 != 0)
    {
      trace_proj_stop();
    }
  else
    {
      trace_proj_start();
    }
#endif
}

static void
twoel_task(UniformTaskCollection *utc, void *_bigd, int bigd_len, void *pldata,
    int pldata_len, vector<void *> data_bufs)
{
  double time = MPI::Wtime();
  assert(_bigd!=NULL);
  assert(bigd_len == sizeof(task_dscr_t));
  assert(pldata!=NULL);
  assert(pldata_len == sizeof(task_plo_t));
  assert(data_bufs.size()==0);

  task_dscr_t *ptdscr = (task_dscr_t*) _bigd;
  task_plo_t *ptplo = (task_plo_t*) pldata;

  //printf("%ld: task\n", ptdscr->taskid);

  int ich;
  Integer lch;
  Integer lo_ik[2], hi_ik[2], lo_jl[2], hi_jl[2];
  Integer i, j, k, l, iloc, jloc, kloc, lloc, ld;
  double gg;
  Integer ione = 1;
  double one = 1.0;

  long lo[4], hi[4], it, jt, kt, lt;

  Integer nbfn;
  double schwmax;
  int cnt1, cnt2;
  long long *icut1, *icut2, *icut3, *icut4;
  Integer g_schwarz, g_dens, g_fock;

  int done_get0 = 0, done_get1 = 0, accum = 0, accumCount = 0;

  buf_t *f_ij, *d_kl, *f_ik, *d_jl, *s_ij, *s_kl;
  buf_t *buf_fij = NULL, *buf_fik = NULL, *buf_dkl = NULL, *buf_djl = NULL;
  buf_t *buf_sij = NULL, *buf_skl = NULL;

  nbfn = ptplo->nbfn;
  schwmax = ptplo->schwmax;
  cnt1 = ptplo->cnt1;
  cnt2 = ptplo->cnt2;
  icut1 = ptplo->icut1;
  icut2 = ptplo->icut2;
  icut3 = ptplo->icut3;
  icut4 = ptplo->icut4;
  g_schwarz = ptplo->g_schwarz;
  g_dens = ptplo->g_dens;
  g_fock = ptplo->g_fock;

  ld = maxnbfn;
  ich = lch = ichunk;

  compute_lo_hi(ptdscr->taskid, nbfn, lo, hi, &it, &jt, &kt, &lt);

  lo_ik[0] = lo[0];
  lo_ik[1] = lo[2];
  hi_ik[0] = hi[0];
  hi_ik[1] = hi[2];
  lo_jl[0] = lo[1];
  lo_jl[1] = lo[3];
  hi_jl[0] = hi[1];
  hi_jl[1] = hi[3];

  assert(ich == hi_jl[0]-lo_jl[0]+1);
  assert(lo[0]%ichunk==0);
  assert(lo[1]%ichunk==0);
  assert(lo[2]%ichunk==0);
  assert(lo[3]%ichunk==0);

//   buf_fij = bufs_request(ptplo);
//   buf_fik = bufs_request(ptplo);
  buf_fij = (buf_t*) ptplo->bmgr->request();
  buf_fik = (buf_t*) ptplo->bmgr->request();
  f_ij = buf_fij;
  f_ik = buf_fik;

    {
      /*lkji*/
      int _lo[4] =
        { lo[3], lo[2], lo[1], lo[0] };
      int _hi[4] =
        { hi[3], hi[2], hi[1], hi[0] };
      unsigned int _loE[4] =
        { lo[3], lo[2], lo[1], lo[0] };
      unsigned int _hiE[4] =
        { hi[3] + 1, hi[2] + 1, hi[1] + 1, hi[0] + 1 };
      int _lo_jl[2] =
        { lo[3], lo[1] };
      int _hi_jl[2] =
        { hi[3], hi[1] };

#if defined(DATA_TRACE)
      trace_proj_event_start(Timer::timestamp(), traceEventTypes["g_schwarz"]);
#endif

      if (0 && NGA_Locate(g_schwarz, &_lo[2]) == GA_Nodeid())
        {
          int ld, subscript[2] =
            { _lo[2] / ichunk, _lo[3] / ichunk };
          NGA_Access_block_grid(g_schwarz, subscript, &s_ij, &ld);
          assert(ld==ichunk);
        }
      else
        {
          //buf_sij = bufs_request(ptplo);
          buf_sij = (buf_t*) ptplo->bmgr->request();
          s_ij = buf_sij;

          /*int idxs[2];
           compute_index(ptdscr, sizeof(task_dscr_t), ptplo, pldata_len,
           SCHWARZ, idxs, 2*sizeof(int));
           DataColl *coll = TheDCR()->intToColl(dataColls[SCHWARZ]);
           int sz = coll->getSize(idxs, 2*sizeof(int));
           coll->get(idxs, 2*sizeof(int), s_ij->buf, sz);*/

          NGA_Get(g_schwarz, &_lo[2], &_hi[2], s_ij->buf, &ich);
          //printf("%f, %f\n", s_ij->buf[0][0], s_ij->buf[0][1]);
        }

#if defined(DATA_TRACE)
      trace_proj_event_stop(Timer::timestamp(), traceEventTypes["g_schwarz"]);
#endif
      //NGA_Get(g_schwarz, &_lo[0], &_hi[0], s_kl, &ich);
      //NGA_Get(g_dens,&_lo[0],&_hi[0], d_kl,&ich);
      //NGA_Get(g_dens,_lo_jl,_hi_jl, d_jl,&ich);
    }

  clean_chunk((double*) f_ij);
  clean_chunk((double *) f_ik);

  for (i = lo[0]; i <= hi[0]; i++)
    {
      iloc = i - lo[0];
      for (j = lo[1]; j <= hi[1]; j++)
        {
            {
              AmListenObjCodelet<NullMutex>* lcodelet;
              if ((lcodelet = theAm().amListeners[0]->progress()) != NULL)
                {
                  lcodelet->execute();
                }
            }
          jloc = j - lo[1];
          if (s_ij->buf[jloc][iloc] * (schwmax) < tol2e)
            {
              *icut1 = *icut1 + (hi[2] - lo[2] + 1) * (hi[3] - lo[3] + 1);
            }
          else
            {
              if (done_get0 == 0)
                {
                  int _lo[4] =
                    { lo[3], lo[2], lo[1], lo[0] };
                  int _hi[4] =
                    { hi[3], hi[2], hi[1], hi[0] };
                  unsigned int _loE[4] =
                    { lo[3], lo[2], lo[1], lo[0] };
                  unsigned int _hiE[4] =
                    { hi[3] + 1, hi[2] + 1, hi[1] + 1, hi[0] + 1 };
                  int _lo_jl[2] =
                    { lo[3], lo[1] };
                  int _hi_jl[2] =
                    { hi[3], hi[1] };
                  int imax = (nbfn / ichunk) + ((nbfn % ichunk) ? 1 : 0);

#if defined(DATA_TRACE)
                  trace_proj_event_start(Timer::timestamp(), traceEventTypes["g_schwarz2"]);
#endif
                  //NGA_Get(g_schwarz, &_lo[2], &_hi[2], s_ij, &ich);
                  if (0 && NGA_Locate(g_schwarz, &_lo[0]) == GA_Nodeid())
                    {
                      int ld, subscript[2] =
                        { _lo[0] / ichunk, _lo[1] / ichunk };
                      NGA_Access_block_grid(g_schwarz, subscript, &s_kl, &ld);
                      assert(ld==ichunk);
                    }
                  else
                    {
                      //buf_skl = bufs_request(ptplo);
                      buf_skl = (buf_t*) ptplo->bmgr->request();
                      s_kl = buf_skl;

                      /*int idxs[2];
                       compute_index(ptdscr, sizeof(task_dscr_t), ptplo, pldata_len,
                       SCHWARZ_2, idxs, 2*sizeof(int));
                       DataColl *coll = TheDCR()->intToColl(dataColls[SCHWARZ_2]);
                       int sz = coll->getSize(idxs, 2*sizeof(int));
                       coll->get(idxs, 2*sizeof(int), s_kl->buf, sz);*/

                      NGA_Get(g_schwarz, &_lo[0], &_hi[0], s_kl->buf, &ich);

                      //NGA_Get(g_dens,&_lo[0],&_hi[0], d_kl,&ich);
                      //NGA_Get(g_dens,_lo_jl,_hi_jl, d_jl,&ich);
                      done_get0 = 1;
                    }
#if defined(DATA_TRACE)
                  trace_proj_event_stop(Timer::timestamp(), traceEventTypes["g_schwarz2"]);
#endif
                }

              /* *cnt1+=1;*/
              for (k = lo[2]; k <= hi[2]; k++)
                {
                  kloc = k - lo[2];
                  for (l = lo[3]; l <= hi[3]; l++)
                    {
                      lloc = l - lo[3];
                      if (s_ij->buf[jloc][iloc] * s_kl->buf[lloc][kloc] < tol2e)
                        {
                          *icut2 = *icut2 + 1;
                        }
                      else
                        {
                          if (done_get1 == 0)
                            {
                              int _lo[4] =
                                { lo[3], lo[2], lo[1], lo[0] };
                              int _hi[4] =
                                { hi[3], hi[2], hi[1], hi[0] };
                              unsigned int _loE[4] =
                                { lo[3], lo[2], lo[1], lo[0] };
                              unsigned int _hiE[4] =
                                { hi[3] + 1, hi[2] + 1, hi[1] + 1, hi[0] + 1 };
                              int _lo_jl[2] =
                                { lo[3], lo[1] };
                              int _hi_jl[2] =
                                { hi[3], hi[1] };
                              unsigned int _lo_jlE[2] =
                                { lo[3], lo[1] };
                              unsigned int _hi_jlE[2] =
                                { hi[3] + 1, hi[1] + 1 };
                              int imax = (nbfn / ichunk)
                                  + ((nbfn % ichunk) ? 1 : 0);

#if defined(DATA_TRACE)
                              trace_proj_event_start(Timer::timestamp(), traceEventTypes["g_dens"]);
#endif
                              //NGA_Get(g_schwarz, &_lo[2], &_hi[2], s_ij, &ich);
                              //NGA_Get(g_schwarz, &_lo[0], &_hi[0], s_kl, &ich);
                              if (0
                                  && NGA_Locate(g_dens, &_lo[0]) == GA_Nodeid())
                                {
                                  int ld, subscript[2] =
                                    { _lo[0] / ichunk, _lo[1] / ichunk };
                                  NGA_Access_block_grid(g_dens, subscript,
                                      &d_kl, &ld);
                                  assert(ld==ichunk);
                                }
                              else
                                {
                                  //buf_dkl = bufs_request(ptplo);
                                  buf_dkl = (buf_t*) ptplo->bmgr->request();
                                  d_kl = buf_dkl;
                                  /*int idxs[2];
                                   compute_index(ptdscr, sizeof(task_dscr_t), ptplo, pldata_len,
                                   DENS, idxs, 2*sizeof(int));
                                   DataColl *coll = TheDCR()->intToColl(dataColls[DENS]);
                                   int sz = coll->getSize(idxs, 2*sizeof(int));
                                   coll->get(idxs, 2*sizeof(int), d_kl->buf, sz);*/

                                  NGA_Get(g_dens, &_lo[0], &_hi[0], d_kl->buf,
                                      &ich);
                                }

#if defined(DATA_TRACE)
                              trace_proj_event_stop(Timer::timestamp(), traceEventTypes["g_dens"]);
                              trace_proj_event_start(Timer::timestamp(), traceEventTypes["g_dens2"]);
#endif

                              if (0
                                  && NGA_Locate(g_dens, &_lo_jl[0])
                                      == GA_Nodeid())
                                {
                                  int ld, subscript[2] =
                                    { _lo_jl[0] / ichunk, _lo_jl[1] / ichunk };
                                  NGA_Access_block_grid(g_dens, subscript,
                                      &d_jl, &ld);
                                  assert(ld==ichunk);
                                }
                              else
                                {
                                  //buf_djl = bufs_request(ptplo);
                                  buf_djl = (buf_t*) ptplo->bmgr->request();
                                  d_jl = buf_djl;
                                  /*int idxs[2];
                                   compute_index(ptdscr, sizeof(task_dscr_t), ptplo, pldata_len,
                                   DENS_2, idxs, 2*sizeof(int));
                                   DataColl *coll = TheDCR()->intToColl(dataColls[DENS_2]);
                                   int sz = coll->getSize(idxs, 2*sizeof(int));
                                   coll->get(idxs, 2*sizeof(int), d_jl->buf, sz);*/
                                  NGA_Get(g_dens, _lo_jl, _hi_jl, d_jl->buf,
                                      &ich);
                                }
#if defined(DATA_TRACE)
                              trace_proj_event_stop(Timer::timestamp(), traceEventTypes["g_dens2"]);
#endif
                              done_get1 = 1;
                            }
                          Integer _i = i + 1, _j = j + 1, _k = k + 1, _l = l
                              + 1;
                          g_(&gg, &_i, &_j, &_k, &_l);
                          /* *cnt2 += 1;*/
                          f_ij->buf[jloc][iloc] = f_ij->buf[jloc][iloc]
                              + gg * d_kl->buf[lloc][kloc];
                          f_ik->buf[kloc][iloc] = f_ik->buf[kloc][iloc]
                              - 0.5 * gg * d_jl->buf[lloc][jloc];
                          accum = 1;
                          accumCount++;
                          *icut3 = *icut3 + 1;
                        }
                    }
                }
            }
        }
    }

  UniformTaskCollSplitHybrid* hybrid =
      reinterpret_cast<UniformTaskCollSplitHybrid*>(utc);

  if (accum)
    {
      /*lkji*/
      int _lo[2] =
        { lo[1], lo[0] };
      int _hi[2] =
        { hi[1], hi[0] };
      unsigned int _loE[2] =
        { lo[1], lo[0] };
      unsigned int _hiE[2] =
        { hi[1] + 1, hi[0] + 1 };
      int _lo_ik[2] =
        { lo[2], lo[0] };
      int _hi_ik[2] =
        { hi[2], hi[0] };
      unsigned int _lo_ikE[2] =
        { lo[2], lo[0] };
      unsigned int _hi_ikE[2] =
        { hi[2] + 1, hi[0] + 1 };
#if defined(DATA_TRACE)
      trace_proj_event_start(Timer::timestamp(), traceEventTypes["g_fock"]);
#endif

      int idxs[2];
      int sz;
      DataColl *coll;
      /*compute_index(ptdscr, sizeof(task_dscr_t), ptplo, pldata_len,
       FOCK, idxs, 2*sizeof(int));
       DataColl *coll = TheDCR()->intToColl(dataColls[FOCK]);
       int sz = coll->getSize(idxs, 2*sizeof(int));
       coll->add(idxs, 2*sizeof(int), f_ij->buf, sz);*/

      NGA_Acc(g_fock, _lo, _hi, f_ij->buf, &ich, &one);

#if defined(DATA_TRACE)
      trace_proj_event_stop(Timer::timestamp(), traceEventTypes["g_fock"]);
      trace_proj_event_start(Timer::timestamp(), traceEventTypes["g_fock2"]);
#endif

      /*compute_index(ptdscr, sizeof(task_dscr_t), ptplo, pldata_len,
       FOCK_2, idxs, 2*sizeof(int));
       coll = TheDCR()->intToColl(dataColls[FOCK_2]);
       sz = coll->getSize(idxs, 2*sizeof(int));
       coll->add(idxs, 2*sizeof(int), f_ik->buf, sz);*/
      NGA_Acc(g_fock, _lo_ik, _hi_ik, f_ik->buf, &ich, &one);

#if defined(DATA_TRACE)
      trace_proj_event_stop(Timer::timestamp(), traceEventTypes["g_fock2"]);
#endif

      if (hybrid)
        hybrid->isNotReadOnly(ptdscr, sizeof(ptdscr));
    }
  else
    {
      list<int> colls;
      colls.push_back(dataColls[SCHWARZ]);
      if (hybrid)
        hybrid->isReadOnly(ptdscr, sizeof(ptdscr), &colls);
    }

  ptplo->bmgr->release(buf_fij);
  ptplo->bmgr->release(buf_dkl);
  ptplo->bmgr->release(buf_fik);
  ptplo->bmgr->release(buf_djl);
  ptplo->bmgr->release(buf_sij);
  ptplo->bmgr->release(buf_skl);

  double taskTime = MPI::Wtime() - time;
  int rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);

  if (accum)
    {
      //printf("%d task %ld took %f accumCount %d\n", rank, ptdscr->taskid,
      //taskTime, accumCount);
    }
  else
    {
      //printf("%d na task %ld took %f\n", rank, ptdscr->taskid, taskTime);
    }
}
