/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_BACKENDS_NATIVE_DRIVER_H_INCLUDED
#define SSTMAC_BACKENDS_NATIVE_DRIVER_H_INCLUDED

#include <sstmac/common/driver_util.h>

#include <sstmac/backends/native/manager.h>
#include <sstmac/backends/native/managerdegraded.h>

//node models
#include <sstmac/hardware/node/simplenode.h>

//nic models
#include <sstmac/hardware/nic/simplenic.h>

//event managers
#include <sstmac/backends/native/eventmap.h>

#include <sstream>

namespace sstmac { namespace native
{

  using namespace std;

  eventmanager::ptr
  run(const sim_parameters::ptr &params, timestamp until, bool print)
  {

    bool errorflag = false;



    interconnect::ptr fab = construct_circuit(params);

    /** Set up static allocation info */
    std::string nodefile = params->get_optional_param("launch_dumpi_mapname","");
    allocationstrategy::set_nodemap_file(nodefile);
    allocationstrategy::set_toplogy(fab->get_topology());

    /** I would like to avoid doing this, but there are
        so many calls to parameter node_num it's not worth
        going through the code to fix them all. Just
        take the interconnect object and override
        the parameter. JJW 01/09/12
    */
    std::stringstream sstr; sstr << fab->slots();
    params->add_param_override("node_num", sstr.str());

    debug dbg_("<debug> sstmac | sim | simulation");

    manager::ptr mgr;

    try{
        if(!params->has_param("manager") || params->get_param("manager") == "simple"){
            mgr = manager::construct();
        }else if(params->get_param("manager") == "degraded"){
            mgr = manager_degraded::construct();
        }

        mgr->set_interconnect(fab);

        mgr->add_nodes(params);

        SSTMAC_DEBUG << "-- nodes added -- \n";


        timestamp runtime = mgr->run(until);



        if (print)
          {
            dbg_ << "final time: ";
            dbg_ << runtime.sec() << "\n";
          }

        return mgr->get_eventmanager();

    } catch (const std::exception& e) {
        mgr->done();
        throw;
    }
  }

} }

#endif
