/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_BACKENDS_OMNETPP_NETWORKMODELS_PACKET_PACKETRACKSWITCH_H_INCLUDED
#define SSTMAC_BACKENDS_OMNETPP_NETWORKMODELS_PACKET_PACKETRACKSWITCH_H_INCLUDED

#include <omnetpp.h>
#include <map>

#include <networkmodels/rackswitch.h>

#include <sstmac/common/stats/stat_time_window.h>

namespace sstmac { namespace omnetpp {


  class packetrackswitch : public rackswitch
  {


  protected:

    virtual void
    handleMessage(cMessage* cmg);
    virtual void
    initialize();
    virtual void
    finish();

    std::map<int, cGate*> inports;
    std::map<int, cGate*> outports;

    void
    tryOutQ(int port);

    std::map<int, cPacketQueue*> outQs;

    std::map<cMessage*, long> transFinish;
    std::map<int, cMessage*> transMsg;

    std::map<int, int> bytes_;
    std::map<cMessage*, simtime_t> entrytimes_;

    static sstmac::stat_time_window_avg::ptr stat_Q_wait_;
    static sstmac::stat_time_window_sum_duration::ptr stat_pwr_;
    static sstmac::stat_time_window_sum::ptr stat_bytes_;

    cMessage* displayUpdate;
    simtime_t updateInterval_;
  };

} }
#endif
