/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <networkmodels/routing/dragonflyrouter.h>

namespace sstmac
{
  namespace omnetpp
  {

    int
    dragonflyrouter::route(NetworkMsg* msg, omnetpp_address::ptr to,
        int inport, bool &dim_change)
    {
      SSTMAC_DEBUG << "me: " << me->to_string() << " routing to "
            << to->to_string() << "\n";

      int hisgroup = to->rack() / 96;
      int hisX;
      int hisY;
      int hisrack;

      if (hisgroup != group_)
        {
          SSTMAC_DEBUG << "my group is " << group_ << ", his group is "
                << hisgroup << "\n";
          int cluster = hisgroup / 5;
          int port = cluster * 5 + hisgroup;
          int factor = rand() % 2;
          hisrack = factor + (port / 5);
          if (hisrack == rack_)
            {
              return nodesPerRack + 22 + (port % 5);
            }
        }
      else
        {
          hisrack = to->rack() - hisgroup * 96;
        }

      if (hisrack == rack_)
        {
          return to->node();
        }
      else
        {
          hisX = hisrack % 16;
          hisY = hisrack / 16;
        }

      if (hisX != myX_)
        {
          return nodesPerRack + hisX;
        }
      else if (hisY != myY_)
        {
          return nodesPerRack + 16 + hisY;
        }
      else
        {
          std::cout << "myX = " << myX_ << ", myY = " << myY_ << ", hisrack = "
              << hisrack << ", hisX = " << hisX << ", hixY = " << hisY << "\n";
          throw ssterror(
              "dragonflyrouter: something's wrong, should be there by now");
        }

    }
  }

}

