/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <networkmodels/routing/meshrouter.h>

namespace sstmac
{
  namespace omnetpp
  {

    int
    meshrouter::route(NetworkMsg* msg, omnetpp_address::ptr to, int inport, bool &dim_change)
    {

      int hisX = to->rack() % numX;
      int hisY = (to->rack() / numX) % numY;


      SSTMAC_DEBUG << "Routing to address " << to->to_string() << ": (" << hisX
          << ", " << hisY << ")";

      int ret = -1;

      dim_change = (inport <= nodesPerRack);  //constrain injection

      if (hisX > myX)
        {

          ret = nodesPerRack + 0;

        }
      else if (hisX < myX)
        {
          ret = nodesPerRack + 1;
        }
      else
        {
          if (inport == nodesPerRack + 0 || inport == nodesPerRack + 1)
            dim_change = true;
          else
            dim_change = dim_change || false;

          if (hisY > myY)
            {
              ret = nodesPerRack + 2;
            }
          else if (hisY < myY)
            {
              ret = nodesPerRack + 3;
            }
          else
            {
                ret = to->node();
                dim_change = false;
            }

        }

      SSTMAC_DEBUG << ": outport " << ret << "\n";

      return ret;
    }

  }
}
