/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_BACKENDS_OMNETPP_NODE_PROCESSOR_ABSTRACTNODEPROCESSOR_H_INCLUDED
#define SSTMAC_BACKENDS_OMNETPP_NODE_PROCESSOR_ABSTRACTNODEPROCESSOR_H_INCLUDED

#include <omnetpp.h>

#include <sstmac/software/ami/environment.h>

#include <omnetppaddress.h>
#include <node_m.h>

#include <sstmac/common/debug.h>

#include <sstmac/common/sim_parameters.h>


#include <sstmac/common/messages/sst_message.h>
#include <sstmac/software/process/softwareid.h>

#include <sstmac/common/ptr_type.h>

#include <sstmac/hardware/node/node.h>
#include <sstmac/common/eventmanager.h>
#include <node/processor/eventhandlerpod.h>
#include <queue>



#include <sstmac/backends/omnetpp/node/processor/fakenic.h>

namespace sstmac
{
  namespace omnetpp
  {

    //class Application;

#define APP_STACK_SIZE 16384

#define OS_MSG_HEADER_SIZE 8
#define APP_SIZE 1000

#define APP_START_DELAY 0.00001

#define SEND_DELAY 0.00001

    class abstractnodeprocessor : public cSimpleModule
    {

      friend class fakeeventmanager;

    protected:

      class fakeeventmanager : public sstmac::eventmanager
      {
        abstractnodeprocessor *parent_;

        fakeeventmanager(abstractnodeprocessor *p) :
            parent_(p)
        {

        }

      public:

        static boost::intrusive_ptr<fakeeventmanager>
        construct(abstractnodeprocessor *p)
        {
          boost::intrusive_ptr<fakeeventmanager> ret(new fakeeventmanager(p));
          return ret;
        }

        virtual void
        clear(const sstmac::timestamp &zero_time = sstmac::timestamp(0))
        {

        }

        virtual void
        run()
        {

        }

        virtual void
        step(){

        }

        virtual const sstmac::timestamp&
        now()
        {
          now_ = parent_->now();
          return now_;
        }

        /// Set off the given eventhandler at the given time.
        virtual void
        schedule(const sstmac::timestamp &start_time,
            const sstmac::eventhandler::ptr &handler,
            const sstmac::sst_message::ptr &msg)
        {
          payloadpod ppod(msg);
          eventhandlerpod epod(handler);

          EventWrapperMsg* emsg = new EventWrapperMsg();
          emsg->setMsgcontent(ppod);
          emsg->setEvhandler(epod);

          double d = start_time.sec();
          simtime_t time(d);

          //have to check for a conversion error
          if(time < parent_->opp_simtime())
            parent_->scheduleAt(parent_->opp_simtime(), emsg);
          else
            parent_->scheduleAt(time, emsg);
        }
      };

    public:
      abstractnodeprocessor();
      virtual
      ~abstractnodeprocessor();


      omnetpp_address::ptr
      get_addr()
      {
        return myid;
      }

      simtime_t
      opp_simtime(){
        return simTime();
      }

    protected:

      virtual void
      handleMessage(cMessage* cmg);
      virtual void
      initialize();
      virtual void
      finish();






      bool
      is_master_core();

      omnetpp_address::ptr myid;
      int coresPerProc, procsPerNode, numOfNodes;

      int nodesPerRack;
      int numRacks;


      cGate* outgate;
      cGate* ingate;

      sstmac::hw::node::ptr node_;
      friend class fakenic;
      fakenic::ptr nic_;

      static debug dbg_;


      virtual sstmac::timestamp
      now();

      // ------------------------------------ //

    };

  }
}
#endif
