/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/backends/omnetpp/omnetppaddress.h>


  // ------- these functions get used if this is in an omnet packet ------ //
  void doPacking(cCommBuffer *buffer, sstmac::omnetpp::omnetpp_address::ptr& addr)
  {
    // addr.dbg_ << "Address: packing \n";

    buffer->pack(addr->rack());
    buffer->pack(addr->node());
    buffer->pack(addr->proc());
    buffer->pack(addr->core());

  }

  void doUnpacking(cCommBuffer *buffer, sstmac::omnetpp::omnetpp_address::ptr& addr)
  {

    int64_t rackval, nodeval, procval, coreval;
    buffer->unpack(rackval);
    buffer->unpack(nodeval);
    buffer->unpack(procval);
    buffer->unpack(coreval);

    addr = sstmac::omnetpp::omnetpp_address::construct(coreval, procval, nodeval, rackval);
    // addr.dbg_ << "Address: unpacked \n";
  }


namespace sstmac { namespace omnetpp {

  long omnetpp_address::numCores;
  long omnetpp_address::numProcs;
  long omnetpp_address::numNodes;
  long omnetpp_address::numRacks;




  // ----- these functions are used if this is in an sst_message ---- //
  void omnetpp_address::serialize(sst_serializer* ser) const
  {
    ser->pack(id.size());

    std::vector<int64_t>::const_iterator it, end = id.end();

    for (it = id.begin(); it != end; it++)
      {
        ser->pack(*it);
      }
  }

  void omnetpp_address::deserialize(sst_serializer* ser)
  {
    //std::vector<int64_t> v;
    id.clear();
    size_t num;
    ser->unpack(num);

    for (size_t i = 0; i < num; i++)
      {
        int64_t val;
        ser->unpack(val);
        id.push_back(val);
      }

    //omnetpp_address::ptr ret = construct(v);
    //return ret;
  }

} }
