//////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2004 Mellanox Technologies LTD. All rights reserved.
//
// This software is available to you under a choice of one of two
// licenses.  You may choose to be licensed under the terms of the GNU
// General Public License (GPL) Version 2, available from the file
// COPYING in the main directory of this source tree, or the
// OpenIB.org BSD license below:
//
//     Redistribution and use in source and binary forms, with or
//     without modification, are permitted provided that the following
//     conditions are met:
//
//      - Redistributions of source code must retain the above
//        copyright notice, this list of conditions and the following
//        disclaimer.
//
//      - Redistributions in binary form must reproduce the above
//        copyright notice, this list of conditions and the following
//        disclaimer in the documentation and/or other materials
//        provided with the distribution.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
// BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
// ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
// CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
//
//////////////////////////////////////////////////////////////////////////////
//
// IB Packets Generator:
// Send IB credits one at a time. 
// 
// Internal Messages: 
// push - inject a new data credit into the Queues
//
// External Messages:
// sent - a vHoQ was sent by the VLA
//
// MODES:
// Normal mode - every push a new data credit will be inejcted  
// Flood mode - on init and "sent" event the Q depth will be checked and filled.
//
/*
#include "ib_m.h"
#include "gen.h"
#include "vlarb.h"
#include "vec_file.h"

Define_Module(IBGenerator)
;

// use static to get a new packet id which is globaly unique
int
IBGenerator::getPacketId()
{
  static int id = 0;
  return (++id);
}

void
IBGenerator::parseIntListParam(char *parName, std::vector<int> &out)
{
  int cnt = 0;
  const char *str = par(parName);
  char *tmpBuf = new char[strlen(str) + 1];
  strcpy(tmpBuf, str);
  char *entStr = strtok(tmpBuf, " ,");
  if (out.size())
    out.clear();
  while (entStr)
    {
      cnt++;
      out.push_back(atoi(entStr));
      entStr = strtok(NULL, " ,");
    }
}

// allocate and init a new data credit
IBDataMsg *
IBGenerator::getNewDataMsg()
{
  IBDataMsg *p_cred;

  char name[128];
  sprintf(name, "data-%d-%d", packetId, creditCounter);
  p_cred = new IBDataMsg(name, IB_DATA_MSG);
  p_cred->setSL(SL);
  p_cred->setVL(packetVL);
  if (RemainPktLength >= creditSize)
    {
      p_cred->setByteLength(creditSize); // length is in bytes
      RemainPktLength -= creditSize;
    }
  else
    {
      p_cred->setByteLength(RemainPktLength); // length is in bytes
      RemainPktLength = 0;
    }
  p_cred->setCreditSn(creditCounter);
  p_cred->setPacketId(packetId);
  p_cred->setSrcLid(srcLid);
  p_cred->setDstLid(dstLid);
  p_cred->setPacketLength(packetLength);
  p_cred->setPacketLengthBytes(packetLengthBytes);
  return p_cred;
}

// initialize the parameters for a new data packet.
// optionally being given a VL to use.
void
IBGenerator::initIBPacketParams(int vl)
{
  int rand = -1;
  // Current packet parameters
  packetId = getPacketId();

  // In case we are in flood mode, generate packet as part of message.
  // Else Generate each packet independantly with length packetLength

  if (firstPacket == 1)
    {
      MsgLength = par("msgLengthInMTUs");
      MsgLength = MsgLength * mtu * creditSize;
      RemainMsgLength = MsgLength;
    }

  // Generate packet Length
  if (RemainMsgLength == 0)
    {
      opp_error(" pktid =%d", packetId);
    }
  packetLengthBytes = getPacketLen();

  // Make sure the minimal size of packet is always 8
  // Packet is at least 20B: LRH(8B) + BTH(12B)
  if (packetLengthBytes % creditSize < 8 && packetLengthBytes % creditSize != 0)
    {
      packetLengthBytes += (8 - packetLengthBytes % creditSize);
    }
  if (packetLengthBytes % creditSize == 0)
    {
      packetLength = packetLengthBytes / creditSize;
    }
  else
    {
      packetLength = packetLengthBytes / creditSize + 1;
    }

  creditCounter = 0; // count from 0 up.
  if (vl < 0)
    packetVL = par("VL");
  else
    packetVL = vl;

  // CHOSING DSTLID
  if (dstSeqMode != DST_BY_PARAM)
    {

      // only change dstLid if we crossed a message
      if (firstPacket || (GenModel == 1 && (RemainMsgLength == MsgLength))
          || (GenModel == 0 && (mtuInMsgIdx + 1 == MsgLength / (creditSize
              * mtu))))
        {

          if (dstBHS)
            {
              if (dstSeqMode == DST_HOT_SPOT && ((simTime() - dstBHS_start)
                  > dstBurstHSDel))
                {
                  if (!ev.isDisabled())
                    {
                      ev << "-I- " << getFullPath() << " HS->SL simTime():"
                          << simTime() << " dstBHS_start:" << dstBHS_start
                          << " diff:" << simTime() - dstBHS_start
                          << " dstBurstHSDel:" << dstBurstHSDel << endl;
                    }
                  dstSeqMode = DST_SEQ_LOOP;
                  dstBHS_start = simTime();
                }
              else if (dstSeqMode == DST_SEQ_LOOP
                  && ((simTime() - dstBHS_start) > dstBurstHSInterDel))
                {

                  if (!ev.isDisabled())
                    {
                      ev << "-I- " << getFullPath() << " SL->HS simTime():"
                          << simTime() << " dstBHS_start:" << dstBHS_start
                          << " diff:" << simTime() - dstBHS_start
                          << " dstBurstHSDel:" << dstBurstHSDel << endl;
                    }
                  dstSeqMode = DST_HOT_SPOT;
                  dstBHS_start = simTime();
                }
            }
          if (dstSeqMode == DST_HOT_SPOT)
            {
              rand = bernoulli(dstHotSpotPerc, 0);
            }
          else
            {
              rand = 0;
            }

          if (dstSeqMode == DST_HOT_SPOT && rand == 1)
            {
              dstLid = parDstLid;
            }
          else if (dstSeqMode == DST_RANDOM)
            {
              dstLid = (*dstSeq)[intuniform(0, dstSeq->size() - 1)];
            }
          else
            {
              dstLid = (*dstSeq)[dstSeqIdx];
              dstSeqIdx = ++dstSeqIdx;
              if (dstSeqIdx == dstSeq->size())
                {
                  dstSeqIdx = 0;
                }
              // In case we are in bursty HS, we do not want to get the
              // parameter Lid outside the window .
              if (dstSeqMode == DST_SEQ_LOOP && dstBHS && dstLid == parDstLid)
                {
                  dstLid = (*dstSeq)[dstSeqIdx];
                  dstSeqIdx = ++dstSeqIdx;
                  if (dstSeqIdx == dstSeq->size())
                    {
                      dstSeqIdx = 0;
                    }
                }
              if ((dstSeqMode == DST_SEQ_ONCE) && (dstSeqIdx == 0))
                {
                  dstSeqDone = 1;
                }
            }
          if (GenModel == 0)
            {
              mtuInMsgIdx = 0;
            }
        }
      else
        {
          if (GenModel == 0)
            {
              mtuInMsgIdx++;
            }
        }
    }
  else
    {
      dstLid = parDstLid;
    }

  if (dstLid == 0)
    {
      opp_error("Error: Packet %d has destination Lid = 0", packetId);
    }
  if (RemainMsgLength <= mtu * creditSize)
    {
      if (GenModel == 0)
        {
          MsgLength = par("msgLengthInMTUs");
          MsgLength = MsgLength * mtu * creditSize;
        }
      else
        {
          if (trafficDist != TRF_APP)
            {
              MsgLength = par("msgLength");
            }
          else
            {
              MsgLength = (*sizeSeq)[dstSeqIdx];
            }
        }
      RemainMsgLength = MsgLength;
    }
  else
    {
      RemainMsgLength = RemainMsgLength - packetLengthBytes;
    }
  RemainPktLength = packetLengthBytes;
  firstPacket = 0;
}

int
IBGenerator::getPacketLen()
{
  int PktLen = 0;
  if (GenModel == 0)
    {
      PktLen = par("packetlength");
      PktLen = PktLen * creditSize;
    }
  else
    {
      if (trafficDist == TRF_MIXLEN)
        {
          int TotProb = PktLenTotProb;
          int prob;
          int res;
          for (int i = 0; i < PktLenDist.size(); i++)
            {
              float k = (float) PktLenProb[i] / TotProb;
              res = bernoulli(k, 990);
              if (res == 1)
                {
                  PktLen = PktLenDist[i];
                  return PktLen;
                }
              else
                {
                  TotProb -= PktLenProb[i];
                }
            }
        }
      else if (trafficDist == TRF_UNI)
        {
          PktLen = par("packetlengthbytes");
          if (GenModel == 0)
            { // for ib_credits, generate packet length only multiple of 64
              if (PktLen % creditSize != 0)
                {
                  PktLen += (creditSize - (PktLen % creditSize));
                }
            }
        }
      else if (trafficDist == TRF_FLOOD || trafficDist == TRF_APP)
        {
          if (RemainMsgLength >= mtu * creditSize)
            {
              PktLen = mtu * creditSize;
            }
          else
            {
              PktLen = RemainMsgLength;
            }
        }
    }
  if (PktLen == 0)
    {
      opp_error(" PktLen is generated to 0");
    }
  return PktLen;
}

void
IBGenerator::initialize()
{
  // init parameters and state variables
  // In non flood mode, parameters:
  burstLength = par("burstLength");
  burstNumber = par("burstNumber");
  burstCounter = burstLength;
  // General parameters
  width = par("width");
  speed = par("speed");
  SL = par("SL");
  srcLid = par("srcLid");
  creditSize = par("creditSize");
  dstHotSpotPerc = par("dstHotSpotPerc");
  mtu = par("mtu");
  parDstLid = par("dstLid");

  // work with packets made of 64B credits. Bytes = 64*numCredits
  GenModel = 0;

  firstPacket = 1;
  firstPacketSent = 1;
  mtuInMsgIdx = 0;
  dstSeqIdx = 0;
  dstSeqDone = 0;
  packetId = 0;
  dstLid = 0;
  TimeLastSent = 0;
  AccBytesSent = 0;

  isDownstreamHoQFree_ = true;
  fromsource_ = gate("fromsource");

  // Destination Sequence Mode
  dstBHS = false;
  const char *dstSeqModePar = par("dstSeqMode");
  if (!strcmp(dstSeqModePar, "dstLid"))
    {
      dstSeqMode = DST_BY_PARAM;
    }
  else if (!strcmp(dstSeqModePar, "dstSeqOnce"))
    {
      dstSeqMode = DST_SEQ_ONCE;
    }
  else if (!strcmp(dstSeqModePar, "dstSeqLoop"))
    {
      dstSeqMode = DST_SEQ_LOOP;
    }
  else if (!strcmp(dstSeqModePar, "dstRandom"))
    {
      dstSeqMode = DST_RANDOM;
    }
  else if (!strcmp(dstSeqModePar, "dstHotSpot"))
    {
      dstSeqMode = DST_HOT_SPOT;
    }
  else if (!strcmp(dstSeqModePar, "dstBurstHotSpot"))
    {
      dstSeqMode = DST_HOT_SPOT;
      dstBHS = true;
      dstBHS_start = simTime();
    }
  else
    {
      dstSeqMode = DST_BY_PARAM;
    }

  // destination related parameters
  if (dstSeqMode != DST_BY_PARAM)
    {
      const char *dstSeqVecFile = par("dstSeqVecFile");
      const int dstSeqIndex = par("dstSeqIndex");
      vecFiles *vecMgr = vecFiles::get();
      dstSeq = vecMgr->getIntVec(dstSeqVecFile, dstSeqIndex);
      if (dstSeq == NULL)
        {
          opp_error("fail to obtain dstSeq vector");
        }
      if (!ev.isDisabled())
        ev << "-I- Defined sequence of " << dstSeq->size() << " LIDs" << endl;
    }
  if (dstBHS)
    {
      dstBurstHSDel = par("dstBurstHSDel");
      dstBurstHSDel = dstBurstHSDel * 1e-3; // (convert to s)
      dstBurstHSInterDel = par("dstBurstHSInterDel");
      dstBurstHSInterDel = dstBurstHSInterDel * 1e-3; // (convert to s)
    }

  initIBPacketParams();
}

// find the VLA and check it HoQ is free...
int
IBGenerator::isRemoteHoQFree(int vl)
{
  // find the VLA connected to the given port and
  // call its method for checking and setting HoQ
  return isDownstreamHoQFree_;
}

void
IBGenerator::sendDataOut(IBDataMsg *p_msg)
{

  double delay = 0;
  // In case we are in flood mode, we want to be able
  // to send packet with delay between them. So far the intraburst delay
  // was not use in flood mode. This is our wayt to use it
  if (trafficDist == TRF_FLOOD || trafficDist == TRF_APP)
    {
      delay = par("intraBurstDelay"); // in nsec
      if (simTime() < TimeLastSent + delay * 1e-9)
        {
          delay = TimeLastSent.dbl() * 1e+9 + delay - simTime().dbl() * 1e+9;
        }
      TimeLastSent = simTime() + delay * 1e-9;
    }
  // time stamp to enable tracking time in Fabric
  p_msg->setTimestamp(TimeLastSent);
  sendDelayed(p_msg, delay * 1e-9, "out");

  if (!ev.isDisabled())
    ev << "-I- " << getFullPath() << " sending " << p_msg->getName()
        << " packetlength(B):" << p_msg->getPacketLengthBytes() << " creditsn:"
        << p_msg->getCreditSn() << " dstLid:" << p_msg->getDstLid()
        << " @time(ns):" << (simTime() + delay * 1e-9) << endl;

  // For oBW calculations
  if (firstPacketSent)
    {
      firstPacketSent = 0;
      FirstPktSendTime = simTime();
    }
  AccBytesSent += p_msg->getByteLength();
}

// push is an internal event for generating a new packet
// should only be happening if floodMode == 0
void
IBGenerator::handlePush(cMessage *p_msg)
{

  if (trafficDist == TRF_FLOOD || trafficDist == TRF_APP)
    {
      opp_error("handlePush called with TRF_FLOOD / TRF_APP");
    }
  // get the new credit
  IBDataMsg *p_cred = getNewDataMsg();

  double delay = 1.0e-3 * 10 * p_cred->getByteLength() / width / speed;
  // is the Q free?
  if (Q[packetVL].empty() && isRemoteHoQFree(packetVL))
    {
      sendDataOut(p_cred);
    }
  else
    {
      // we can not send so we Q
      if (!ev.isDisabled())
        ev << "-I- " << getFullPath() << " queue packet:" << p_cred->getName()
            << " Time:" << simTime() << endl;

      Q[packetVL].insert(p_cred);
    }

  // Do we need a new IB packet?
  if (++creditCounter == packetLength)
    {
      initIBPacketParams();

      //  Is it last packet of burst ?
      if (--burstCounter == 0)
        {
          if ((trafficDist != TRF_FLOOD || trafficDist == TRF_APP))
            {
              burstNumber--;
            }
          // we need to send more burst?
          if ((trafficDist == TRF_FLOOD) || trafficDist == TRF_APP
              || burstNumber > 0)
            {
              burstCounter = burstLength;
              delay = par("interBurstDelay");
              delay = delay * 1e-3;
              delay += 1.0e-3 * 10 * p_cred->getByteLength() / width / speed;
              scheduleAt(simTime() + delay * 1e-6, p_msg);
            }
        }
      else
        {
          // schedule next sending within burst
          delay = par("intraBurstDelay");
          delay = delay * 1e-3;
          delay += 1.0e-3 * 10 * p_cred->getByteLength() / width / speed;
          scheduleAt(simTime() + delay * 1e-6, p_msg);
        }
    }
  else
    {
      // but not in addition to the current data being sent
      scheduleAt(simTime() + delay * 1e-6, p_msg);
    }
}

void
IBGenerator::handleSent(IBSentMsg *p_sent)
{
  int vl = p_sent->getVL();
  // We can not just send - need to see if the HoQ is free...
  // NOTE : since we LOCK the HoQ when asking if HoQ is free we
  // must make sure we have something to send before we ask about it
  if (!Q[vl].empty())
    {
      if (isRemoteHoQFree(vl))
        {
          IBDataMsg *p_msg = (IBDataMsg *) Q[vl].pop();
          if (!ev.isDisabled())
            ev << "-I- " << getFullPath() << " de-queue packet:"
                << p_msg->getName() << " at time " << simTime() << endl;
          sendDataOut(p_msg);
        }
      else
        {
          if (!ev.isDisabled())
            ev << "-I- " << getFullPath() << " HoQ not free for vl:" << vl
                << endl;
        }
    }
  else
    {
      if (!ev.isDisabled())
        ev << "-I- " << getFullPath() << " nothing to send on vl:" << vl
            << endl;
    }

  delete p_sent;
}

void
IBGenerator::handleMessage(cMessage *p_msg)
{
  if (p_msg->getArrivalGate() == fromsource_) // a new message from the nic
    {

    }
  else
    {
      int msgType = p_msg->getKind();
      if (msgType == IB_SENT_MSG)
        {
          handleSent((IBSentMsg *) p_msg);
        }
      else if (msgType == IB_HOQ_FREE)
        {
          IBHoQFree* free = check_and_cast<IBHoQFree*> (p_msg);
          isDownstreamHoQFree_ = free->getIsFree();
        }
      else
        {

          handlePush(p_msg);
        }
    }
}

void
IBGenerator::finish()
{
  double oBW = AccBytesSent / (simTime() - FirstPktSendTime);
  ev << "STAT: " << getFullPath() << " Gen Output BW (B/s):" << oBW << endl;
}
*/

