/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_COMMON_EVENTHANDLER_H_INCLUDED
#define SSTMAC_COMMON_EVENTHANDLER_H_INCLUDED

#include <sstmac/common/messages/sst_message.h>

#include <sstmac/common/ptr_type.h>

namespace sstmac {

  /**
   * The main interface for something that can respond to an event (sst_message).
   */
  class eventhandler : public virtual ptr_type
  {
  public:

    typedef boost::intrusive_ptr<eventhandler> ptr;

  protected:

    static eventhandler::ptr ignore;

    eventhandler()
    {

    }

  public:
    virtual std::string to_string() const{
      return "eventhandler";
    }

    virtual ~eventhandler()
    {
      //std::cout << "calling eventhandler destructor" << std::endl;
    }

    virtual void handle(const sst_message::ptr &msg)
    {

    }



  };
} // end of namespace sstmac
#endif
