/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_COMMON_MESSAGES_SST_SERIALIZER_H_INCLUDED
#define SSTMAC_COMMON_MESSAGES_SST_SERIALIZER_H_INCLUDED

#include <sstmac/common/timestamp.h>
#include <sstmac/common/richpair.h>
#include <sstmac/common/sstmac-config.h>
#include <typeinfo>

#include <sstmac/common/messages/payload.h>

namespace sstmac
{

  class nodeaddress;
  /**
   * Abstract base class for a serializer for running parallel
   * simulations. Different backends will want to implement this differently.
   */
  class sst_serializer
  {
  public:
    virtual boost::intrusive_ptr<nodeaddress>
    get_deser_addr() = 0;

    virtual
    ~sst_serializer()
    {
    }

    /**
     * Pure virtual pack an integer
     * @param i the integer
     */
    virtual void
    pack(int i) = 0;

    /**
     * Pure virtual unpack an integer
     * @param i Value returned in this variable
     */
    virtual void
    unpack(int &i) = 0;

    virtual void
    pack(long i) = 0;
    virtual void
    unpack(long &i) = 0;

    virtual void
    pack(long long unsigned int i) = 0;
    virtual void
    unpack(long long unsigned int &i) = 0;


    virtual void
    pack(long long int i) = 0;
    virtual void
    unpack(long long int &i) = 0;

    virtual void
    pack(size_t i) = 0;
    virtual void
    unpack(size_t &i) = 0;

    virtual void
    pack(std::string i) = 0;
    virtual void
    unpack(std::string &i) = 0;

    virtual void
    pack(timestamp i) = 0;
    virtual void
    unpack(timestamp &i) = 0;

    virtual void
    pack(bool i) = 0;
    virtual void
    unpack(bool &i) = 0;

    virtual void
    pack(double i) = 0;
    virtual void
    unpack(double &i) = 0;

    virtual void
    pack(richpair<int, int> i) = 0;
    virtual void
    unpack(richpair<int, int> &i) = 0;

    /**
     * Function that looks at the deserialization name of
     * the payload that was serialized, and makes a new one.
     * Children of payload need to register their class in
     * this function in order to be deserialized
     * @param ser the serializer instance
     * @return a smart pointer to a new payload object
     */
    static payload::const_ptr
    deserialize_content(sst_serializer* ser);

  };
} // end of namespace sstmac
#endif
