/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2011 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macro directory.
 */

#ifndef SSTMAC_COMMON_RICHPAIR_H_INCLUDED
#define SSTMAC_COMMON_RICHPAIR_H_INCLUDED

namespace sstmac {
  
  template <class T1, class T2>
  class richpair {
  public:
    T1 first;
    T2 second;
    // First and second are unitialized with this CTOR.
    richpair() {} 
    richpair(const T1 &a1, const T2 &a2): first(a1), second(a2) {} 
    richpair(const richpair<T1,T2> &rp): first(rp.first), second(rp.second) {}
    richpair(const std::pair<T1,T2> &rp): first(rp.first), second(rp.second) {}
    bool operator==(const richpair<T1,T2> &rp) const {
      return first == rp.first && second == rp.second;
    }
    bool operator<(const richpair<T1,T2> &rp) const {
      return first<rp.first || (first==rp.first && second<rp.second);
    }
    bool operator!=(const richpair<T1,T2> &rp) const {
      return ! ((*this) == rp);
    }
    bool operator>=(const richpair<T1,T2> &rp) const {
      return ! ((*this) < rp);
    }
    bool operator>(const richpair<T1,T2> &rp) const {
      return rp < (*this);
    }
    bool operator<=(const richpair<T1,T2> &rp) const {
      return !(rp < (*this));
    }
    richpair<T1,T2> operator+(const richpair<T1,T2> &rp) const {
      return richpair<T1,T2>(first + rp.first, second + rp.second);
    }
  };
  template <class T1, class T2>
  inline std::ostream& operator<<(std::ostream &o, const richpair<T1,T2> &b){
    o << "(" << b.first << "," << b.second << ")";
    return o;
  }

}

#endif
