/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef MEMORYMODEL_H_
#define MEMORYMODEL_H_

#include <sstmac/common/eventmodule.h>
#include <sstmac/hardware/failure/failable.h>

namespace sstmac
{
  namespace hw
  {

    class memorymodel : public eventmodule, public failable
    {
    protected:
      memorymodel(long nodeid);

      long nodeid_;
      static debug dbg_;

    public:
      typedef boost::intrusive_ptr<memorymodel> ptr;

      virtual
      ~memorymodel() { }

      virtual void
      do_access(sst_message::ptr msg) = 0;

      virtual std::string
      to_string() const
      {
        return "memorymodel";
      }

      virtual long
      failure_id() const
      {
        return nodeid_;
      }

      class link : public ptr_type
      {

      protected:
        double bw_;
        timestamp lat_;

        int current_accesses_;

        link(double bw, timestamp lat) :
          bw_(bw), lat_(lat), current_accesses_(0)
        {

        }

      public:
        typedef boost::intrusive_ptr<link> ptr;

        static ptr
        construct(double bw, timestamp lat)
        {
          return ptr(new link(bw, lat));
        }

        virtual ~link() { }

        virtual std::string
        to_string() const {
          return "memorymodel::link";
        }

        virtual timestamp
        new_access(long size);

        virtual void
        access_done();
      };

    };

  }
} /* namespace sstmac */
#endif /* MEMORYMODEL_H_ */
