/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/memory/simple_memorymodel.h>

namespace sstmac
{
  namespace hw
  {

    simple_memorymodel::simple_memorymodel(double mem_bw, timestamp mem_lat, eventhandler::ptr whendone, long nodeid) : memorymodel(nodeid),
       done_(whendone)
    {
      self_ = ptr(this);
      link_ = link::construct(mem_bw, mem_lat);
    }

    simple_memorymodel::~simple_memorymodel()
    {
      link_ = link::ptr();
    }

    void
    simple_memorymodel::handle(const sst_message::ptr &msg)
    {
      if(msg->get_type() == sst_message::FAILURE){
          failure_occurred(msg, self_);
          return;
      }
      link_->access_done();

      done_->handle(msg);
    }

    void
    simple_memorymodel::do_access(sst_message::ptr msg)
    {
      use_me();



      SSTMAC_DEBUG << "simple_memorymodel: doing access of " << msg->get_byte_length() << " \n";


      timestamp done = link_->new_access(msg->get_byte_length());
      schedule(now() + done, self_, msg);
    }

  }
} /* namespace sstmac */
