/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_HARDWARE_NETWORK_CONGESTION_CONSTANTLATENCY_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_CONGESTION_CONSTANTLATENCY_H_INCLUDED

#include <sstmac/hardware/network/congestion/latencymodel.h>
#include <sstmac/common/ptr_type.h>

namespace sstmac
{
  namespace hw
  {

    /**
     * The simplest latency model.  All transfers are assumed to have
     * a constant latency.
     */
    class constantlatency : public latencymodel
    {
      /// The latency we dole out.
      timestamp latency_;

      /// And off we go.
      constantlatency(const timestamp &lat) :
          latency_(lat)
      {
      }

    public:
      typedef boost::intrusive_ptr<constantlatency> ptr;

      /// May we offer you some constant latency?
      static constantlatency::ptr
      construct(const timestamp &lat)
      {
        return constantlatency::ptr(new constantlatency(lat));
      }

      /// No more Mr. Nice Constant Latency.
      virtual
      ~constantlatency()
      {
      }

      /// What sort of latency are we talking about?
      virtual timestamp
      latency(data) const
      {
        return latency_;

      }

      virtual std::string
      to_string() const
      {
        std::stringstream ss;
        ss << "(latency: " << latency_.sec() << ")";
        return ss.str();
      }
    };

  }
} // end of namespace sstmac.

#endif
