/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/network/congestion/sharedbuffer.h>
#include <sstmac/hardware/network/congestion/sharedbuffer-flowregistration.h>

#include <sstmac/common/errors.h>
#include <sstmac/common/debug.h>

namespace sstmac { namespace hw  {

  

  // Defaults:
  //  head_(true) - A flow registration is constructed only at arrival
  //  tail_(false) - There are no messages of size 0
  //  ejection_(false) - Only needs to be set true once; will be checked
  //  injection_(falst) - Same as ejection
  //  has_vc_(false) - Assumed vc-less until proven otherwise
  sharedbuffer::flowregistration::flowregistration()
    : head_(true), tail_(false), ejection_(false), injection_(false),
      has_vc_(false)
  {}

  // Construct me.
  sharedbuffer::flowregistration::ptr
  sharedbuffer::flowregistration::construct()
  {
    return flowregistration::ptr
      (new flowregistration());
  }

  // Destruct me.
  sharedbuffer::flowregistration::~flowregistration() throw()
  {}

} } //end of namespace sstmac

