/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_HARDWARE_NETWORK_CONGESTION_SHAREDBUFFER_FLOWREGISTRATION_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_CONGESTION_SHAREDBUFFER_FLOWREGISTRATION_H_INCLUDED

#include <sstmac/hardware/network/congestion/sharedbuffer.h>
#include <sstmac/common/eventhandler.h>

namespace sstmac { namespace hw  {

/**
 * Contains the state information about a particular flow at a particular vertex.
 */
class sharedbuffer::flowregistration : public ptr_type {

public:
        /// The time this flow is expected to depart.
        timestamp expected_departure_time_;
        /// Are we the tail?
        bool tail_;
        /// Are we the head?
        bool head_;
        /// Edge we rode in on.
        edge_index input_edge_;
        /// Edge we're taking out of here.
        edge_index output_edge_;
        /// Flow rate into this vertex.
        double input_bw_;
        /// Flow rate out of this vertex.
        double output_bw_;
        /// How many bytes this flow occupies in the buffer.
        double buffer_occupancy_;
        /// Last time we updated the buffer occupancy.
        timestamp last_buffer_update_time_;
        /// Is this the ejection node for this flow?
        bool ejection_;
        /// Is this the injection node for this flow?
        bool injection_;
        /// Have we been assigned an output virtual channel?
        bool has_vc_;

protected:
        /// Construct a new flow registration.
        flowregistration();

public:

        typedef boost::intrusive_ptr<flowregistration> ptr;

        /// Construct a new flow registration.
        static flowregistration::ptr
        construct();

        /// Destructor.
        virtual ~flowregistration() throw ();

        virtual std::string to_string() const {
                return "sharedbuffer::flowregistration";
        }

};
} } //end of namespace sstmac

#endif
