/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */


// tbird.h: Class for the tbird network.
//
// Author: Curtis Janssen <cljanss@ca.sandia.gov>

#ifndef SSTMAC_HARDWARE_NETWORK_TOPOLOGY_TBIRD_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_TOPOLOGY_TBIRD_H_INCLUDED

#include <sstmac/hardware/network/topology/topology.h>

namespace sstmac { namespace hw  {

  /** The tbird network generates an oversubscribed fat tree network like
      that used in the tbird machine. */
  class tbird: public topology {
    void init_tbird();

  protected:
    //tbird(TiXmlHandle);
    tbird();

  public:
    typedef boost::intrusive_ptr<tbird> ptr;

    static tbird::ptr construct();
    virtual ~tbird();

    void shortest_path(const vertex&v1, const vertex&v2, walk&w) const;
    void route(const vertex &v1, const vertex &v2, walk &w) const;
    int64_t diameter() const;
    double a() const;
    int64_t b() const;
    bool vertex_symmetric() const;
    std::string name() const;
  };

} } //end of namespace sstmac

#endif
