/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/nic/simplenic.h>
//#include <sstmac/simulator/networkdata.h>
//#include <sstmac/process/recvhandler.h>
#include <sstmac/common/errors.h>

namespace sstmac
{
  namespace hw
  {

    // Pointer types.

    using namespace sstmac::hw;

    //
    // Hi.
    //
    simplenic::simplenic(const node::ptr &parent,
        const interconnect::ptr &interconn) :
        networkinterface(parent, interconn)

    {
    }

    //
    // Hello.
    //
    simplenic::ptr
    simplenic::construct(const node::ptr &parent,
        const interconnect::ptr &interconn)
    {
      return simplenic::ptr(new simplenic(parent, interconn));
    }

    //
    // Goodbye.
    //
    simplenic::~simplenic() throw ()
    {
    }

    //
    // Schedule a send operation to happen at the given time.
    //
    void
    simplenic::send(const nodeaddress::ptr &recver,
        const sst_message::ptr &payload, const eventhandler::ptr &completion)
    {
      use_me();

      SSTMAC_DEBUG << "simplenic::send(" << recver->to_string() << ", "
            << payload->to_string() << ", " << completion << ")\n";

      if (payload == 0)
        {
          std::cout << "I'm here throwing this payload error" << std::endl;
          throw nullerror("simplenic::send: null payload pointer.");
        }

      nodeaddress::ptr raw = recver;

      if (node_map_.find(raw) == node_map_.end())
        {
          std::cout << "I'm here throwing this node map error" << std::endl;
          std::cout << "looking for: " << raw << " " << raw->to_string()
              << std::endl;
          std::cout << "size of node_map: " << node_map_.size() << std::endl;
          throw ssterror(
              "simplenic::send: can't find receiver nic in node map");
        }

      networkinterface::ptr otherguy = node_map_[raw];
      SSTMAC_DEBUG << "simplenic:send - got receiver address: "
            << raw->to_string() << ", interface: "
            << otherguy->get_parent_id()->to_string() << "\n";

      if (otherguy == 0)
        {
          std::cout << "I'm here throwing this null recver error" << std::endl;
          throw nullerror("simplenic::send - null receiving nic");
        }

      timestamp delayed = now(); //don't model delay for now
      eventhandler::ptr casted = boost::dynamic_pointer_cast<eventhandler>(
          otherguy);

      SSTMAC_DEBUG << "simplenic:send - sending now \n";

      interconn_->send(delayed, payload, completion, casted);

      spy_->add_one(payload->fromaddr_->unique_id(),
          payload->toaddr_->unique_id());
      spy_bytes_->add(payload->fromaddr_->unique_id(),
                payload->toaddr_->unique_id(), payload->get_byte_length());
    }

    networkinterface::ptr
    simplenic::clone(const node::ptr &parent,
        const sstmac::hw::interconnect::ptr &interconn) const
    {
      return simplenic::construct(parent, interconn);
    }

    void
    simplenic::handle(const sst_message::ptr &msg)
    {
      SSTMAC_DEBUG << "simplenic::handle - received message \n";

      parent_->handle(msg);
    }

  }
} // end of namespace sstmac.
