/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/node/degradednode.h>

#include <sstmac/hardware/processor/simpleprocessor-config.h>
#include <sstmac/hardware/processor/instructionprocessor-config.h>
#include <sstmac/hardware/node/simplenode.h>
#include <sstmac/hardware/memory/simple_memorymodel.h>
#include <sstmac/hardware/processor/migratingcomputescheduler.h>
#include <sstmac/common/errors.h>
#include <sstmac/software/services/launch/machine_descriptor.h>

namespace sstmac
{
  namespace hw
  {

    void
    degradednode::initialize(nodeaddress::ptr my_id, eventmanager::ptr man)
    {
      node::initialize(my_id, man);

      double tcycle = 1.0 / (freq_ / factor_);

      timestamp tflop(tcycle * 10); // 10 clock cycles for a floating point
      timestamp tmemop(
          memory_latency_.sec() + 1.0 / (memory_bandwidth_ / factor_));

      sstmac::hw::simpleprocessor_config::ptr cfg =
          sstmac::hw::simpleprocessor_config::construct(cores_, freq_, 10, 30);

      compute_sched_ = sstmac::hw::migratingcomputescheduler::construct(my_id->unique_id(), cores_,
          preempt_);
      compute_sched_->set_eventmanager(man);
      proc_model_ = sstmac::hw::simpleprocessor::construct(cfg, msg_handler_,
          compute_sched_);

      sstmac::hw::instructionprocessor_config::ptr pcfg =
          sstmac::hw::instructionprocessor_config::construct(cores_,
              INST_FLOP_LABEL, tflop, INST_MEMOP_LABEL, tmemop,
              INST_MEMOP_LOCAL_LABEL, tmemop, INST_MEMOP_RAND_LABEL, tmemop);

      proc_inst_ = sstmac::hw::instructionprocessor::construct(pcfg,
          msg_handler_, compute_sched_);

      mem_model_ = sstmac::hw::simple_memorymodel::construct(memory_bandwidth_,
          memory_latency_, msg_handler_, my_id->unique_id());
      mem_model_->set_eventmanager(man);

      if (params_ && params_->has_param("disk_type"))
        {
          if (params_->get_param("disk_type") == "vesper")
            {
              disk_ = sstmac::hw::disk_controller::construct(my_id->unique_id(), params_,  msg_handler_, man);
            }
        }

      os_->start_apps();

      launcher_->start();

    }

  }
}
