/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_HARDWARE_PROCESSOR_PROCESSOR_H_INCLUDED
#define SSTMAC_HARDWARE_PROCESSOR_PROCESSOR_H_INCLUDED


#include <sstmac/common/eventmodule.h>

#include <sstmac/common/ptr_type.h>
#include <sstmac/hardware/processor/processor_config.h>

#include <sstmac/common/messages/sst_message.h>

#include <sstmac/common/debug.h>
#include <list>
#include <queue>

#include <sstmac/common/stats/stat_time_window.h>
#include <sstmac/common/stats/stats_common.h>

#include <sstmac/hardware/processor/computescheduler.h>

namespace sstmac { namespace hw 
  {

  /**
   * An interface for processor models
   */
  class processor :  public eventhandler
  {
  public:

    typedef boost::intrusive_ptr<processor> ptr;

    /// Goodbye.
    virtual
    ~processor()
    {
    }

    /// Get a new processor model of the same type as the current one.
    virtual processor::ptr
    clone() const = 0;

    /// Process a compute kernel.
    /// \param inst instructions object that is processed.
    /// \return timestamp time to process instructions.
    virtual void
    compute(const sst_message::ptr& inst, long thread_id) = 0;

    processor_config::ptr
    getProcCfg();



    virtual void
    handle(const sst_message::ptr &msg);

    virtual std::string to_string() const {
      return "processor";
    }

  protected:
    processor(processor_config::ptr cfg,
        const eventhandler::ptr &completion,
        computescheduler::ptr sched);

    static debug dbg_;
    processor_config::ptr proc_cfg;

    // stats
    static stat_time_window_sum::ptr stat_proc_;
    static stat_time_window_sum_duration::ptr stat_energy_;
    static stat_sum::ptr stat_pwr_;


    eventhandler::ptr completion_;

    void
    request_compute(const timestamp& comp,
        const sst_message::ptr& to_deliver, long thread_id);


    computescheduler::ptr csched_;
  };

  } } // end of namespace sstmac

#endif
