/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_HARDWARE_PROCESSOR_SIMPLEPROCESSOR_H_INCLUDED
#define SSTMAC_HARDWARE_PROCESSOR_SIMPLEPROCESSOR_H_INCLUDED

#include <sstmac/hardware/processor/processor.h>
//#include <sstmac/simulator/ppid.h>
#include <sstmac/common/timestamp.h>
//#include <sstmac/simulator/eventhandler.h>
#include <sstmac/common/ptr_type.h>
#include <sstmac/hardware/processor/simpleprocessor-config.h>

//#include <sstmac/simulator/power_config.h>
//#include <sstmac/simulator/powermanager.h>

namespace sstmac { namespace hw 
  {


  /**
   * A very simple processor model
   */
  class simpleprocessor : public processor, virtual public nocopy
  {

  private:

    simpleprocessor_config::ptr config_;



  protected:

    /// Protected constructor.
    simpleprocessor(const simpleprocessor_config::ptr& cfg,
        const eventhandler::ptr &completion,
        computescheduler::ptr sched);

  public:
    typedef boost::intrusive_ptr<processor> ptr;

    /// Goodbye.
    virtual
    ~simpleprocessor()
    {
    }

    /// Create simple processor model.
    /// \Param cfg model configuration object
    static simpleprocessor::ptr
    construct(const simpleprocessor_config::ptr& cfg,
        const eventhandler::ptr &completion,
        computescheduler::ptr sched);

    /// Inherited from processor.
    processor::ptr
    clone() const;

    /// Process a compute kernel.
    /// \param inst is the instructions object that will be run.
    /// \return timestamp time to process instructions.
    void
    compute(const sst_message::ptr& inst, long thread_id);

    virtual std::string to_string() const {
      return "simpleprocessor";
    }

  };

  } } // end of namespace sstmac

#endif
