/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/libraries/actor/actorid.h>
#include <sstmac/common/errors.h>
#include <iostream>

namespace sstmac { namespace sw {

  //
  // Create a new valid actorid.
  //
  actorid::actorid(const nodeaddress::ptr node, int32_t rank, appid aid) :
    node_(node), rank_(rank), aid_(aid)
  {
    if(rank_ < 0)
      throw valueerror("actorid(...):  Not a valid id.");
  }
  
  //
  // Create a blank (invalid) actorid.
  //
  actorid::actorid() :
    node_(), rank_(-1), aid_(-1)
  {}

  //
  // Goodbye.
  //
  actorid::~actorid() {
  }

  //
  // Test whether this is a valid actor identifier for sends 
  //
  bool actorid::is_real() const {
    return (rank_ >= 0);
  }

  //
  // Test whether this actorid matches the given identifier.
  //
  bool actorid::matches(const actorid &other) const {
    if(rank_ < -1) // wildcard
      return true;
    else
      return ((node_ == other.node_) && (rank_ == other.rank_) && (aid_ == other.aid_));
  }

  //
  // An invalid actor identifier.  This is the same as calling actorid().
  actorid actorid::invalid() {
    return actorid();
  }

  //
  // A wildcard match for any actor identifier.
  //
  actorid actorid::any() {
    actorid id;
    id.rank_ = -2;
    return id;
  }

  //
  // Utility.
  //
  std::ostream& operator<<(std::ostream &os, const actorid &id) {
    os << id.to_string();
    return os;
  }


} } // end of namespace sstmac.
