/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_CHECKPOINT_RESTARTABLE_APP_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_CHECKPOINT_RESTARTABLE_APP_H_INCLUDED

#include <sstmac/common/ptr_type.h>
#include <checkpoint/lib_checkpoint.h>

namespace sstmac
{
  namespace sw
  {

    class restartable_app : virtual public ptr_type
    {

    protected:

      lib_checkpoint::ptr checklib_;



    public:
      typedef boost::intrusive_ptr<restartable_app> ptr;

      virtual std::string
      to_string() const
      {
        return "restartable_app";
      }

      virtual void
      init_restartable(OS_API::ptr os, softwareid id,
          sim_parameters::ptr params)
      {

        checklib_ = lib_checkpoint::construct(id);
        os->register_lib(checklib_);


      }

      virtual void
      init_restartable(lib_checkpoint::ptr lib)
      {
        checklib_ = lib;

      }

      virtual lib_checkpoint::ptr
      get_check_lib()
      {
        return checklib_;
      }
      //----- wrap these for convenience ----
    protected:
      virtual void
      checkpoint()
      {
        checklib_->checkpoint();
      }

      virtual void
      finalize_checkpoint()
      {
        checklib_->finalize_checkpoint();
      }

      template<typename Type>
        void
        register_var(std::string n, Type* val)
        {

          checklib_->register_var<Type> (n, val);
        }

      template<typename Type>
        void
        register_fake(std::string n, size_t size)
        {

          checklib_->register_fake<Type> (n, size);
        }

      virtual bool
      was_restarted()
      {
        return checklib_->restarted();
      }

    };
  }
}
#endif
