/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_COMPUTE_LIB_COMPUTE_LOOPS_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_COMPUTE_LIB_COMPUTE_LOOPS_H_INCLUDED

#include <sstmac/software/libraries/compute/lib_compute_inst.h>
#include <sstmac/software/libraries/compute/lib_compute_memmove.h>

namespace sstmac
{
  namespace sw
  {

    class lib_compute_loops : public lib_compute_inst,
        public lib_compute_memmove
    {
      // ------- constructor / boost stuff -------------//

    protected:
      lib_compute_loops(softwareid id) :
          lib_compute_inst(id), lib_compute_memmove(id)
      {
        libname_ = "computelibgtc" + id.to_string();
      }

      lib_compute_loops(std::string id) :
          lib_compute_inst(id), lib_compute_memmove(id)
      {
        libname_ = id;
      }

      bool doing_memory_;

    public:
      typedef boost::intrusive_ptr<lib_compute_loops> ptr;

      virtual
      ~lib_compute_loops()
      {
      }

      static ptr
      construct(softwareid id)
      {
        return ptr(new lib_compute_loops(id));
      }

      static ptr
      construct(std::string id)
      {
        return ptr(new lib_compute_loops(id));
      }

      template<typename Type>
        void
        compute_array_init(long arraysize)
        {
          doing_memory_ = true;
          lib_compute_memmove::compute_block_access(
              arraysize * sizeof(Type));
          doing_memory_ = false;
        }

      /**
       * compute_loop - compute for a loop
       * @param from
       * @param to
       * @param numlines
       */
      void
      compute_loop(int from, int to, int numlines);

      /**
       * compute_loop2 - compute for one loop nested in an outer one
       * @param from1
       * @param to1
       * @param from2
       * @param to2
       * @param numlines
       */
      void
      compute_loop2(int from1, int to1, int from2, int to2, int numlines);

      /**
       * compute_loop3 - compute for one loop nested in loop nested in an outer one
       * @param from1
       * @param to1
       * @param from2
       * @param to2
       * @param from3
       * @param to3
       * @param numlines
       */
      void
      compute_loop3(int from1, int to1, int from2, int to2, int from3, int to3,
          int numlines);

      /**
       * compute_loop4 - compute for quadruple nested loop

       * @param numlines
       */
      void
      compute_loop4(int from1, int to1, int from2, int to2, int from3, int to3,
          int from4, int to4, int numlines);

      void
      compute_fft();

      static int mem_op_size_;
      static double mem_line_ratio_;
      static double flop_line_ratio_;

    protected:

      virtual bool
      supported()
      {
        return lib_compute_inst::supported() && lib_compute_memmove::supported();
      }

    };

  } //end of namespace sstmac
}

#endif
