/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/software/libraries/compute/lib_sleep.h>

#include <sstmac/software/process/key.h>

namespace sstmac { namespace sw
{

  debug lib_sleep::dbg_;


  void
  lib_sleep::sleep(timestamp time)
  {
    if (supported())
      {
        sst_message::ptr cmsg = sst_message::construct(
            get_lib_name(), time);

        SSTMAC_DEBUG << "lib_sleep: executing compute kernel \n";
        os_->execute_kernel(AMI::AMI_COMPUTE, AMI::AMI_COMP_SLEEP, cmsg);

        SSTMAC_DEBUG << "lib_sleep: blocking \n";
        wait_this(cmsg);
      }
    else
      {
        throw valueerror(
            "lib_sleep requires the presence of a processor model which can sleep");
      }
  }

} } //end of namespace sstmac
