/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_COMPUTE_MEM_ACCESS_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_COMPUTE_MEM_ACCESS_H_INCLUDED

#include <sstmac/common/messages/sst_message.h>

namespace sstmac
{
  namespace sw
  {

    class mem_access : public sst_message
    {
    protected:
      mem_access(long bytes, std::string libn) :
        sst_message(libn, timestamp(0), sst_message::COMPUTE, 0),
            bytes_(bytes), srcthread_(-1), dstthread_(-1)
      {

      }

      mem_access(long bytes, std::string libn, long srcth, long destth) :
        sst_message(libn, timestamp(0), sst_message::COMPUTE, 0),
            bytes_(bytes), srcthread_(srcth), dstthread_(destth)
      {

      }

      long bytes_;

    public:
      typedef boost::intrusive_ptr<mem_access> ptr;

      static ptr
      construct(long bytes, std::string libn)
      {
        return ptr(new mem_access(bytes, libn));
      }

      static ptr
      construct(long bytes, std::string libn, long srcth, long dstth)
      {
        return ptr(new mem_access(bytes, libn, srcth, dstth));
      }

      virtual long
      get_byte_length() const
      {
        return bytes_;
      }
      long uid_;
      long srcthread_;
      long dstthread_;

    };

  }
}

#endif
