/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <string>
#include <iostream>

#include <sstmac/software/process/app.h>
#include <sstmac/common/messages/eventdata.h>

using namespace std;
using namespace sstmac;
using namespace sstmac::sw;

  app::ptr
  get_current_app()
  {
    thread::ptr t = OS_API::get_current_thread();
    app::ptr app;
    try
      {
        app = boost::dynamic_pointer_cast<sstmac::sw::app>(t);
      }
    catch (std::exception& e)
      {
        std::cerr << e.what() << std::endl;
        return app::ptr();
      }

    if (!app)
      {
        throw sstmac::ssterror(
            "sstcpp::get_current_app - app didn't cast");
      }

    return app;
  }

  extern "C" void
  SSTMAC_compute(double seconds)
  {
    app::ptr app = get_current_app();
    app->compute( timestamp(seconds) );
  }

  extern "C" void
  SSTMAC_compute_block_access(long long bytes)
  {
    app::ptr app = get_current_app();
    app->compute_block_access( bytes );
  }
