/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <math.h>
#include <sstmac/software/libraries/filesystem/local_file_system.h>
#include <sstmac/hardware/disk/vesper/disk.h>
#include <sstmac/software/libraries/filesystem/disk_message.h>

namespace sstmac
{
  namespace sw
  {

    bool
    local_file_system::supported()
    {
      return os_->kernel_supported(AMI::AMI_COMPUTE, AMI::AMI_COMP_DISKACCESS);
    }

    openfile::ptr
    local_file_system::open(std::string name)
    {
      /*if (supported())
        {
          //For now, it's a random sector.  Not doing a real file system with a directory
          //  to look things up
          int sec = rand() % sstmac::hw::Disk::logicalBlocks;
          disk_message::ptr dm = disk_message::construct(get_lib_name(), disk_message::OPEN, 0,
              sec, 1, 0);

          os_->execute_kernel(AMI::AMI_COMPUTE, AMI::AMI_COMP_DISKACCESS, dm);

          wait();

          openfile::ptr ret = openfile::construct(name, sec);

          return ret;
        }
      return openfile::ptr();*/
    }

    void
    local_file_system::read(openfile::ptr file, long bytes)
    {
      if (!file->is_open())
        {
          throw valueerror(
              "local_file_system: trying to read from a closed file");
        }
      int sec = file->get_sector() + file->get_offset_sectors();
      int numsec = ceil((double)bytes / (sstmac::hw::Disk::blockSize * 1000000.0));
      disk_message::ptr dm = disk_message::construct(get_lib_name(), disk_message::READ, bytes,
          sec, numsec, 0);

      os_->execute_kernel(AMI::AMI_COMPUTE, AMI::AMI_COMP_DISKACCESS, dm);

      wait_this(dm);
    }
    void
    local_file_system::write(openfile::ptr file, long bytes)
    {
      if (!file->is_open())
        {
          throw valueerror(
              "local_file_system: trying to write to a closed file");
        }
      int sec = file->get_sector() + file->get_offset_sectors();
      int numsec = ceil((double)bytes / (sstmac::hw::Disk::blockSize * 1000000.0));
      disk_message::ptr dm = disk_message::construct(get_lib_name(), disk_message::WRITE, bytes,
          sec, numsec, 0);

      os_->execute_kernel(AMI::AMI_COMPUTE, AMI::AMI_COMP_DISKACCESS, dm);

      wait_this(dm);
    }

    void
    local_file_system::seek(openfile::ptr file, long bytes)
    {
      file->seek(bytes);
    }

    void
    local_file_system::close(openfile::ptr file)
    {
      int sec = file->get_sector();
      disk_message::ptr dm = disk_message::construct(get_lib_name(), disk_message::CLOSE, 0,
          sec, 1, 0);
      os_->execute_kernel(AMI::AMI_COMPUTE, AMI::AMI_COMP_DISKACCESS, dm);

      wait_this(dm);

      file->close();
    }

  }
} /* namespace sstmac */
