/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_HPX_AGAS_AGAS_SERVER_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_HPX_AGAS_AGAS_SERVER_H_INCLUDED

#include <sstmac/software/services/service.h>
#include <sstmac/software/libraries/hpx/agas/globalid.h>
#include <sstmac/software/libraries/hpx/agas/lva.h>
#include <hpx/parcels/parcel.h>
#include <compute/lib_compute_memmove.h>
#include <map>
#include <queue>

namespace hpx
{

  class agas_server : public sstmac::sw::service
  {

#define MAX_OPS 1
#define OP_SIZE 64

  protected:
    unsigned long long gid_count_;
    sstmac::nodeaddress::ptr locality_;

    long WAITING_ID;

#ifdef SSTMAC_HAVE_BOOST_1_37
    boost::unordered_map<unsigned long long, lva> lvamap_;
    boost::unordered_map<long, sstmac::sw::key::ptr> waiting_;
#else
    std::map<unsigned long long, lva> lvamap_;
    std::map<long, sstmac::sw::key::ptr> waiting_;
#endif

    hpx::parcels::parcel::ptr rx_;

    std::queue<hpx::parcels::parcel::ptr> pending_;

    sstmac::sw::lib_compute_memmove::ptr lib_mem_;
    int computing_;

    static sstmac::debug dbg_;

  public:
    typedef boost::intrusive_ptr<agas_server> ptr;

    static ptr
    construct()
    {
      return ptr(new agas_server());
    }

    virtual
    ~agas_server()
    {
    }

    virtual void
    handle(const sstmac::sst_message::ptr &msg);

    virtual service::ptr
    clone()
    {
      return construct();
    }

    virtual void
    start()
    {
      service::start();

      lib_mem_ = sstmac::sw::lib_compute_memmove::construct(
          "agas-server-lib-mem");
      lib_mem_->init_handler(self_);
      lib_mem_->init_waiter(wait_);
      register_lib(lib_mem_);

      locality_ = os_->get_my_addr();
      gid_count_ = locality_->unique_id() * globalid::get_quantum();
    }

    static std::string
    lib_name()
    {
      return "hpx_agas_server";
    }

    // ------ these are the useful functions ------ //

    globalid
    new_gid(lva::ENTITY_TYPE t);

    sstmac::nodeaddress::ptr
    resolve_locality(const globalid &gid);

    // -------------------------------------- //

  protected:
    agas_server() :
        computing_(0), WAITING_ID(0)
    {

      if (!dbg_.is_real())
        {
          dbg_ = sstmac::debug("<debug> hpx | hpx_agas_server",
              sstmac::logger::blue);
          sstmac::logger::register_logger(&dbg_);
        }

      libname_ = lib_name();
    }

   // void
  //  service_one();

  };

}

#endif
