/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_HPX_AGAS_GLOBALID_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_HPX_AGAS_GLOBALID_H_INCLUDED

#include <sstmac/common/logger.h>

#include <iostream>
#include <sstream>
#include <math.h>

namespace hpx
{

  struct globalid
  {
  public:

    unsigned long long gid_;
#define numbits sizeof(long long)*8

  protected:

    static long num_localities_;
    static long long quantum_;

  public:

    static long
    get_quantum()
    {
      return quantum_;
    }

    bool
    matches(const globalid &other) const;

    std::string
    to_string() const;

    static void
    set_num_localities(long num)
    {
      num_localities_ = num;
      double max = pow(2, numbits);
      quantum_ = max / num_localities_;

    }

    long
    get_node_uid() const
    {
      if (quantum_ <= 0)
        {
          throw sstmac::ssterror(
              "globalid::get_node_uid() - problem with the quantum");
        }
      return gid_ / (quantum_);
    }

    globalid(unsigned long long id) :
        gid_(id)
    {
    }

    globalid() :
        gid_(-1)
    {
    }

  };

  globalid&
  operator++(globalid &value);
  globalid
  operator++(globalid &value, int);
  bool
  operator==(const globalid &a, const globalid &b);
  bool
  operator!=(const globalid &a, const globalid &b);
  bool
  operator<(const globalid &a, const globalid &b);
  bool
  operator>(const globalid &a, const globalid &b);
  bool
  operator<=(const globalid &a, const globalid &b);
  bool
  operator>=(const globalid &a, const globalid &b);

  globalid
  operator+(const globalid &a, const globalid &b);

  std::ostream&
  operator<<(std::ostream &os, const globalid &id);

}

namespace boost
{
  size_t
  hash_value(const hpx::globalid &t);
}

#endif
