/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_HPX_AGAS_LVA_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_HPX_AGAS_LVA_H_INCLUDED

#include <sstmac/common/nodeaddress.h>
#include <sstmac/software/libraries/hpx/agas/globalid.h>

namespace hpx
{

  class lva
  {
  public:

    enum ENTITY_TYPE
    {
      APP, THREAD, COMPONENT
    };

    virtual std::string
    to_string() const
    {
      std::stringstream ss;
      ss << "lva(" << gid_ << ", " << addr_->to_string() << ", " << type_
          << ")";
      return ss.str();
    }



    lva(globalid id, sstmac::nodeaddress::ptr addr, ENTITY_TYPE t) :
        gid_(id), addr_(addr), type_(t)
    {

    }

    lva() : gid_(0) {

    }

    sstmac::nodeaddress::ptr
    addr(){
      return addr_;
    }

  protected:
    globalid gid_;
    sstmac::nodeaddress::ptr addr_;
    ENTITY_TYPE type_;
  };

}

#endif
