/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <hpx/hpx_app.h>

namespace hpx
{

  hpx_thread_manager::ptr
  hpx_app::get_current_thread_manager()
  {
    thread::ptr t = sstmac::sw::OS_API::get_current_thread();
    hpx_app::ptr app = boost::dynamic_pointer_cast<hpx_app>(t);

    if (!app)
      {
        boost::intrusive_ptr<hpx_thread> pxth = boost::dynamic_pointer_cast<
            hpx_thread>(t);

        if (!pxth)
          {

            throw sstmac::ssterror(
                "hpx_app::get_current_thread_manager() - app didn't cast, make sure app is an hpx_app or an hpx_thread");
          }
        else
          {
            return pxth->get_thread_man();
          }
      }
    else
      {

        return app->threadman_;
      }
  }

  agas_server::ptr
  hpx_app::get_current_agas_server()
  {
    thread::ptr t = sstmac::sw::OS_API::get_current_thread();
    hpx_app::ptr app = boost::dynamic_pointer_cast<hpx_app>(t);

    if (!app)
      {
        boost::intrusive_ptr<hpx_thread> pxth = boost::dynamic_pointer_cast<
            hpx_thread>(t);

        if (!pxth)
          {

            throw sstmac::ssterror(
                "hpx_app::get_current_thread_manager() - app didn't cast, make sure app is an hpx_app or an hpx_thread");
          }
        else
          {
            return pxth->get_agas_server();
          }
      }
    else
      {

        return app->agas_;
      }
  }

  void
  hpx_app::hpx_block(sstmac::sw::key::ptr blocker)
  {

    thread::ptr t = sstmac::sw::OS_API::get_current_thread();
    hpx_app::ptr app = boost::dynamic_pointer_cast<hpx_app>(t);

    if (!app)
      {
        boost::intrusive_ptr<hpx_thread> pxth = boost::dynamic_pointer_cast<
            hpx_thread>(t);

        if (!pxth)
          {

            throw sstmac::ssterror(
                "hpx_app::hpx_block() - app didn't cast, make sure app is an hpx_app or an hpx_thread");
          }
        else
          {
            pxth->hpx_block(blocker);
          }
      }
    else
      {
        app->os_->block(blocker);
      }

  }

  void
  hpx_app::hpx_thread_swap()
  {

    thread::ptr t = sstmac::sw::OS_API::get_current_thread();
    hpx_app::ptr app = boost::dynamic_pointer_cast<hpx_app>(t);

    if (!app)
      {
        boost::intrusive_ptr<hpx_thread> pxth = boost::dynamic_pointer_cast<
            hpx_thread>(t);

        if (!pxth)
          {

            throw sstmac::ssterror(
                "hpx_app::hpx_thread_swap() - app didn't cast, make sure app is an hpx_app or an hpx_thread");
          }
        else
          {
            pxth->compute(sstmac::timestamp(hpx_thread::THREAD_SWAP_OVERHEAD));
          }
      }
    else
      {
        app->compute(sstmac::timestamp(hpx_thread::THREAD_SWAP_OVERHEAD));

      }

  }

  void
  hpx_app::hpx_unblock(sstmac::sw::key::ptr blocker)
  {
    thread::ptr t = sstmac::sw::OS_API::get_current_thread();
    hpx_app::ptr app = boost::dynamic_pointer_cast<hpx_app>(t);

    if (!app)
      {
        boost::intrusive_ptr<hpx_thread> pxth = boost::dynamic_pointer_cast<
            hpx_thread>(t);

        if (!pxth)
          {

            throw sstmac::ssterror(
                "hpx_app::hpx_unblock() - app didn't cast, make sure app is an hpx_app or an hpx_thread");
          }
        else
          {
            pxth->hpx_unblock(blocker);
          }
      }
    else
      {
        app->os_->unblock(blocker);
      }

  }

  void
  hpx_app::hpx_send_msg(sstmac::sst_message::ptr msg)
  {
    thread::ptr t = sstmac::sw::OS_API::get_current_thread();
    hpx_app::ptr app = boost::dynamic_pointer_cast<hpx_app>(t);

    if (!app)
      {
        boost::intrusive_ptr<hpx_thread> pxth = boost::dynamic_pointer_cast<
            hpx_thread>(t);

        if (!pxth)
          {

            throw sstmac::ssterror(
                "hpx_app::hpx_send_msg() - app didn't cast, make sure app is an hpx_app or an hpx_thread");
          }
        else
          {
            pxth->hpx_send(msg);
          }
      }
    else
      {
        app->os_->execute_kernel(sstmac::sw::AMI::AMI_COMM,
            sstmac::sw::AMI::AMI_COMM_SEND, msg);
      }
  }

  sstmac::nodeaddress::ptr
  hpx_app::gbl_find_here()
  {
    thread::ptr t = sstmac::sw::OS_API::get_current_thread();
    hpx_app::ptr app = boost::dynamic_pointer_cast<hpx_app>(t);

    if (!app)
      {
        boost::intrusive_ptr<hpx_thread> pxth = boost::dynamic_pointer_cast<
            hpx_thread>(t);

        if (!pxth)
          {

            throw sstmac::ssterror(
                "hpx_app::hpx_send_msg() - app didn't cast, make sure app is an hpx_app or an hpx_thread");
          }
        else
          {
            return pxth->find_here();
          }
      }
    else
      {
        return app->find_here();
      }
  }

  sstmac::sw::environment::const_ptr
  hpx_app::get_env()
  {
    thread::ptr t = sstmac::sw::OS_API::get_current_thread();
    hpx_app::ptr app = boost::dynamic_pointer_cast<hpx_app>(t);

    if (!app)
      {
        boost::intrusive_ptr<hpx_thread> pxth = boost::dynamic_pointer_cast<
            hpx_thread>(t);

        if (!pxth)
          {

            throw sstmac::ssterror(
                "hpx_app::hpx_send_msg() - app didn't cast, make sure app is an hpx_app or an hpx_thread");
          }
        else
          {
            return pxth->get_env(app->id_.app_);
          }
      }
    else
      {
        return app->os_->get_env(app->id_.app_);
      }
  }

  sstmac::sw::OS_API::ptr
  hpx_app::get_os()
  {
    thread::ptr t = sstmac::sw::OS_API::get_current_thread();
    hpx_app::ptr app = boost::dynamic_pointer_cast<hpx_app>(t);

    if (!app)
      {
        boost::intrusive_ptr<hpx_thread> pxth = boost::dynamic_pointer_cast<
            hpx_thread>(t);

        if (!pxth)
          {

            throw sstmac::ssterror(
                "hpx_app::get_os() - app didn't cast, make sure app is an hpx_app or an hpx_thread");
          }
        else
          {
            return pxth->get_os();
          }
      }
    else
      {
        return app->os_;
      }
  }

  void
  hpx_app::compute(const sstmac::timestamp &ti)
  {
    thread::ptr t = sstmac::sw::OS_API::get_current_thread();
    hpx_app::ptr app = boost::dynamic_pointer_cast<hpx_app>(t);

    if (!app)
      {
        boost::intrusive_ptr<hpx_thread> pxth = boost::dynamic_pointer_cast<
            hpx_thread>(t);

        if (!pxth)
          {

            throw sstmac::ssterror(
                "hpx_app::get_os() - app didn't cast, make sure app is an hpx_app or an hpx_thread");
          }
        else
          {
            return pxth->compute(ti);
          }
      }
    else
      {
        app->app::compute(ti);
      }
  }

  void
  hpx_app::hpx_register_lib(sstmac::sw::library::ptr lib)
  {
    thread::ptr t = sstmac::sw::OS_API::get_current_thread();
    hpx_app::ptr app = boost::dynamic_pointer_cast<hpx_app>(t);

    if (!app)
      {
        boost::intrusive_ptr<hpx_thread> pxth = boost::dynamic_pointer_cast<
            hpx_thread>(t);

        if (!pxth)
          {

            throw sstmac::ssterror(
                "hpx_app::get_os() - app didn't cast, make sure app is an hpx_app or an hpx_thread");
          }
        else
          {
            return pxth->register_lib(lib);
          }
      }
    else
      {
        app->app::register_lib(lib);
      }
  }
}
