/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_HPX_LCOS_EAGER_FUTURE_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_HPX_LCOS_EAGER_FUTURE_H_INCLUDED

#include <hpx/lcos/promise.h>
#include <hpx/runtime/hpx_thread.h>
#include <hpx/runtime/actions/hpx_action_params.h>
#include <hpx/runtime/hpx_thread_manager.h>
#include <hpx/hpx_app.h>
namespace hpx
{
  namespace lcos
  {

    template<typename Result>
      class eager_future : public promise<Result>
      {
      public:
        typedef boost::intrusive_ptr<eager_future<Result> > ptr;

      protected:
        ptr self_;

        eager_future(std::string action_name, sstmac::nodeaddress::ptr loc,
            hpx_action_params::ptr params,
            boost::intrusive_ptr<hpx::parcels::depayloadizer<Result> > de) :
            promise<Result>(de)
        {
          self_ = ptr(this);
          hpx_thread_manager::ptr tman = hpx_app::get_current_thread_manager();

          tman->start_action(action_name, params, self_, loc);

        }

        eager_future(std::string action_name, globalid gid,
            hpx_action_params::ptr params,
            boost::intrusive_ptr<hpx::parcels::depayloadizer<Result> > de) :
            promise<Result>(de)
        {
          self_ = ptr(this);
          hpx_thread_manager::ptr tman = hpx_app::get_current_thread_manager();

          tman->start_action(action_name, params, self_, gid);

        }

      public:

        static ptr
        construct(std::string action_name, sstmac::nodeaddress::ptr loc,
            hpx_action_params::ptr params,
            boost::intrusive_ptr<hpx::parcels::depayloadizer<Result> > de =
                hpx::parcels::depayloadizer<Result>::construct())
        {
          return ptr(new eager_future(action_name, loc, params, de));
        }

        static ptr
        construct(std::string action_name, globalid gid,
            hpx_action_params::ptr params,
            boost::intrusive_ptr<hpx::parcels::depayloadizer<Result> > de =
                hpx::parcels::depayloadizer<Result>::construct())
        {
          return ptr(new eager_future(action_name, gid, params, de));
        }

        virtual
        ~eager_future()
        {
        }

      };

  }
}

#endif
