/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_HPX_LCOS_REMOTE_CALLBACK_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_HPX_LCOS_REMOTE_CALLBACK_H_INCLUDED

#include <hpx/lcos/base_lco_result.h>
#include <hpx/lcos/promise.h>
#include <hpx/hpx_app.h>
#include <hpx/parcels/remote_action_result.h>
#include <hpx/runtime/hpx_thread_manager.h>

namespace hpx
{
  namespace lcos
  {

    /**
     * Mutex in SST/macro is actually quite simple, since only one thread is ever
     * active at a time, there's never a chance for real resource contention.
     */
    class remote_callback : public base_lco_result
    {

    protected:
      long transid_;
      sstmac::nodeaddress::ptr backto_;

     /* remote_callback(long id, sstmac::nodeaddress::ptr a,
          sstmac::sw::OS_API::ptr os) :
          base_lco_result(os), transid_(id), backto_(a)
      {

      }*/

      remote_callback(long id, sstmac::nodeaddress::ptr a) :
          transid_(id), backto_(a)
      {

      }
    public:

      typedef boost::intrusive_ptr<remote_callback> ptr;

      static ptr
      construct(long transid, sstmac::nodeaddress::ptr a)
      {
        return ptr(new remote_callback(transid, a));
      }

      /*static ptr
      construct(long transid, sstmac::nodeaddress::ptr a,
          sstmac::sw::OS_API::ptr os)
      {
        return ptr(new remote_callback(transid, a, os));
      }*/

      virtual void
      set_value(sstmac::payload::const_ptr load, sstmac::sw::OS_API::ptr os)
      {
        hpx::parcels::remote_action_result::ptr msg =
            hpx::parcels::remote_action_result::construct(load, transid_, os->now());
        msg->toaddr_ = backto_;
        msg->set_lib_name(hpx_thread_manager::lib_name());
        hpx_app::hpx_send_msg(msg);
      }

      virtual sstmac::payload::const_ptr
      get_value()
      {
        throw sstmac::ssterror(
            "hpx::lcos::remote_callback::get_value() -- this shouldn't really be called");
      }

      virtual
      ~remote_callback()
      {
      }

    };

  }
}

#endif
