/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_HPX_PARCELS_PARCEL_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_HPX_PARCELS_PARCEL_H_INCLUDED

#include <sstmac/common/messages/sst_message.h>
#include <sstmac/common/stats/stat_time_window.h>

namespace hpx { namespace parcels
{

  class parcel : public sstmac::sst_message
  {

  public:
    typedef boost::intrusive_ptr<parcel> ptr;

    enum PARCEL_TYPE {REMOTE_ACTION, REMOTE_ACTION_RESULT, RESOLVE_GID, RESOLVE_GID_RESPONSE};


    static ptr
    construct(PARCEL_TYPE t){
      return ptr(new parcel(t));
    }

    PARCEL_TYPE
    type() const {
      return type_;
    }

    virtual
    ~parcel() { }

  protected:
    parcel(PARCEL_TYPE t) : sst_message(sst_message::HPX), type_(t) {

    }

    PARCEL_TYPE type_;

    static std::map<PARCEL_TYPE, sstmac::stat_time_window_sum::ptr> stats_;

  };

} }
#endif
