/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_HPX_PARCELS_REMOTE_ACTION_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_HPX_PARCELS_REMOTE_ACTION_H_INCLUDED

#include <hpx/parcels/parcel.h>
#include <hpx/hpx_app.h>
namespace hpx
{

  namespace parcels
  {

    class remote_action : public parcel
    {

    public:
      typedef boost::intrusive_ptr<remote_action> ptr;

      static ptr
      construct(std::string t, hpx_action_params::ptr p, long trans,
          sstmac::timestamp now)
      {
        if (stats_[REMOTE_ACTION] == sstmac::stat_time_window_sum::ptr())
          {
            stats_[REMOTE_ACTION] = sstmac::stat_time_window_sum::construct(
                "<stats> hpx_parcels | hpx", "hpx_parcels_remote_action.csv");
          }
        stats_[REMOTE_ACTION]->record(now, 1);
        return ptr(new remote_action(t, p, trans));
      }

      static ptr
      construct(std::string t, hpx_action_params::ptr p, long trans,
          globalid gid, sstmac::timestamp now)
      {
        if (stats_[REMOTE_ACTION] == sstmac::stat_time_window_sum::ptr())
          {
            stats_[REMOTE_ACTION] = sstmac::stat_time_window_sum::construct(
                "<stats> hpx_parcels | hpx", "hpx_parcels_remote_action.csv");
          }
        stats_[REMOTE_ACTION]->record(now, 1);

        return ptr(new remote_action(t, p, trans, gid));
      }

      std::string
      which() const
      {
        return which_;
      }

      hpx::hpx_action_params::ptr
      params() const
      {
        return params_;
      }

      long
      trans_id() const
      {
        return trans_id_;
      }

      bool
      is_gid() const
      {
        return isgid_;
      }

      globalid
      gid() const
      {
        return gid_;
      }

      virtual long
      get_byte_length() const
      {
        // which + trans_id + gid + params
        return 4 + 8 + 16 + (params_ ? params_->get_byte_length() : 0);
      }

      virtual
      ~remote_action()
      {
      }

      virtual std::string
      to_string() const
      {
        return "hpx::parcels::remote_action";
      }

    protected:
      remote_action(std::string which, hpx_action_params::ptr p, long trans) :
          parcel(REMOTE_ACTION), which_(which), params_(p), trans_id_(trans), isgid_(
              false)
      {

      }

      remote_action(std::string which, hpx_action_params::ptr p, long trans,
          globalid gid) :
          parcel(REMOTE_ACTION), which_(which), params_(p), trans_id_(trans), gid_(
              gid), isgid_(true)
      {

      }

      std::string which_;
      hpx::hpx_action_params::ptr params_;
      long trans_id_;
      globalid gid_;
      bool isgid_;

    };

  }
}
#endif
