/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <hpx/runtime/hpx_action_thread.h>
#include <hpx/runtime/hpx_thread_manager.h>

namespace hpx
{

  long hpx_action_thread::COMPUTE_LIB_COUNT = 0;

  void
  hpx_action_thread::run()
  {
    SSTMAC_DEBUG << "hpx_action_thread -- computing creation overhead.. \n";
    std::stringstream ss;
    ss << "hpx_action_thread_compute_lib-" << COMPUTE_LIB_COUNT++;
    libcomp_ =
        sstmac::sw::lib_compute_time::construct(ss.str());
    os_->register_lib(libcomp_);
    libcomp_->compute(sstmac::timestamp(THREAD_CREATION_OVERHEAD + THREAD_SWAP_OVERHEAD));


    SSTMAC_DEBUG << "hpx_action_thread: running thread \n";
    sstmac::payload::const_ptr res;
    if (comp_)
      res = action_->act(comp_, params_);
    else
      res = action_->act(params_);

    lco_->set_value(res, os_);

    os_->unregister_lib(libcomp_);
  }
}
