/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_HPX_RUNTIME_HPX_ACTION_THREAD_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_HPX_RUNTIME_HPX_ACTION_THREAD_H_INCLUDED

#include <hpx/runtime/hpx_thread.h>
#include <hpx/runtime/actions/hpx_action.h>
#include <hpx/lcos/base_lco_result.h>
#include <hpx/components/hpx_component.h>


namespace hpx
{


  class hpx_action_thread : public hpx_thread
  {
  protected:
    typedef boost::intrusive_ptr<hpx::lcos::base_lco_result> lco_type;

    boost::intrusive_ptr<actions::hpx_action> action_;
    hpx_action_params::ptr params_;
    lco_type lco_;
    hpx::components::hpx_component::ptr comp_;



  public:
    typedef boost::intrusive_ptr<hpx_action_thread> ptr;

    static ptr
    construct(boost::intrusive_ptr<hpx_thread_manager> man,
        boost::intrusive_ptr<agas_server> agas,
        boost::intrusive_ptr<actions::hpx_action> act,
        hpx_action_params::ptr params, lco_type prom)
    {
      return ptr(new hpx_action_thread(man, agas, act, params, prom));
    }

    static ptr
    construct(boost::intrusive_ptr<hpx_thread_manager> man,
        boost::intrusive_ptr<agas_server> agas,
        boost::intrusive_ptr<actions::hpx_action> act,
        hpx_action_params::ptr params, lco_type prom,
        hpx::components::hpx_component::ptr comp)
    {
      return ptr(new hpx_action_thread(man, agas, act, params, prom, comp));
    }

    virtual
    ~hpx_action_thread()
    {
    }

  protected:
    hpx_action_thread(boost::intrusive_ptr<hpx_thread_manager> man,
        boost::intrusive_ptr<agas_server> agas,
        boost::intrusive_ptr<actions::hpx_action> act,
        hpx_action_params::ptr params, lco_type prom,
        hpx::components::hpx_component::ptr comp) :
        hpx_thread(man, agas), action_(act), params_(params), lco_(prom), comp_(
            comp)
    {

    }

    hpx_action_thread(boost::intrusive_ptr<hpx_thread_manager> man,
        boost::intrusive_ptr<agas_server> agas,
        boost::intrusive_ptr<actions::hpx_action> act,
        hpx_action_params::ptr params, lco_type prom) :
        hpx_thread(man, agas), action_(act), params_(params), lco_(prom), comp_(
            hpx::components::hpx_component::ptr())
    {

    }

    static long COMPUTE_LIB_COUNT;

    virtual void
    run();

  };

}

#endif
