/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <hpx/runtime/hpx_thread.h>
#include <sstmac/software/process/os_api.h>
#include <hpx/runtime/hpx_thread_manager.h>

namespace hpx
{
  sstmac::debug hpx_thread::dbg_;

  double hpx_thread::THREAD_CREATION_OVERHEAD = 1e-6;
  double hpx_thread::THREAD_SWAP_OVERHEAD = 1e-7;

  hpx_thread::hpx_thread(boost::intrusive_ptr<hpx_thread_manager> man,
      boost::intrusive_ptr<agas_server> agas) :
      threadman_(man), agas_(agas)
  {
    if (!dbg_.is_real())
      {
        dbg_ = sstmac::debug("<debug> hpx | hpx_thread", sstmac::logger::blue);
        sstmac::logger::register_logger(&dbg_);
      }
  }

  boost::intrusive_ptr<hpx_thread_manager>
  hpx_thread::get_thread_man()
  {
    return threadman_;
  }

  boost::intrusive_ptr<agas_server>
  hpx_thread::get_agas_server()
  {
    return agas_;
  }

  void
  hpx_thread::hpx_block(sstmac::sw::key::ptr blocker)
  {

    os_->block(blocker);

  }

  void
  hpx_thread::hpx_unblock(sstmac::sw::key::ptr blocker)
  {
    os_->unblock(blocker);
  }

  void
  hpx_thread::hpx_send(sstmac::sst_message::ptr msg)
  {
    os_->execute_kernel(sstmac::sw::AMI::AMI_COMM,
        sstmac::sw::AMI::AMI_COMM_SEND, msg);
  }

  sstmac::nodeaddress::ptr
  hpx_thread::find_here()
  {
    return os_->get_my_addr();
  }

  sstmac::sw::environment::const_ptr
  hpx_thread::get_env(sstmac::sw::appid aid)
  {
    return os_->get_env(aid);
  }

  sstmac::sw::OS_API::ptr
  hpx_thread::get_os()
  {
    return os_;
  }

  void
  hpx_thread::compute(const sstmac::timestamp &t)
  {
    if (!libcomp_)
      {
        throw sstmac::ssterror(
            "hpx_thread::compute -- compute lib has not been initialized");
      }
    libcomp_->compute(t);
  }

  void
  hpx_thread::register_lib(sstmac::sw::library::ptr lib)
  {
    os_->register_lib(lib);
  }



}
