/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_HPX_RUNTIME_HPX_THREAD_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_HPX_RUNTIME_HPX_THREAD_H_INCLUDED

#include <sstmac/software/process/thread.h>
#include <sstmac/software/process/key.h>
#include <sstmac/common/messages/sst_message.h>
#include <sstmac/software/ami/environment.h>
#include <sstmac/software/process/os_api.h>
#include <sstmac/software/libraries/compute/lib_compute_time.h>

namespace hpx
{
  class hpx_thread_manager;
  class agas_server;

  class hpx_thread : public sstmac::sw::thread
  {
  protected:

    boost::intrusive_ptr<hpx_thread_manager> threadman_;
    boost::intrusive_ptr<agas_server> agas_;

    static sstmac::debug dbg_;

    sstmac::sw::lib_compute_time::ptr libcomp_;

  public:
    typedef boost::intrusive_ptr<hpx_thread> ptr;

    static double THREAD_CREATION_OVERHEAD;
    static double THREAD_SWAP_OVERHEAD;

    virtual
    ~hpx_thread()
    {
    }

    boost::intrusive_ptr<hpx_thread_manager>
    get_thread_man();

    boost::intrusive_ptr<agas_server>
    get_agas_server();

    void
    hpx_block(sstmac::sw::key::ptr blocker);

    void
    hpx_unblock(sstmac::sw::key::ptr blocker);

    void
    hpx_send(sstmac::sst_message::ptr msg);

    sstmac::nodeaddress::ptr
    find_here();

    sstmac::sw::environment::const_ptr
    get_env(sstmac::sw::appid aid);

    sstmac::sw::OS_API::ptr
    get_os();


    void
    register_lib(sstmac::sw::library::ptr lib);

    void
    compute(const sstmac::timestamp &t);

  protected:
    hpx_thread(boost::intrusive_ptr<hpx_thread_manager> man,
        boost::intrusive_ptr<agas_server> agas);

  };

}

#endif
