/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_MPI_MPI_QUEUE_MPIQUEUE_RECVREQUEST_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_MPI_MPI_QUEUE_MPIQUEUE_RECVREQUEST_H_INCLUDED

#include <sstmac/software/libraries/mpi/mpi_queue/mpiqueue.h>
#include <sstmac/common/nocopy.h>

namespace sstmac { namespace sw {

  /**
   * A nested type to handle individual mpi receive requests.
   */
  class mpiqueue::recvrequest : virtual public nocopy, public ptr_type {
  public:
     typedef boost::intrusive_ptr<recvrequest> ptr;

  private:

    friend class mpiqueue;

    /// The queue to whom we belong.
    boost::intrusive_ptr<mpiqueue> queue_;

    /// State flag to see whether we are in an ack/send cycle.
    bool acked_;

    /// The parameters I will be matching on.
    mpiid source_;
    mpitag tag_;
    mpicomm comm_;
    mpimessage::category cat_;

    /// The parameters I will not be matching on, but are good error checks.
    int count_;
    mpitype type_;

    mpirequest_t key_;

    /// The object that will be notified upon completion.
    eventhandler::ptr completion_;

  protected:
    /// Hello.
    recvrequest(const mpirequest_t key, const mpiqueue::ptr& queue, int count,
                const mpitype &type, const mpiid &source, const mpitag &tag,
                const mpicomm &comm, mpimessage::category cat,
                eventhandler::ptr &completion);



  public:
    virtual std::string to_string() const{
            return "recvrequest";
          }

    /// Hello.
    static recvrequest::ptr
    construct(const mpirequest_t key, const mpiqueue::ptr& queue, int count,
              const mpitype &type, const mpiid &source, const mpitag &tag,
              const mpicomm &comm, mpimessage::category cat,
              eventhandler::ptr &completion);

    /// Goodbye.
    virtual ~recvrequest();

    /// Do we match the given message?
    inline bool matches(const mpimessage::ptr& msg) const {
      return ((cat_ == msg->cat()) && (comm_.id() == msg->commid()) &&
              source_.matches(msg->source()) && tag_.matches(msg->tag()));
    }

    /// We be done.
    void handle(
                const mpimessage::ptr& mess);
  };

} } // end of namespace sstmac

#endif
