/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_MPI_MPI_STRATEGIES_MPI_COLLECTIVE_ENGINES_MPIRINGBARRIERENGINE_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_MPI_MPI_STRATEGIES_MPI_COLLECTIVE_ENGINES_MPIRINGBARRIERENGINE_H_INCLUDED

#include <sstmac/software/libraries/mpi/mpi_strategies/mpi_collective/mpicollective.h>
#include <sstmac/software/libraries/mpi/mpi_comm/mpicomm.h>
#include <sstmac/software/libraries/mpi/mpi_types/mpitag.h>

#include <sstmac/software/libraries/mpi/mpi_strategies/mpi_iterator/topologyiterator.h>
#include <sstmac/software/libraries/mpi/mpistatus.h>
#include <sstmac/software/libraries/mpi/mpi_queue/mpiqueue.h>

namespace sstmac { namespace sw {

  // Important types.




  /**
   * A basic mpikernel to provide an MPI_Barrier.
   */
  class mpiringbarrierengine : public mpicollective {
    /// The iterator to whom we entrust decisions on where to send and receive.
    topologyiterator::ptr iter_;
    /// Our MPI tag.
    mpitag tag_;


    bool forward_;

    /// Fire off the next set of send and receive operations.
    void sendrecv(bool firsttime);

  protected:
    /// Hi.
    mpiringbarrierengine(const topologyiterator::ptr& iter,
                     const mpirequest::ptr& the_key,
                     const mpiqueue::ptr& queue,
                     const mpitag &tag,
                     const mpicomm &comm,
                     eventhandler::ptr completion);

    /// Callback method to indicate that a send operation has completed.
    virtual void send_complete(
                               int count,
                               const mpitype &type,
                               const mpitag &tag,
                               const mpiid &dest);

    /// Callback method to indicate that a receive operation has completed.
    virtual void recv_complete(
                               int count,
                               const mpistatus_t& stat);

    /// Callback method to indicate that a copy is done.
    /// Does nothing, since this object doesn't copy data.
    virtual void copy_complete(
                               int count,
                               const mpitype &type,
                               const mpitag &tag,
                               const mpiid &source);
    
  public:
    typedef boost::intrusive_ptr<mpiringbarrierengine> ptr;

    /// Hi.
    static mpiringbarrierengine::ptr
    construct(const topologyiterator::ptr& iter,
              const mpirequest::ptr& the_key,
              const mpiqueue::ptr& queue,
              const mpitag &tag,
              const mpicomm &comm,
              eventhandler::ptr completion);

    /// Goodbye.
    virtual ~mpiringbarrierengine() throw();

    /// Start this kernel.
    virtual void start();
  };

} } // end of namespace sstmac

#endif
