/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/libraries/mpi/mpi_strategies/mpicorebarrier.h>
#include <sstmac/software/libraries/mpi/mpi_strategies/mpi_collective/engines/mpibarrierengine.h>
#include <sstmac/software/libraries/mpi/mpi_strategies/mpi_collective/engines/mpilinearbarrierengine.h>
#include <sstmac/software/libraries/mpi/mpi_strategies/mpi_collective/engines/mpiringbarrierengine.h>

#include <sstmac/software/libraries/mpi/mpi_strategies/mpi_iterator/hfmiterator.h>
#include <sstmac/software/libraries/mpi/mpi_strategies/mpi_iterator/rditerator.h>
#include <sstmac/software/libraries/mpi/mpi_strategies/mpi_iterator/btscatteriterator.h>
#include <sstmac/software/libraries/unblockhandler.h>
#include <sstmac/common/eventhandler.h>

namespace sstmac
{
  namespace sw
  {

    // Pointer types.

    //
    // Hello.
    //
    mpicorebarrier::mpicorebarrier(BARRIER_IMPL i) :
        impl_(i)
    {
    }

    //
    // Hi.
    //
    mpicorebarrier::ptr
    mpicorebarrier::construct(BARRIER_IMPL i)
    {
      return mpicorebarrier::ptr(new mpicorebarrier(i));
    }

    //
    // Goodbye.
    //
    mpicorebarrier::~mpicorebarrier() throw ()
    {
    }

    //
    // Build a kernel to synchronize nodes.
    //
    void
    mpicorebarrier::execute(const mpirequest_t& thekey,
        const mpiqueue::ptr& queue, const mpitag &tag, const mpicomm &comm,
        OS_API::ptr &os) const
    {

      key::ptr k = thekey->get_key();
      eventhandler::ptr completion = unblockhandler::construct(os, k);

      // Here we might want to pick different iterators depending on message size.
      topologyiterator::ptr iter;

      switch (impl_)
        {
      case HFM:
        {
          iter = hfmiterator::construct(comm.rank().id, comm.size().id);
          mpibarrierengine::ptr engine = mpibarrierengine::construct(iter,
              thekey, queue, tag, comm, completion);
          engine->start();
          break;
        }

      case LINEAR:
        {
          mpilinearbarrierengine::ptr engine =
              mpilinearbarrierengine::construct(iter, thekey, queue, tag, comm,
                  mpiid(0), completion);
          engine->start();
          break;
        }

      case RING:
        mpiringbarrierengine::ptr engine = mpiringbarrierengine::construct(
            iter, thekey, queue, tag, comm, completion);
        engine->start();
        break;

        }

    }

  }
} // end of namespace sstmac
