/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/libraries/mpi/mpi_strategies/mpicoregather.h>

#include <sstmac/software/libraries/unblockhandler.h>
#include <sstmac/software/libraries/mpi/mpi_strategies/mpi_collective/engines/mpibtreegather.h>

namespace sstmac { namespace sw
{

  
  //
  // Hi.
  //
  mpicoregather::mpicoregather()
  {
  }

  //
  // Hello.
  //
  mpicoregather::ptr
  mpicoregather::construct()
  {
    return mpicoregather::ptr(new mpicoregather);
  }

  //
  // Perform a fused reduce-gather operation
  //
  void
  mpicoregather::execute(const mpirequest_t& thekey, const
      mpiqueue::ptr& queue, int sendcount, const mpitype &sendtype, int recvcount,
      const mpitype &recvtype, const mpiid &root, const mpitag &tag,
      const mpicomm &comm, const payload::const_ptr& content,
      OS_API::ptr &os) const
  {
    // return mpibtreegather::construct(thekey, queue, sendcount, sendtype,
    //                                  recvcount, recvtype, root, tag, comm,
    //                                  content, magic);

    key::ptr k = thekey->get_key();
    eventhandler::ptr completion = unblockhandler::construct(os, k);
    mpibtreegather::ptr engine = mpibtreegather::construct(
        thekey, queue, sendcount, sendtype, recvcount, recvtype, root, tag,
        comm, content, completion);

    engine->start();

  }

} } // end of namespace sstmac
