/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2011 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/libraries/mpi/mpiapi-persistent.h>
#include <sstmac/software/libraries/mpi/mpistatus.h>
#include <sstmac/software/process/key.h>
#include <sstmac/common/errors.h>

namespace sstmac { namespace sw {

  


  mpiapi::persistent::persistent() :
    mpirequest(key::construct())
  {}

  mpiapi::persistent::~persistent() throw() {
  }

  mpiapi::persistentsend::
  persistentsend(mpiapi::ptr api,
                 sendfunc_t func, int count, const mpitype &type,
                 const mpiid &target, const mpitag &tag, const mpicomm &comm,
                 const payload::const_ptr& content) :
    parent_(api), func_(func), count_(count),
    type_(type), target_(target), tag_(tag), comm_(comm), content_(content)
  {
    if(parent_ == NULL || func_ == NULL)
      throw illformederror("mpiapi::persistentsend:  Invalid parent info.");

  }

  mpiapi::persistentsend::ptr mpiapi::persistentsend::
  construct(mpiapi::ptr api,
            sendfunc_t func, int count, const mpitype &type,
            const mpiid &target, const mpitag &tag, const mpicomm &comm,
            const payload::const_ptr& content)
  {
    persistentsend::ptr retval
      (new persistentsend(api, func, count, type, target, tag, comm, content));

    return retval;
  }

  mpiapi::persistentsend::~persistentsend() throw() {
  }

  //
  // Start this request.
  //
  void mpiapi::persistentsend::start() {

    ((*parent_).*func_)(count_, type_, target_, tag_, comm_,
                      persistentsend::ptr(this), content_);
  }


  mpiapi::persistentrecv::
  persistentrecv(mpiapi::ptr api, int count, const mpitype &type,
                 const mpiid &dest, const mpitag &tag, const mpicomm &comm) :
    parent_(api), count_(count), type_(type),
    dest_(dest), tag_(tag), comm_(comm)
  {
    if(parent_ == NULL)
      throw illformederror("mpiapi::persistentrecv:  Invalid parent info.");

  }

  mpiapi::persistentrecv::ptr mpiapi::persistentrecv::
  construct(mpiapi::ptr api, int count, const mpitype &type,
            const mpiid &dest, const mpitag &tag, const mpicomm &comm)
  {
    persistentrecv::ptr retval
      (new persistentrecv(api, count, type, dest, tag, comm));

    return retval;
  }

  mpiapi::persistentrecv::~persistentrecv() throw() {
  }

  //
  // Start this request.
  //
  void mpiapi::persistentrecv::start() {


    parent_->do_irecv(count_, type_, dest_, tag_, comm_, ptr(this));
  }


} } // end of namespace sstmac
