/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2011 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/libraries/mpi/sstmac_mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <iostream>

extern "C" void
mpi_init_(int &err)
{
  err = MPI_Init(0, 0);
}

extern "C" void
mpi_finalize_(int &err)
{
  err = MPI_Finalize();
}

extern "C" void
mpi_comm_world_(int &world)
{
  world = (int) MPI_Comm_World();
}

extern "C" void
mpi_abort_(int &comm, int &code, int &err)
{
  printf("MPI_Abort was called with code %i", code);
  exit(1);
}

extern "C" void
mpi_comm_rank_(int &comm, int &rank, int &err)
{
  err = MPI_Comm_rank(comm, &rank);
}

extern "C" void
mpi_comm_size_(int &comm, int &size, int &err)
{
  err = MPI_Comm_size(comm, &size);
}

extern "C" void
mpi_send0_(int &count, int &datatype, int &dest, int &tag,
    int &comm, int &err)
{

  err = MPI_Send(NULL, count, datatype, dest, tag, comm);

}

extern "C" void
mpi_send1_(long &buf, int &count, int &datatype, int &dest, int &tag,
    int &comm, int &err)
{
  if (count > 1)
    {
      printf("fortran mpi_send: buffers not yet supported \n");
      exit(1);
    }
  err = MPI_Send(&buf, count, datatype, dest, tag, comm);
}
/*
 extern "C" int
 mpi_send2_(long *buf, int &count, int &datatype, int &dest, int &tag,
 int &comm, int &err){
 printf("calling this version of send\n");
 err = MPI_Send(buf, count, datatype, dest, tag, comm);
 }*/

extern "C" void
mpi_send3_(double &buf, int &count, int &datatype, int &dest,
    int &tag, int &comm, int &err)
{
  if (count > 1)
    {
      printf("fortran mpi_send: buffers not yet supported \n");
      exit(1);
    }
  err = MPI_Send(&buf, count, datatype, dest, tag, comm);
}

extern "C" void
mpi_recv0_(int &count, MPI_Datatype &datatype, int &source, int &tag,
    MPI_Comm &comm, int *status, int &err)
{
  MPI_Status cstat;

  err = MPI_Recv(NULL, count, datatype, source, tag, comm, &cstat);

  status[0] = cstat.MPI_SOURCE;
  status[1] = cstat.MPI_TAG;
  status[2] = cstat.MPI_ERROR;
  status[3] = cstat.BYTES_RECV;
}

extern "C" void
mpi_recv1_(long &buf, int &count, MPI_Datatype &datatype, int &source,
    int &tag, MPI_Comm &comm, int *status, int &err)
{
  MPI_Status cstat;

  if (count > 1)
    {
      printf("fortran mpi_recv: buffers not yet supported \n");
      exit(1);
    }

  err = MPI_Recv(&buf, count, datatype, source, tag, comm, &cstat);
  status[0] = cstat.MPI_SOURCE;
  status[1] = cstat.MPI_TAG;
  status[2] = cstat.MPI_ERROR;
  status[3] = cstat.BYTES_RECV;
}
/*
 extern "C" int
 mpi_recv2_(long *buf, int &count, MPI_Datatype &datatype, int &source, int &tag,
 MPI_Comm &comm, int *status, int &err){
 MPI_Status cstat;
 printf("calling this version of recv \n");
 err = MPI_Recv(buf, count, datatype, source, tag, comm, &cstat);
 status[0] = cstat.MPI_SOURCE;
 status[1] = cstat.MPI_TAG;
 status[2] = cstat.MPI_ERROR;
 status[3] = cstat.BYTES_RECV;
 }*/

extern "C" void
mpi_recv3_(double &buf, int &count, MPI_Datatype &datatype, int &source,
    int &tag, MPI_Comm &comm, int *status, int &err)
{
  MPI_Status cstat;

  if (count > 1)
    {
      printf("fortran mpi_recv: buffers not yet supported \n");
      exit(1);
    }
  err = MPI_Recv(&buf, count, datatype, source, tag, comm, &cstat);
  status[0] = cstat.MPI_SOURCE;
  status[1] = cstat.MPI_TAG;
  status[2] = cstat.MPI_ERROR;
  status[3] = cstat.BYTES_RECV;
}

extern "C" void
mpi_wait_(MPI_Request &request, int *status, int &err)
{
  MPI_Status cstat;
  err = MPI_Wait(&request, &cstat);
  status[0] = cstat.MPI_SOURCE;
  status[1] = cstat.MPI_TAG;
  status[2] = cstat.MPI_ERROR;
  status[3] = cstat.BYTES_RECV;
}

extern "C" void
mpi_waitany_(int &count, MPI_Request array_of_requests[], int &index,
    int *status, int& err)
{
  MPI_Status cstat;
  err = MPI_Waitany(count, array_of_requests, &index, &cstat);
  status[0] = cstat.MPI_SOURCE;
  status[1] = cstat.MPI_TAG;
  status[2] = cstat.MPI_ERROR;
  status[3] = cstat.BYTES_RECV;
}

extern "C" void
mpi_waitsome_(int &incount, MPI_Request array_of_requests[], int &outcount,
    int array_of_indices[], int* array_of_statuses[], int &err)
{
  MPI_Status *cstat;
  err = MPI_Waitsome(incount, array_of_requests, &outcount, array_of_indices,
      cstat);
  for (int i = 0; i < outcount; i++)
    {
      array_of_statuses[i][0] = cstat[i].MPI_SOURCE;
      array_of_statuses[i][1] = cstat[i].MPI_TAG;
      array_of_statuses[i][2] = cstat[i].MPI_ERROR;
      array_of_statuses[i][3] = cstat[i].BYTES_RECV;
    }
}

extern "C" void
mpi_get_count_(int *status, MPI_Datatype &datatype, int &count, int &err)
{
  MPI_Status cstat;
  err = MPI_Get_count(&cstat, datatype, &count);
  status[0] = cstat.MPI_SOURCE;
  status[1] = cstat.MPI_TAG;
  status[2] = cstat.MPI_ERROR;
  status[3] = cstat.BYTES_RECV;
}

extern "C" void
mpi_test_(MPI_Request &request, int &flag, int *status, int &err)
{
  MPI_Status cstat;
  err = MPI_Test(&request, &flag, &cstat);
  status[0] = cstat.MPI_SOURCE;
  status[1] = cstat.MPI_TAG;
  status[2] = cstat.MPI_ERROR;
  status[3] = cstat.BYTES_RECV;
}

extern "C" void
mpi_cancel_(MPI_Request &request, int &err)
{
  err = MPI_Cancel(&request);
}

extern "C" void
mpi_request_free_(MPI_Request &request, int &err)
{
  err = MPI_Request_free(&request);
}

extern "C" void
mpi_comm_dup_(MPI_Comm &comm, MPI_Comm &newcomm, int &err)
{
  err = MPI_Comm_dup(comm, &newcomm);
}

extern "C" void
mpi_comm_split_(MPI_Comm &comm, int &color, int &key, MPI_Comm &newcomm,
    int err)
{
  err = MPI_Comm_split(comm, color, key, &newcomm);
}

extern "C" void
mpi_comm_free_(MPI_Comm &comm, int &err)
{
  err = MPI_Comm_free(&comm);
}

extern "C" void
mpi_comm_create_(MPI_Comm &comm, MPI_Group &group, MPI_Comm &newcomm, int &err)
{
  err = MPI_Comm_create(comm, group, &newcomm);
}

extern "C" void
mpi_comm_group_(MPI_Comm &comm, MPI_Group &group, int err)
{
  err = MPI_Comm_group(comm, &group);
}

extern "C" void
mpi_group_incl_(MPI_Group &group, int &n, int *ranks, MPI_Group &newgroup,
    int &err)
{
  err = MPI_Group_incl(group, n, ranks, &newgroup);
}

extern "C" void
mpi_group_free_(MPI_Group &group, int &err)
{
  err = MPI_Group_free(&group);
}

extern "C" void
mpi_sendrecv0_(int &sendcount, MPI_Datatype &sendtype, int &dest, int &sendtag,
    int &recvcount, MPI_Datatype &recvtype, int &source, int &recvtag,
    MPI_Comm &comm, int *status, int &err)
{
  MPI_Status cstat;
  err = MPI_SendRecv(NULL, sendcount, sendtype, dest, sendtag, NULL, recvcount,
      recvtype, source, recvtag, comm, &cstat);
  status[0] = cstat.MPI_SOURCE;
  status[1] = cstat.MPI_TAG;
  status[2] = cstat.MPI_ERROR;
  status[3] = cstat.BYTES_RECV;
}

extern "C" void
mpi_isend0_(int &count, MPI_Datatype &datatype, int &dest, int &tag,
    MPI_Comm &comm, MPI_Request &request, int &err)
{
  err = MPI_Isend(NULL, count, datatype, dest, tag, comm, &request);
}

extern "C" void
mpi_irecv0_(int &count, MPI_Datatype &datatype, int &source, int &tag,
    MPI_Comm &comm, MPI_Request &request, int &err)
{
  err = MPI_Irecv(NULL, count, datatype, source, tag, comm, &request);
}

extern "C" void
mpi_recover_payload_(int *buf, MPI_Request *request, int &err)
{
  printf("fortran mpi: mpi_recover_payload is not implemented yet");
  exit(1);
}

extern "C" void
mpi_allreduce0_(int &count, int &datatype, int &op, int &comm,
    int &err)
{
  err = MPI_Allreduce(NULL, NULL, count, datatype, op, comm);
}

extern "C" void
mpi_reduce0_(int &count, int &datatype, int &op, int &root,
    int &comm, int &err)
{
  std::cerr << "MPI_Type: " << datatype << std::endl;
  err = MPI_Reduce(NULL, NULL, count, datatype, op, root, comm);
}

extern "C" void
mpi_barrier_(int &comm, int &err)
{
  err = MPI_Barrier(comm);
}

extern "C" void
mpi_bcast0_(int &count, int &datatype, int &root, int &comm,
    int &err)
{
  err = MPI_Bcast(NULL, count, datatype, root, comm);
}

extern "C" void
mpi_gather0_(int &sendcnt, int &sendtype, int &recvcnt,
    int &recvtype, int &root, int &comm, int &err)
{
  err
      = MPI_Gather(NULL, sendcnt, sendtype, NULL, recvcnt, recvtype, root, comm);
}

extern "C" void
mpi_allgather0_(int &sendcount, int &sendtype, int &recvcount,
    int &recvtype, int &comm, int &err)
{
  err = MPI_Allgather(NULL, sendcount, sendtype, NULL, recvcount, recvtype,
      comm);
}

extern "C" void
mpi_scatter0_(int &sendcnt, int &sendtype, int &recvcnt,
    int &recvtype, int &root, int &comm, int &err)
{
  err = MPI_Scatter(NULL, sendcnt, sendtype, NULL, recvcnt, recvtype, root,
      comm);
}

extern "C" void
mpi_alltoall0_(int &sendcount, int &sendtype, int &recvcount,
    int &recvtype, int &comm, int &err)
{
  err
      = MPI_Alltoall(NULL, sendcount, sendtype, NULL, recvcount, recvtype, comm);
}


extern "C" void
mpi_wtime_(double &t)
{
  t = MPI_Wtime();
}

