/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/software/process/app.h>

namespace sstmac
{
  namespace sw
  {

    debug app::dbg_;

    void
    app::compute(timestamp time)
    {

      if (!computetime_)
        {
          computetime_ = lib_compute_time::construct(id_);
          register_lib(computetime_);
        }
      computetime_->compute(time);
    }

    void
    app::compute_inst(const eventdata::ptr &cmsg)
    {
      if (!computeinst_)
        {
          computeinst_ = lib_compute_inst::construct(id_);
          register_lib(computeinst_);
        }
      computeinst_->compute_inst(cmsg);
    }

    void
    app::compute_block_access(long bytes)
    {
      if (!computememmove_)
        {
          computememmove_ = lib_compute_memmove::construct(id_);
          register_lib(computememmove_);
        }
      computememmove_->compute_block_access(bytes);
    }

    void
    app::sleep(timestamp time)
    {
      if (!sleeplib_)
        {
          sleeplib_ = lib_sleep::construct(id_);
          register_lib(sleeplib_);
        }
      sleeplib_->sleep(time);
    }

    std::string
    app::get_compute_name()
    {

      if (!computetime_)
        {
          computetime_ = lib_compute_time::construct(id_);
          register_lib(computetime_);
        }
      return computetime_->get_lib_name();
    }

    std::string
    app::get_compute_inst_name()
    {
      if (!computeinst_)
        {
          computeinst_ = lib_compute_inst::construct(id_);
          register_lib(computeinst_);
        }
      return computeinst_->get_lib_name();
    }

    std::string
    app::get_compute_memmove_name()
    {
      if (!computememmove_)
        {
          computememmove_ = lib_compute_memmove::construct(id_);
          register_lib(computememmove_);
        }
      return computememmove_->get_lib_name();
    }

    std::string
    app::get_sleep_name()
    {
      if (!sleeplib_)
        {
          sleeplib_ = lib_sleep::construct(id_);
          register_lib(sleeplib_);
        }
      return sleeplib_->get_lib_name();
    }

  }
}
