/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_PROCESS_APP_H_INCLUDED
#define SSTMAC_SOFTWARE_PROCESS_APP_H_INCLUDED

#include <sstmac/software/libraries/compute/lib_compute_inst.h>
#include <sstmac/software/libraries/compute/lib_compute_time.h>
#include <sstmac/software/libraries/compute/lib_compute_memmove.h>
#include <sstmac/software/libraries/compute/lib_sleep.h>
#include <sstmac/common/ptr_type.h>
#include <sstmac/software/process/thread.h>
#include <sstmac/common/sim_parameters.h>
//#include <sstmac/hardware/processor/eventdata.h>
#include <string>
#include <list>

namespace sstmac
{
  namespace sw
  {

    /**
     * The app derived class adds to the thread base class by providing
     * facilities to allow applications to simulate computation.
     * Messaging models are supported through derivation of this class,
     * as in the mpiapp class.
     */
    class app : public thread
    {
      lib_compute_inst::ptr computeinst_;
      lib_compute_time::ptr computetime_;
      lib_compute_memmove::ptr computememmove_;
      lib_sleep::ptr sleeplib_;

    protected:
    
      sim_parameters::ptr params_;

      app(softwareid id) :
          id_(id)
      {
        if (!dbg_.is_real())
          {
            dbg_ = debug("<debug> thread | app", logger::green);
            logger::register_logger(&dbg_);
          }
      }

      static debug dbg_;

      softwareid id_;

    public:

      softwareid
      get_sid() const {
        return id_;
      }
      // convenience functions for apps
      // public to allow C interface
      void
      compute(timestamp time);

      void
      compute_inst(const eventdata::ptr &cmsg);

      void
      compute_block_access(long bytes);

      void
      sleep(timestamp time);

      std::string
      get_compute_name();

      std::string
      get_compute_inst_name();

      std::string
      get_sleep_name();

      std::string
      get_compute_memmove_name();

      typedef boost::intrusive_ptr<app> ptr;

      /// Goodbye.
      virtual
      ~app()
      {
      }
        


      virtual void
      consume_params(const sim_parameters::ptr &params) = 0;

      virtual app::ptr
      clone(softwareid newid) = 0;

      //called when killing the app, in case you want to check or clean anything up before destructor
      virtual void
      kill()
      {

        thread::kill();
      }

      sim_parameters::ptr
      parameters() {
        return params_;
      }

      virtual void
      init_libs(sim_parameters::ptr params) {
        params_ = params;
      }

      virtual void
      skeleton_main() = 0;

      virtual void
      run()
      {
        skeleton_main();
      }
    };

  }
} // end of namespace sstmac.

#endif
