/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/software/process/appdescr.h>
#include <sstmac/common/driver_util.h>
//#include <sstmac/backends/native/manager.h>
//#include <sstmac/backends/native/driver.h>

namespace sstmac {

std::map<const char*, SSTAppDesc_base*>* SSTAppDesc_build::appmap_ = 0;

void
SSTAppDesc_build::register_app(const char* name, SSTAppDesc_base* descr)
{
    if (appmap_ == 0)
        appmap_ = new std::map<const char*, SSTAppDesc_base*>;

    (*appmap_)[name] = descr;
}

void
SSTAppDesc_build::init_apps(sim_parameters::ptr params)
{
    if (appmap_ == 0)
        return;
    
    std::map<const char*, SSTAppDesc_base*>::const_iterator it, end = appmap_->end();
    for (it = appmap_->begin(); it != end; ++it)
    {
        const char* name = it->first;
        app::ptr application = it->second->build_app();
        application->init_libs(params);
        sw::skeletonfactory::apps[it->first] = application;
    }
}


} //end namespace sstmac

