/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_PROCESS_BLOCKWAITER_H_INCLUDED
#define SSTMAC_SOFTWARE_PROCESS_BLOCKWAITER_H_INCLUDED

#include <sstmac/software/process/waiter.h>
#include <sstmac/software/process/os_api.h>
#include <sstmac/software/process/key.h>

namespace sstmac
{
  namespace sw
  {

    /**
     * A waiter that blocks through the OS, which should only be called from children
     * of thread.  Otherwise, the main DES thread will try to swap a context to nothing.
     */
    class blockwaiter : public waiter
    {
    public:

      typedef boost::intrusive_ptr<blockwaiter> ptr;

    protected:
      OS_API::ptr os_;
      key::ptr key_;

      blockwaiter(OS_API::ptr &os, key::ptr &k) :
          os_(os), key_(k)
      {

      }

    public:
      virtual std::string
      to_string() const
      {
        return "blockwaiter";
      }

      virtual
      ~blockwaiter()
      {

      }

      static ptr
      construct(OS_API::ptr &os, key::ptr &k)
      {
        ptr ret(new blockwaiter(os, k));
        return ret;
      }

      virtual void
      wait()
      {
        os_->block(key_);
      }

    };
  }
} // end of namespace sstmac
#endif
