/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/process/key.h>
#include <sstmac/common/alloc.h>
#include <iostream>
#include <stdlib.h>

namespace sstmac { namespace sw {
  
  // Pointer types.
  
  

  // The key objects are very short-lived, so we don't feel like
  // continually allocating and deallocating those bastards.
  namespace keypool {
    static std::vector<key*>* chunks = 0;
  } // end of namespace keypool.

key::ptr key::construct() {
      return key::ptr(new key);
}

key::~key(){

}

void* key::operator new(size_t size) {
  if(! keypool::chunks) {
    keypool::chunks = new std::vector<key*>;
    keypool::chunks->reserve(1024);
  }

  void* memptr;
  if(!keypool::chunks->empty()) {
    memptr = keypool::chunks->back();
    keypool::chunks->pop_back();
  }
  else{
      memptr = ::malloc(size);
  }
  return memptr;
}

void key::add_blocker(thread_data_t t)
{
    blocking_t::iterator it = blocking_.find(t);
    if (it != blocking_.end())
    {
        throw illformederror("key is blocking the same thread twice");
    }
    blocking_.insert(t);
}

void key::remove_blocker(thread_data_t t)
{
    blocking_.erase(t);
}

/**
    I am painfully aware that I am leaking memory here like a sieve... but for
    time being I'm not going to sweat what is basically 50K of leaked memory
    JJW 01/09/12
*/
void key::operator delete(void* ptr){
  keypool::chunks->push_back(static_cast<key*>(ptr));
}

} } // end of namespace sstmac.
