/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_BACKENDS_NATIVE_LAUNCH_RACKALLOCATOR_H_INCLUDED
#define SSTMAC_BACKENDS_NATIVE_LAUNCH_RACKALLOCATOR_H_INCLUDED

#include <sstmac/software/services/launch/allocation/allocationstrategy.h>

namespace sstmac
{
  namespace sw
  {

    /**
     * This is an Allocator that attempts to allocate whole racks at a time
     * for an application.
     */
    class rackallocator : public allocationstrategy
    {
    protected:
      rackallocator(int numOfNodes, int coresPerProc, int procsPerNode,
          int nodesPerRack, int numRacks);

    public:
      typedef boost::intrusive_ptr<rackallocator> ptr;

      static ptr
      construct(int non, int cpp, int ppn, int npr, int nr)
      {
        return ptr(new rackallocator(non, cpp, ppn, npr, nr));
      }

      static ptr
      construct(sim_parameters::ptr params)
      {
        return construct(params->get_long_param("node_num"), params->get_long_param("node_cores"),
            params->get_long_param("node_procs"), params->get_long_param("rack_nodes"), params->get_long_param("rack_num"));
      }

      virtual void
      allocate(int nnode_requested,
          std::set<nodeaddress::ptr, nodeaddress::nodeaddresscomp> &allocation);

      virtual
      ~rackallocator() throw ()
      {
      }

    protected:
      int currentRack;

      int cpp_;
      int ppn_;
      int npr_;
      int nr_;
    };
  }
}
#endif
