/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/software/services/launch/indexing/dumpiindexer.h>

#include <fstream>
#include <sstmac/common/errors.h>

#include <sstmac/software/skeletons/undumpi/dumpiutil.h>
#include <sstmac/software/skeletons/undumpi/dumpimeta.h>
#include <dumpi/libundumpi/libundumpi.h>
#include <sstmac/software/services/launch/allocation/dumpiallocation.h>

namespace sstmac
{
  namespace sw
  {

    using namespace std;

    void
    dumpiindexer::allocate(
        const std::set<nodeaddress::ptr, nodeaddress::nodeaddresscomp> &nodes,
        int ppn,
        std::vector<nodeaddress::ptr> &result,
        int nproc)
    {
      static std::string here("dumpiindexing::allocate: ");

      dumpimeta::ptr meta = dumpimeta::construct(metaname_);

      std::set<nodeaddress::ptr, nodeaddress::nodeaddresscomp> nodeidset;
      int nrank = getnumprocs(meta);
      result.resize(nrank);

      std::cout << "nrank: " << nrank << std::endl;
      for (int i = 0; i < nrank; i++)
        {
          std::string fname = dumpi_file_name(i, meta->dirplusfileprefix_);
          dumpi_profile *profile = undumpi_open(fname.c_str());
          if (profile == NULL)
            {
              throw ioerror(here + "unable to open \"" + fname + "\"");
            }
          dumpi_header *header = dumpi_build_header();
          if (header == NULL)
            {
              throw ioerror(here + " header is null");
            }
          dumpi_read_header(profile, header);
          if (header->hostname == NULL || *header->hostname == 0)
            {
              throw ioerror(here + " hostname is null or empty");
            }
          std::string hostname(header->hostname);
#if 0
          char tmp[100];
          if (i < 10)            sprintf(tmp, "nid0000%d", i);
          else if (i < 100)      sprintf(tmp, "nid000%d", i);
          else if (i < 1000)     sprintf(tmp, "nid00%d", i);
          else if (i < 10000)    sprintf(tmp, "nid0%d", i);
          else if (i < 100000)   sprintf(tmp, "nid%d", i);
          hostname = tmp;
#endif

          dumpi_free_header(header);

          dumpiallocation::nodemap_t::const_iterator
                    aptr_it = dumpiallocation::hostnamemap_.find(hostname),
                    end = dumpiallocation::hostnamemap_.end();

          if (aptr_it == end)
          {
                std::stringstream sstr;
                sstr << hostname << " from dumpi file does not exist in node map.";

                dumpiallocation::nodemap_t::const_iterator
                        it = dumpiallocation::hostnamemap_.begin();

                if (it == end)
                {
                    sstr << " No hostnames are registered with dumpiallocation."
                         << " This is perhaps not surprising then."
                         << " Maybe check that launch_allocation is set to dumpi.";
                }
                else
                {
                    sstr << endl << "Valid hostnames are: ";
                    for ( ; it != end; ++it)
                    {
                        sstr << endl << it->first;
                    }
                    sstr << endl << endl
                         << "Are you sure the dumpi file and node map"
                            " are from the same machine?";
                }
                throw sstmac::inputerror(sstr.str());
          }

          nodeaddress::ptr aptr = aptr_it->second;

          SSTMAC_DEBUG << here << " rank " << i << " is on " << hostname << ", at " << aptr->to_string() << "\n";


          result[i] = aptr;

        }

    }
  }
}
