/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/software/services/servdescr.h>
#include <sstmac/common/driver_util.h>
#include <sstmac/hardware/node/node.h>
//#include <sstmac/backends/native/manager.h>
//#include <sstmac/backends/native/driver.h>

namespace sstmac
{
  namespace sw
  {

    std::map<const char*, SSTServDesc_base*>* SSTServDesc_build::servmap_ = 0;

    void
    SSTServDesc_build::register_service(const char* name, SSTServDesc_base* descr)
    {
      if (servmap_ == 0)
        servmap_ = new std::map<const char*, SSTServDesc_base*>;

      (*servmap_)[name] = descr;
    }

    void
    SSTServDesc_build::start_services(sstmac::hw::node::ptr node)
    {
      if (servmap_ == 0)
        return;

      std::map<const char*, SSTServDesc_base*>::const_iterator it, end =
          servmap_->end();
      for (it = servmap_->begin(); it != end; ++it)
        {
          const char* name = it->first;
          service::ptr serv = it->second->build_service();
          node->add_service(serv);
        }
    }

  }
} //end namespace sstmac

