/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/common/sstmac-config.h>

#include <sstmac/software/skeletons/skeletonfactory.h>

#include <sstmac/software/process/softwareid.h>

#include <sstmac/software/skeletons/test/app_helloworld.h>
#include <sstmac/software/skeletons/test/mpipingpong.h>
#include <sstmac/software/skeletons/test/mpipairwise.h>
#include <sstmac/software/skeletons/test/mpi_testall.h>
#include <sstmac/software/skeletons/test/sstmac_mpi_testall.h>
#include <sstmac/software/skeletons/test/test_file_system.h>
#include <sstmac/software/skeletons/test/test_nppfs.h>
#include <sstmac/software/skeletons/test/test_mpi_barrier.h>
#include <sstmac/software/skeletons/undumpi/parsedumpi.h>
#include <sstmac/software/libraries/mpi/user_mpiapp_cxx.h>
#include <sstmac/software/libraries/mpi/user_mpiapp_c.h>
#include <sstmac/software/libraries/mpi/user_mpiapp_f90.h>
#include <sstmac/software/skeletons/test/test_compute.h>
#include <sstmac/software/skeletons/test/test_compute_f.h>
//#include <sstmac/software/skeletons/examples/example_tascel.h>

namespace sstmac
{
  namespace sw
  {

    std::map<std::string, app::ptr> skeletonfactory::apps;
    bool skeletonfactory::initialized_ = false;

    void
    skeletonfactory::init_skeletons(sim_parameters::ptr params)
    {
      if (!initialized_)
        { //only do it once
          appid dumma(-1);
          taskid dummt(-1);
          softwareid dumms(dumma, dummt);

          skeletonfactory::apps["hello"] =
              sstmac::sw::app_helloworld::construct(dumms);
          skeletonfactory::apps["hello"]->init_libs(params);

          skeletonfactory::apps["MPI_pingpong"] =
              sstmac::sw::mpipingpong::construct(dumms);
          skeletonfactory::apps["MPI_pingpong"]->init_libs(params);

          skeletonfactory::apps["MPI_pairwise"] =
              sstmac::sw::mpipairwise::construct(dumms);
          skeletonfactory::apps["MPI_pairwise"]->init_libs(params);

          skeletonfactory::apps["MPI_testall"] =
              sstmac::sw::mpi_testall::construct(dumms);
          skeletonfactory::apps["MPI_testall"]->init_libs(params);

          skeletonfactory::apps["SSTMAC_MPI_testall"] =
              sstmac::sw::sstmac_mpi_testall::construct(dumms);
          skeletonfactory::apps["SSTMAC_MPI_testall"]->init_libs(params);

          skeletonfactory::apps["test_compute"] =
              sstmac::sw::test_compute::construct(dumms);
          skeletonfactory::apps["test_compute"]->init_libs(params);

          skeletonfactory::apps["test_mpi_barrier"] =
              sstmac::sw::test_mpi_barrier::construct(dumms);
          skeletonfactory::apps["test_mpi_barrier"]->init_libs(params);

          skeletonfactory::apps["test_file_system"] =
              sstmac::sw::test_file_system::construct(dumms);
          skeletonfactory::apps["test_file_system"]->init_libs(params);

          skeletonfactory::apps["test_nppfs"] =
              sstmac::sw::test_nppfs::construct(dumms);
          skeletonfactory::apps["test_nppfs"]->init_libs(params);

          //run apps using native MPI interfaces, expects C++ linkage
          skeletonfactory::apps["user_mpiapp_cxx"] =
              sstmac::sw::user_mpiapp_cxx::construct(dumms);
          skeletonfactory::apps["user_mpiapp_cxx"]->init_libs(params);

          //run apps using native MPI interfaces, expects C linkage
          skeletonfactory::apps["user_mpiapp_c"] = 
              sstmac::sw::user_mpiapp_c::construct(dumms);
          skeletonfactory::apps["user_mpiapp_c"]->init_libs(params);

          skeletonfactory::apps["parsedumpi"] =
              sstmac::sw::parsedumpi::construct(dumms);
          skeletonfactory::apps["parsedumpi"]->init_libs(params);

        #ifdef SSTMAC_HAVE_FORTRAN
          skeletonfactory::apps["test_compute_f"] =
              sstmac::sw::test_compute_f::construct(dumms);
          skeletonfactory::apps["test_compute_f"]->init_libs(params);

          //run apps using native MPI interfaces, expects F90 linkage
          skeletonfactory::apps["user_mpiapp_f90"] =
              sstmac::sw::user_mpiapp_f90::construct(dumms);
          skeletonfactory::apps["user_mpiapp_f90"]->init_libs(params);
        #endif

          initialized_ = true;
        }
    }

    void
    skeletonfactory::print_factory()
    {
      std::cout << "Contents of skeleton factory are: " << std::endl;

      std::map<std::string, app::ptr>::iterator it, end = apps.end();
      for (it = apps.begin(); it != end; it++)
        {
          std::cout << it->first << std::endl;
        }
    }

  }
} //end of namespace sstmac
