/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_SKELETONS_TEST_MPIPAIRWISE_H_INCLUDED
#define SSTMAC_SOFTWARE_SKELETONS_TEST_MPIPAIRWISE_H_INCLUDED

#include <sstmac/software/libraries/mpi/mpiapp.h>
#include <vector>

namespace sstmac
{
  namespace sw
  {

    /**
     * Basic MPI pairwise send-receive.
     */
    class mpipairwise : public mpiapp
    {
    public:
      typedef boost::intrusive_ptr<mpipairwise> ptr;
    private:

      /// How many iterations we do.
      int iterations_, count_;

      std::vector<mpiid> from_;
      std::vector<mpiid> to_;
      std::vector<timestamp> when_;

    protected:
      /// Hi.
      mpipairwise(softwareid id) :
          mpiapp(id, to_string())
      {

      }

    public:

      /// Hello.
      static ptr
      construct(softwareid id)
      {
        return ptr(new mpipairwise(id));
      }

      /// Goodbye.
      virtual
      ~mpipairwise() throw ();

      /// Get a copy.
      app::ptr
      clone(softwareid id);

      /// Go.
      void
      skeleton_main();

      virtual void
      consume_params(const sim_parameters::ptr &params);

      virtual std::string
      to_string() const
      {
        return "mpi_pairwise";
      }
    };

  }
} //end of namespace sstmac

#endif
