/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/software/skeletons/test/test_nppfs.h>

namespace sstmac
{
  namespace sw
  {
    using namespace std;

    void
    test_nppfs::consume_params(const sim_parameters::ptr &params)
    {

      access_size_ = params->get_int_param("testfilesystem_access_size");
      seek_length_ = params->get_int_param("testfilesystem_seek_length");


      fs_ = nppfs::construct(id_);
      register_lib(fs_);
    }

    void
    test_nppfs::skeleton_main()
    {
      int rank = id_.task_.id;

      SSTMAC_DEBUG << "test_nppfs[" << rank << "]: Hello World, i'm tester number "
            << id_.task_.id << ", about to begin accessing " << access_size_
            << " bytes \n";

      openfile::ptr file = fs_->open("/usr/local/purplemonkey");
      SSTMAC_DEBUG << "test_nppfs[" << rank << "]: file is open, begin read \n";

      timestamp t1 = now();
      fs_->read(file, access_size_);
      timestamp t2 = now();

      SSTMAC_DEBUG << "test_nppfs[" << rank << "]: read took " << (t2 - t1) << "\n";
      SSTMAC_DEBUG << "test_nppfs[" << rank << "]: begin write \n";

      t1 = now();
      fs_->write(file, access_size_);
      t2 = now();

      SSTMAC_DEBUG << "test_nppfs[" << rank << "]: write took " << (t2 - t1) << "\n";

      SSTMAC_DEBUG << "test_nppfs[" << rank << "]: seeking \n";

      fs_->seek(file,  seek_length_);

      t1 = now();
      fs_->write(file, access_size_ * 2);
      t2 = now();

      SSTMAC_DEBUG << "test_nppfs[" << rank << "]: big write after seek took " << (t2 - t1)
            << "\n";

      fs_->close(file);

    }

  }
} //end of namespace sstmac
