#ifndef SDMA_ANALYSIS_SERVER_H_
#define SDMA_ANALYSIS_SERVER_H_

#include <sstmac/software/services/service.h>
#include "sdma_work_lib.h"
#include <sstmac/software/libraries/filesystem/nppfs.h>
#include <queue>
#include "sdma_message.h"

namespace sdma
{

  class sdma_analysis_server : public sstmac::sw::service
  {
  public:
    typedef boost::intrusive_ptr<sdma_analysis_server> ptr;

  protected:

    class work_handler : public eventhandler
    {
    protected:
      work_handler()
      {
        buildup_ = 0;
        factor_ = 4; //4 is totally arbitrary, can make this a parameter later
      }

    public:

      typedef boost::intrusive_ptr<work_handler> ptr;
      static ptr
      construct()
      {
        return ptr(new work_handler());
      }

      virtual
      ~work_handler()
      {
      }

      virtual std::string
      to_string() const
      {
        return "sdma_disk_server::work_handler";
      }

      virtual void
      handle(const sstmac::sst_message::ptr &msg);

      sdma_analysis_server::ptr parent_;

      long factor_;

      int buildup_;

    };

    class pfs_handler : public eventhandler
    {
    public:
      typedef boost::intrusive_ptr<pfs_handler> ptr;
      static ptr
      construct()
      {
        return ptr(new pfs_handler());
      }

      virtual
      ~pfs_handler()
      {
      }

      virtual std::string
      to_string() const
      {
        return "sdma_disk_server::pfs_handler";
      }

      virtual void
      handle(const sstmac::sst_message::ptr &msg);

      sdma_analysis_server::ptr parent_;

    };

    static sstmac::debug dbg_;

    std::queue<sdma_message::ptr> work_queue_;

    bool working_;
    sstmac::sst_message::ptr work_msg_;

    sdma_work_lib::ptr work_;
    sstmac::sw::nppfs::ptr pfs_;
    sstmac::sw::openfile::ptr resultfile_;

    sstmac::nodeaddress::ptr myaddr_;

    long to_write_;

    sdma_analysis_server()
    {
      working_ = false;

      if (!dbg_.is_real())
        {
          dbg_ = sstmac::debug(
              "<debug> sdma | sdma_server | sdma_analysis_server",
              sstmac::logger::red);
          sstmac::logger::register_logger(&dbg_);
        }

      to_write_ = 0;

    }

  public:

    virtual
    ~sdma_analysis_server()
    {
    }

    virtual std::string
    to_string() const
    {
      return "sdma_analysis_server";
    }

    virtual void
    register_child_libs()
    {
      work_ = sdma_work_lib::construct(to_string() + "work-lib");

      work_handler::ptr workh = work_handler::construct();
      ptr slf(this);
      workh->parent_ = slf;
      work_->init_handler(workh);
      work_->init_waiter(wait_);
      register_lib(work_);

      pfs_ = sstmac::sw::nppfs::construct(to_string() + "pfs");
      pfs_handler::ptr pfsh = pfs_handler::construct();
      pfsh->parent_ = slf;
      pfs_->init_handler(pfsh);
      pfs_->init_waiter(wait_);

      register_lib(pfs_);
    }

    virtual void
    handle(const sstmac::sst_message::ptr &msg);

    virtual void
    handle_pfs(const sstmac::sw::file_metadata::ptr &msg);

    virtual void
    work() = 0;

    virtual void
    send_along(long bytes, sstmac::nodeaddress::ptr next) = 0;


    virtual void
    work_done();

  };

}

#endif
