/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SDMA_APP_H_
#define SDMA_APP_H_

#include <sstmac/software/libraries/mpi/mpiapp.h>
#include <sstmac/software/libraries/mpi/sstmac_mpi.h>


#include "sdma_lib.h"


namespace sdma
{

  /**
   * App that can test out the SDMA system
   */
  class sdma_app : virtual public sstmac::sw::mpiapp
  {

  protected:
    /// Hi.
    sdma_app(sstmac::sw::softwareid id) :
        sstmac::sw::mpiapp(id, to_string())
    {

    }

  public:
    typedef boost::intrusive_ptr<sdma_app> ptr;

    /// Hello.
    static ptr
    construct(sstmac::sw::softwareid id)
    {
      return ptr(new sdma_app(id));
    }

    /// Goodbye.
    virtual
    ~sdma_app() throw ()
    {
    }

    /// Get a copy.
    virtual sstmac::sw::app::ptr
    clone(sstmac::sw::softwareid id)
    {
      return construct(id);
    }

    /// Go.
    void
    skeleton_main();

    virtual void
    consume_params(const sstmac::sim_parameters::ptr &params);

    virtual std::string
    to_string() const
    {
      return "sdma_app";
    }

    void
    init_sdma_system();

  protected:

    std::string sdma_type_;
    sdma_lib::ptr sdma_;

    long numiter_;
    long data_size_;
    timestamp sleep_;

    int stages_;
    double stage_conc_;
  };
}

#endif
