#include "sdma_disk_server.h"
#include "sdma_message.h"
#include <sstmac/sstmacro.h>
#include "sdma_lib.h"

using namespace sstmac;
using namespace sstmac::sw;

namespace sdma
{
  SSTRegisterServ("sdma_disk_server", sdma_disk_server)
  ;


  void
  sdma_disk_server::work()
  {
    if (!working_)
      {
        if (work_queue_.size() > 0)
          {
            working_ = true;
            sdma_message::ptr smsg = work_queue_.front();

            SSTMAC_DEBUG << "sdma_server[" << myaddr_->to_string() << "]: working on a go-disk message \n";

            sdma_go_disk::ptr go = boost::dynamic_pointer_cast<sdma_go_disk>(
                smsg);

            //first, get it from disk
            pfs_->read(go->get_file(), go->get_work_bytes());

          }
      }
  }

  /**
   * This gets called to handle messages from the file system library
   * @param msg
   */
  void
  sdma_disk_server::handle_pfs(const file_metadata::ptr &meta)
  {
    SSTMAC_DEBUG << "sdma_disk_server[" << myaddr_->to_string() << "]::handle_pfs \n";

    if (meta->op_ == file_metadata::CREATE_RESP) //this was to create the results file

      {
        SSTMAC_DEBUG
              << "sdma_disk_server[" << myaddr_->to_string() << "]: got a file creation resp, can now write \n";
        resultfile_ = pfs_->process_openfile_meta(meta);
        pfs_->write(resultfile_, to_write_, false);
      }
    else if (meta->op_ == file_metadata::READ_RESP) //this is for using the staged disk method

      {
        SSTMAC_DEBUG << "sdma_disk_server[" << myaddr_->to_string() << "]: got a read resp, now working \n";
        sdma_message::ptr smsg = work_queue_.front();
        sdma_go_disk::ptr go = boost::dynamic_pointer_cast<sdma_go_disk>(smsg);
        //now we can work, start the memory access
        work_->mem_access(go->get_work_bytes());
      }
    else if (meta->op_ == file_metadata::WRITE_RESP)
      { // this is for the staged disk method writing to next stage
        sdma_message::ptr smsg = work_queue_.front();
        sdma_go_disk::ptr mess =
            boost::dynamic_pointer_cast<sdma_go_disk>(smsg);
        mess->bytes_to_access_ = writing_bytes_;

        SSTMAC_DEBUG
              << "sdma_disk_server[" << myaddr_->to_string() << "]: got a write resp, notifying next stage "
              << writing_to_->to_string() << "\n";
        mess->toaddr_ = writing_to_;
        mess->set_lib_name(sdma_disk_server::get_server_name());
        mess->set_needs_ack(false);

        os_->execute_kernel(AMI::AMI_COMM, AMI::AMI_COMM_SEND, mess);

        work_done();
      }
  }

  void
  sdma_disk_server::send_along(long bytes, sstmac::nodeaddress::ptr next)
  {
    writing_bytes_ = bytes;
    writing_to_ = next;
    sdma_go_disk::ptr go = boost::dynamic_pointer_cast<sdma_go_disk>(work_queue_.front());
    pfs_->write(go->get_file(), bytes, true);

  }

}
