#ifndef SDMA_MESSAGE_H_
#define SDMA_MESSAGE_H_

#include <sstmac/common/messages/sst_message.h>
#include <sstmac/software/libraries/filesystem/openfile.h>

namespace sdma
{

  class sdma_message : public sstmac::sst_message
  {
  public:
    enum MESS_TYPE
    {
      GO_DISK_ANALYSIS, GO_STAGED_ANALYSIS
    };

  protected:
    long bytes_;  //this is how big the message is
    MESS_TYPE type_;


    sdma_message(long bytes, MESS_TYPE t, long work) :
      sst_message(sst_message::SDMA), bytes_(bytes), type_(t), bytes_to_access_(work)
    {

    }

  public:
    long bytes_to_access_;  //this is how much data we're dealing with

    typedef boost::intrusive_ptr<sdma_message> ptr;

    static ptr
    construct(long bytes, MESS_TYPE t, long work)
    {
      return ptr(new sdma_message(bytes, t, work));
    }

    virtual
    ~sdma_message()
    {
    }

    virtual long
    get_byte_length() const
    {
      return bytes_;
    }

    MESS_TYPE
    get_sdma_type()
    {
      return type_;
    }

    long
    get_work_bytes() const
    {
      return bytes_to_access_;
    }
  };

  class sdma_go_disk : public sdma_message
  {
  protected:
    sstmac::sw::openfile::ptr file_;

    sdma_go_disk(sstmac::sw::openfile::ptr file, long bytes) :
      sdma_message(32, sdma_message::GO_DISK_ANALYSIS, bytes), file_(file)
    {

    }

  public:
    typedef boost::intrusive_ptr<sdma_go_disk> ptr;

    static ptr
    construct(sstmac::sw::openfile::ptr file, long bytes)
    {
      return ptr(new sdma_go_disk(file, bytes));
    }

    virtual
    ~sdma_go_disk()
    {
    }

    sstmac::sw::openfile::ptr
    get_file() const
    {
      return file_;
    }


  };

  class sdma_go_staged : public sdma_message
  {
  protected:


    sdma_go_staged(long bytes) :
      sdma_message(bytes, sdma_message::GO_STAGED_ANALYSIS, bytes)
    {

    }

  public:
    typedef boost::intrusive_ptr<sdma_go_staged> ptr;

    static ptr
    construct(long bytes)
    {
      return ptr(new sdma_go_staged(bytes));
    }

    virtual
    ~sdma_go_staged()
    {
    }

  };

}

#endif
