#ifndef SDMA_WORK_LIB_H_
#define SDMA_WORK_LIB_H_

#include <sstmac/software/libraries/compute/lib_compute_inst.h>
#include <sstmac/software/libraries/compute/lib_compute_memmove.h>

namespace sdma
{

  class sdma_work_lib : public sstmac::sw::lib_compute_inst,
      public sstmac::sw::lib_compute_memmove
  {
    // ------- constructor / boost stuff -------------//

  protected:
    sdma_work_lib(sstmac::sw::softwareid id) :
      lib_compute_inst(id), lib_compute_memmove(id)
    {
      if (!dbg_.is_real())
        {
          dbg_ = sstmac::debug("<debug> sdma | sdma_work_lib", sstmac::logger::red);
          sstmac::logger::register_logger(&dbg_);
        }

      libname_ = "sdma_work_lib" + id.to_string();
    }

    sdma_work_lib(std::string id) :
      lib_compute_inst(id), lib_compute_memmove(id)
    {

      if (!dbg_.is_real())
        {
          dbg_ = sstmac::debug("<debug> sdma | sdma_work_lib", sstmac::logger::red);
          sstmac::logger::register_logger(&dbg_);
        }

      libname_ = id;
    }

    bool doing_memory_;

  public:
    typedef boost::intrusive_ptr<sdma_work_lib> ptr;

    virtual
    ~sdma_work_lib()
    {
    }

    static ptr
    construct(sstmac::sw::softwareid id)
    {
      return ptr(new sdma_work_lib(id));
    }

    static ptr
    construct(std::string id)
    {
      return ptr(new sdma_work_lib(id));
    }

    void
    mem_access(long bytes);

    void
    compute(long bytes);

    bool
    doing_mem() const {
      return doing_memory_;
    }
  protected:

    virtual bool
    supported()
    {
      return lib_compute_inst::supported() && lib_compute_memmove::supported();
    }

    static sstmac::debug dbg_;

  };
}

#endif
